/*
 * Window clipboard drawing
 */

/*  2002 Serge Rosmorduc,
 * quipe EC.art, Laboratoire LRIA, Universit de Paris 8
 *
 * 15 rue Catulienne
 * 93236 St Denis Cedex
 * FRANCE
 *
 * Tout droits rservs.
 * 
 * voyez le fichier ../LICENCE.txt pour plus d'informations
 * See the ../LICENCEE.txt file for more information
 *
 * $Id: TKSO_WCB.c,v 1.1.1.1 2002/04/16 13:55:00 rosmord Exp $
 *
 */

/* * $Log: TKSO_WCB.c,v $
 * * Revision 1.1.1.1  2002/04/16 13:55:00  rosmord
 * * Imported using tkCVS
 * *
 *
 */

#define _POSIX_SOURCE 1

#include "chaine.h"
#include "tgraph.h"
#include "tgraphImpl.h"
#include "TKSO_WCB.h"
#include <tk.h>
#include "general.h"
#include <malloc.h>
#include <math.h>
#include <windows.h>

#ifndef _GNUC_
#define inline
#endif

#ifndef PI
#define PI		3.14159265358979323846
#endif

#define TKSESHDEGREE (180*64.0)

#define WCBCAST(x) ((TKSO_WCB*)(x))

static void cleanP(TKSESH_output *);
static void DrawLineP(TKSESH_output *, TKSESH_point, TKSESH_point);
static void DrawLinesP(TKSESH_output *, int, TKSESH_point *);
static void fillPolygonP(TKSESH_output *, int , TKSESH_point * , int);
static void DrawArcP(TKSESH_output *, TKSESH_point, double, double, int, int);
static double String_widthP(TKSESH_output *, char *);
static void DrawStringP(TKSESH_output *, TKSESH_point, char *);
static void fill_rectangleP(TKSESH_output *, TKSESH_rectangle);

static TKSO_device
TKSO_WCB_driver= {
	"WCB",
	sizeof(TKSO_WCB),
	cleanP,
	DrawLineP,
	DrawLinesP,
	TKSO_genericHachureP,
	fillPolygonP,
	DrawArcP,
	String_widthP,
	DrawStringP,
	fill_rectangleP,
	0,
	0
};

/* TKSO_init --
	 Initialise a TKSESH_output. 
	 Doesn't set up a drawing surface 
	 (which is done by some TKSO_init_draw or some other function)
	 */

#define WCBBLACK RGB(0,0,0)
#define WCBRED RGB(255,0,0)
#define WCBWHITE RGB(255,255,255)

#define TBLACKPEN 0
#define TBLACKBRUSH 1
#define TREDPEN 2
#define TREDBRUSH 3
#define TWHITEPEN 4
#define TWHITEBRUSH 5
#define TWIDEBLACKPEN 6
#define TWIDEREDPEN 7

static void 
fix_state_fill(TKSESH_output *o);

static void 
fix_state_draw(TKSESH_output *o);



TKSO_WCB *
TKSO_WCB_init(double x1, double y1,
							double x2, double y2, double lineHeight, int ENHANCED)
{
	int w, h;
	TKSO_WCB *o; 
	RECT rect;
	LOGPEN lpen;
	LOGBRUSH lbrush;
	
	/* for mapping mode. */
	int mapMode;
	double mapModeScale;
	

	o= NOUV(TKSO_WCB);

	/* 
	 * w and h are always expressed in O.01 milimeter units
	 * (window DC convention).
	 * 
	 * lineHeight is expressed in mm. x and y here are
	 * expressed in units such that one line is 1 unit high.
	 */
	
	w= (x2-x1)* 100 * lineHeight;
	h= (y2-y1)*  100 * lineHeight;

	o->enmh= ENHANCED;
	
	if (ENHANCED) {
		/*
		 * Now, we choose a good mapping mode.
		 * the idea is that a glyph should be at least 1000 units hight.
		 * Reminder :
		 * LOMETRIC : 0.1 mm
		 * LOENGLISH : 0.01 inch = 0.0254 cm= 0.254 mm
		 * HIMETRIC : 0.01 mm
		 * HIENGLISH : 0.001 inch = 0.0254 mm
		 * TWIPS : 1/1440 inch
		 *
		 * mapModeScale converts h and w (expressed in 0.01 mm units)
		 * into actual DC units
		 *
		 */
		if (lineHeight/0.254 >= 1000) 
			{
				mapMode= MM_LOENGLISH;
				mapModeScale=  0.01 / 0.254 ;
			} 
		else if (lineHeight/0.1 >= 1000)
			{
				mapMode= MM_LOMETRIC;
				mapModeScale= 0.01 / 0.1 ;
			}
		else if (lineHeight/0.0254 >= 1000)
			{
				mapMode= MM_HIENGLISH;
				mapModeScale= 0.01 / 0.0254 ;
			}
		else /* if (lineHeight/0.01 >= 1000) */
			{
				mapMode= MM_HIMETRIC;
				mapModeScale= 1; /* 0.01 / 0.01 */
			}

		o->mappingMode= mapMode;

		rect.top= 0;
		rect.left= 0;
		rect.bottom= h;
		rect.right= w;

		o->dc= CreateEnhMetaFile(o->deviceDC, 
														 (LPTSTR) NULL,
														 &rect, 
														 (LPTSTR)NULL); 
		TKSO_init_plain(TKSOC(o),x1, y1, x2, y2, w*mapModeScale, h*mapModeScale);
	} else {
		/* plain old metafile */
		o->dc= CreateMetaFile((LPTSTR) NULL);
		mapModeScale= 1.0;
		TKSO_init_plain(TKSOC(o),x1, y1, x2, y2, w*mapModeScale, h*mapModeScale);
	}

	TKSOC(o)->isLowRes= FALSE;
	TKSODRIVER(o)= &TKSO_WCB_driver;

	/* set by TKSO_init_plain */
	/* TKSO_SET_SCALE(TKSOC(o), w*mapModeScale/(x2-x1)); */
 
	o->deviceDC= CreateDC("DISPLAY", 0, 0, 0);

	// o->olddc= SaveDC(o->dc);

	if (o->enmh)
		{
			SetMapMode(o->dc, mapMode);
		}
	else 
		{
			/* For plain old metafile, we need to use either MM_ISOTROPIC
			 * or MM_ANISOTROPIC
			 */
			
			/* the metafile shouldn't contain a call to SetMapMode */
			/* SetMapMode(o->dc, MM_ISOTROPIC);*/
			o->mappingMode= MM_ANISOTROPIC;

			/*
			 * for plain metafiles in MM_ISOTROPIC mode,
			 * calls should be made to SetWindowExtEx and SetWindowOrgEx,
			 * but not to ViewPort or setMapMode functions.
			 */
			SetWindowOrgEx(o->dc, 0, 0, NULL);
			/* picture dimensions */
			/* SetWindowExtEx(o->dc, w, h, NULL); */
			SetWindowExtEx(o->dc, w, h, NULL);
			o->mmW= w;
			o->mmH= h;
		}

	//SetMapMode(o->dc, MM_ANISOTROPIC);
	//SetWindowExtEx(o->dc, w, h, NULL);
	//SetWindowOrgEx(o->dc, 0, 0, NULL);
	//SetViewportExtEx(o->dc, w, h, NULL);

	SetROP2(o->dc, R2_COPYPEN);
	
	// WCBSetsPolyFillMode(o->dc, WINDING);
	// WCBSetMapMode(o->dc, MM_HIENGLISH);
	// WCBSetWindowExt(o->dc, w, h);
	// WCBSetMapMode(o->dc, MM_LOMETRIC);
	// WCBSetViewportExt(o->dc, w, h);
	SetBkColor(o->dc, WCBBLACK);
	SetBkMode(o->dc, TRANSPARENT);
	 
	/* Null Pen */
	lpen.lopnWidth.x = 0;
	lpen.lopnWidth.y = 0;         /* Unused in LOGPEN */
	lpen.lopnStyle = PS_NULL;    /* For now...convert 'style' in the future */
	lpen.lopnColor = WCBBLACK;       /* Assume we're getting a COLORREF */
	
	o->nullPen= CreatePenIndirect(&lpen);
	/* Black pen */
	/* CreatePenIndirect(o->dc, PS_SOLID, 0, WCBBLACK);*/
	lpen.lopnWidth.x = 1;
	lpen.lopnWidth.y = 1;         /* Unused in LOGPEN */
	lpen.lopnStyle = PS_SOLID;    /* For now...convert 'style' in the future */
	lpen.lopnColor = WCBBLACK;       /* Assume we're getting a COLORREF */
	 
	o->blackPen= CreatePenIndirect(&lpen);

	// WCBSelectObject(WCBCAST(o)->dc, 0);
	/* CreateBrushIndirect(o->dc, 0, WCBBLACK, BS_SOLID); */
	lbrush.lbStyle= BS_SOLID;
	lbrush.lbColor= WCBBLACK;
	lbrush.lbHatch= 0;
	 
	o->blackBrush= CreateBrushIndirect(&lbrush);
	// WCBSelectObject(WCBCAST(o)->dc, 1);
	 
	/* Red pen */
	lpen.lopnColor= WCBRED;
	o->redPen= CreatePenIndirect(&lpen);

	// WCBSelectObject(WCBCAST(o)->dc, 2);
	lbrush.lbColor= WCBRED;
	/* CreateBrushIndirect(o->dc, 0, WCBRED, BS_SOLID);*/
	o->redBrush= CreateBrushIndirect(&lbrush);

	/* WCBSelectObject(WCBCAST(o)->dc, 3); */

	/* White pen */	
	lpen.lopnColor= WCBWHITE;
	o->whitePen= CreatePenIndirect(&lpen);

	/* CreatePenIndirect(o->dc, PS_SOLID, 0, WCBWHITE); */

	// WCBSelectObject(WCBCAST(o)->dc, 4);
	lbrush.lbColor= WCBWHITE;
	o->whiteBrush= CreateBrushIndirect(&lbrush);

	// WCBSelectObject(WCBCAST(o)->dc, 5);
	/* wide black pen */
	/* CreatePenIndirect(o->dc, PS_SOLID, 50, WCBBLACK); */
	lpen.lopnColor= WCBBLACK;
	lpen.lopnWidth.x = 25;
	lpen.lopnWidth.y = 25;         
	o->wideBlackPen= CreatePenIndirect(&lpen);
	/* wide red pen */
	lpen.lopnColor= WCBRED;
	o->wideRedPen= CreatePenIndirect(&lpen);

	return o;
}

int
TKSO_WCB_free(TKSO_WCB *o)
{
	/* Delete all objects in o : */
	DeleteObject(o->whitePen);
	DeleteObject(o->redPen);
	DeleteObject(o->blackPen);
	DeleteObject(o->wideBlackPen);
	DeleteObject(o->wideRedPen);
	DeleteObject(o->blackBrush);
	DeleteObject(o->whiteBrush);
	DeleteObject(o->redBrush);
	
	//RestoreDC(o->dc, o->olddc);


	if (!OpenClipboard(NULL))
		{
			DeleteDC(o->deviceDC);
			return FALSE;
		}
	
	if (! EmptyClipboard()) {
		DeleteDC(o->deviceDC);
		CloseClipboard();
		return FALSE;
	}

	if (o->enmh)
		{
			HENHMETAFILE metafile;
      HENHMETAFILE hcopy; /* I don't think we need to copy metafile */
			metafile= CloseEnhMetaFile(o->dc);
			hcopy= CopyEnhMetaFile(metafile, NULL);
			DeleteEnhMetaFile(metafile);
			SetClipboardData(CF_ENHMETAFILE, hcopy);
		}
	else 
		{
			GLOBALHANDLE   hGlobal ;
      LPMETAFILEPICT pMFP ;

			HMETAFILE metafile= CloseMetaFile(o->dc);
			
      hGlobal= GlobalAlloc (GHND | GMEM_SHARE, sizeof (METAFILEPICT)) ;
      pMFP = (LPMETAFILEPICT) GlobalLock (hGlobal) ;
			
			/*
			 * if mapping mode is ISOTROPIC, positive xExt and yExt
			 * are picture size in 0.01 mm units
			 */
			
      pMFP->mm   = o->mappingMode ;
      pMFP->xExt = o->mmW ;
      pMFP->yExt = o->mmH ;
      pMFP->hMF  = metafile  ;
			
      GlobalUnlock (hGlobal) ;
			
      SetClipboardData (CF_METAFILEPICT, hGlobal) ;
		}
	CloseClipboard(); 
	DeleteDC(o->deviceDC);
	return TRUE;
}

WCBpoint
TKSESH_point_to_WCBpoint(TKSO_WCB *o, TKSESH_point p)
{
	WCBpoint m;
	
	m.x= (int)((p.x - TKSOWIN(o).bottom_left.x)* TKSOSCALE(o));
	m.y= (int)((p.y - TKSOWIN(o).top_right.y)* TKSOSCALE(o));
	if ((o->mappingMode == MM_ISOTROPIC) || (o->mappingMode == MM_ANISOTROPIC)) {
		m.y= (int)((TKSOWIN(o).top_right.y-p.y)* TKSOSCALE(o));
	}
	return m;
}

static void 
cleanP(TKSESH_output * o)
{
	/* Do nothing */
}

static void 
DrawLineP(TKSESH_output * o, TKSESH_point a, TKSESH_point b)
{
	WCBpoint a1, b1;
	a1= TKSESH_point_to_WCBpoint(WCBCAST(o), a);
	b1= TKSESH_point_to_WCBpoint(WCBCAST(o), b);
	MoveToEx(WCBCAST(o)->dc, a1.x, a1.y, NULL);
	fix_state_draw(o);
	LineTo(WCBCAST(o)->dc, b1.x, b1.y);
}

static void 
DrawLinesP(TKSESH_output * o, int np, TKSESH_point * l)
{
	WCBpoint m;
	int i;
	static Chaine chp= NULL;

	fix_state_draw(o);
	if (chp == empty_chaine) 
		init_Chaine(&chp, sizeof(WCBpoint));
	
	for (i=0; i< np; i++)
		{
			m= TKSESH_point_to_WCBpoint(WCBCAST(o), l[i]);
			add_to_Chaine(chp, &m);
		}
	
	PolylineTo(WCBCAST(o)->dc, Chaine_to_data(chp), np);
	reset_Chaine(chp);
}

/* NOT USED. 
 * First written for WMF.
 * The following code would have been the logical path. Unfortunately,
 * WMF Hatching is very poorly supported in many systems.
 *

 static void 
	 hachureP(TKSESH_output * o, TKSESH_rectangle rect)
	 {
	 WMFpoint bottomleft, topright;
	 bottomleft= TKSESH_point_to_WMFpoint(WMFCAST(o), rect.bottom_left);
	 topright= TKSESH_point_to_WMFpoint(WMFCAST(o), rect.top_right);
	 WMFSelectObject(WMFCAST(o)->wmf, TWHITEPEN);
	 WMFSelectObject(WMFCAST(o)->wmf, TSHADEBRUSH);
	 WMFSetBKMode(WMFCAST(o)->wmf, TRANSPARENT);
	 WMFRectangle(WMFCAST(o)->wmf, bottomleft.x, topright.y, topright.x, bottomleft.y);
	 WMFSetBKMode(WMFCAST(o)->wmf, OPAQUE);
	 }
*/

static void 
fillPolygonP(TKSESH_output * o, int nl, TKSESH_point * l, int sens)
{
	WCBpoint m;
	int i;
	TKSO_state s= o->state;
	static Chaine chp= NULL;

	if (chp == empty_chaine) 
		init_Chaine(&chp, sizeof(WCBpoint));
	
	if (!sens)
		TKSO_set_state(o, white);
	fix_state_fill(o);
 
	for (i=0; i< nl; i++)
		{
			m= TKSESH_point_to_WCBpoint(WCBCAST(o), l[i]);
			add_to_Chaine(chp, &m);
		}
	
	Polygon(WCBCAST(o)->dc, Chaine_to_data(chp), nl);
	reset_Chaine(chp);
	TKSO_set_state(o, s);
}

/*
 * See tgraph.h for more info on DrawArcP
 */

static void 
DrawArcP(TKSESH_output * o, TKSESH_point topLeft, double w, double h, int a1, int a2)
{
	/*
	 * Well, this one is a bit tricky.
	 */
	TKSESH_point center, bottomRight, start, end;
	WCBpoint topleft1, bottomright1, start1, end1;
	
	/* 
	 * We start getting rid of SESHfixed. 386 computers without
	 * FPU are getting uncommon
	 * NOW, SESHfixed == double
	 */
	center.x= topLeft.x + w/2;
	center.y= topLeft.y - h/2;
	
	bottomRight.x = center.x + w / 2;
	bottomRight.y= center.y - h / 2;
	/* xstart, ystart define a line from c with angle a1.
	 * For example, 
	 * xstart= cos(a1) * w + center.x
	 * ystart= sin(a1) * w + center.y (note that the w is NOT an error)
	 * of course, then, a1 is expressed as degrees * 64.
	 */
	start.x= cos((a1/TKSESHDEGREE)*PI) * w/2 + center.x;
	start.y= sin((a1/TKSESHDEGREE)*PI) * h/2 + center.y;
	end.x=  cos(((a1+a2)/TKSESHDEGREE)*PI) * w/2 + center.x;
	end.y= sin(((a1+a2)/TKSESHDEGREE)*PI) * h/2 + center.y;
	/* Now, all we have to to is to translate these points */
	topleft1= TKSESH_point_to_WCBpoint(WCBCAST(o), topLeft);
	bottomright1= TKSESH_point_to_WCBpoint(WCBCAST(o), bottomRight);
	start1= TKSESH_point_to_WCBpoint(WCBCAST(o), start);
	end1= TKSESH_point_to_WCBpoint(WCBCAST(o), end);
	/* We draw the arc */

	fix_state_draw(o);
	Arc(WCBCAST(o)->dc,
			topleft1.x, topleft1.y,
			bottomright1.x, bottomright1.y,
			start1.x, start1.y,
			end1.x, end1.y);
}

static double 
String_widthP(TKSESH_output * o, char * txt)
{
	return 0.0;
}

static void 
DrawStringP(TKSESH_output * o, TKSESH_point p, char * txt)
{
	
}

static void 
fill_rectangleP(TKSESH_output * o, TKSESH_rectangle r)
{
	/* NO OP NOW */
}

static void 
fix_state_draw(TKSESH_output *o)
{
	switch(o->state)
		{
		case white :
			SelectObject(WCBCAST(o)->dc, WCBCAST(o)->whitePen);
			break;
		case red :
			SelectObject(WCBCAST(o)->dc, WCBCAST(o)->wideRedPen);
			break;
		case gram :
		case gramend :
		case blue :
		case aux :
		case grayfill :
		case invert :
		case black:
			SelectObject(WCBCAST(o)->dc, WCBCAST(o)->wideBlackPen);
			break;
		}
}

static void 
fix_state_fill(TKSESH_output *o)
{
	switch(o->state)
		{
		case white :
			SelectObject(WCBCAST(o)->dc, WCBCAST(o)->nullPen);
			SelectObject(WCBCAST(o)->dc, WCBCAST(o)->whiteBrush);
			break;
		case red :
			SelectObject(WCBCAST(o)->dc, WCBCAST(o)->nullPen);
			SelectObject(WCBCAST(o)->dc, WCBCAST(o)->redBrush);
			break;
		case gram :
		case gramend :
		case blue :
		case aux :
		case grayfill :
		case invert :
		case black:
			SelectObject(WCBCAST(o)->dc, WCBCAST(o)->nullPen);
			SelectObject(WCBCAST(o)->dc, WCBCAST(o)->blackBrush);
			break;
		}
}

/* Let's preserve the present presentation
 * Local Variables:
 * tab-width: 2
 * c-block-comments-indent-p: nil
 * End:
 */
