/**
 * Functions to transform an internal structure into manuel de codage code
 */

/*  2000 Serge Rosmorduc,
 * quipe EC.art, Laboratoire LRIA, Universit de Paris 8
 *
 * 15 rue Catulienne
 * 93236 St Denis Cedex
 * FRANCE
 *
 * Tout droits rservs.
 * 
 * voyez le fichier ../LICENCE.txt pour plus d'informations
 * See the ../LICENCEE.txt file for more information
 *
 * $Id: cad2text.c,v 1.1.1.1 2002/04/16 13:55:00 rosmord Exp $
 *
 */

/*
 * $Log: cad2text.c,v $
 * Revision 1.1.1.1  2002/04/16 13:55:00  rosmord
 * Imported using tkCVS
 *
 */


#define _POSIX_SOURCE 1

#include <stdio.h>

#ifndef __MWERKS__	
/* #include <varargs.h>  */
#endif
 
#include <strings.h>

#include "hache.h"
#include "duplique.h"
#include "quadra.h"
#include "general.h"
#include "bigstr.h"

/**********************************************************************
																BEWARE 
								The two following macroes are used to
							ease the reading of the text and to ensure
											 malloced memory is freed

											 SO:
YOU SHOULD USE APP2BSC(x) when x is STATIC data.
YOU SHOULD USE APP2BS(x) when x has been malloced and thus should be freed

**********************************************************************/

#define APP2BSC(x) append_to_BigString(bs,(x))
#define APP2BS(x)  append_to_BigString(bs,(ttt=(x))); LIBERE(ttt)

enum direction { droite_a_gauche, gauche_a_droite};


PRIVATE char *
qd2txt(Quadra *q,	int externe, int protected, int breakit, int red, int shaded);

PRIVATE char *
affiche_un_hieroglyphe(char * s,
											 Typehachure hachure,
											 int revert,
											 int size, 
											 int rotation,
											 int ignore,
											 int islong
											 )
{
	static char result[100];

	strcpy(result, s);

	if (revert && rotation == 0)
		strcat(result,"\\");
	if (size != 100)
	{
		char taille[6];
		sprintf(taille,"\\%i", size);
		strcat(result, taille);
	}
	if (ignore)
		strcat(result, "\\i");
	if (islong == 1)
		strcat(result, "\\l");
	if (islong == 0)
		strcat(result, "\\l0");
	if (revert && rotation != 0)
	{
		char rotate[6];
		switch (rotation)
			{
			case 90:
				strcpy(rotate, "\\t1");
				break;
			case 180:
				strcpy(rotate, "\\t2");
				break;
			case 270:
				strcpy(rotate, "\\t3");
				break;
			case 360:
				strcpy(rotate, "\\t4");
				break;
			default:
				sprintf(rotate,"\\T%i", rotation);
			}
		strcat(result, rotate);
	}
	
	if (!revert && rotation != 0)
	{
		char rotate[6];
		switch (rotation)
			{
			case 90:
				strcpy(rotate, "\\r1");
				break;
			case 180:
				strcpy(rotate, "\\r2");
				break;
			case 270:
				strcpy(rotate, "\\r3");
				break;
			case 360:
				strcpy(rotate, "\\r4");
				break;
			default:
				sprintf(rotate,"\\R%i", rotation);
			}
		strcat(result, rotate);
	}
	switch(hachure)
	{
	case S_aucune:
		break;
	case S_ga:
		strcat(result, "#//");
		break;
	case S_ha:
		strcat(result, "#h/");
		break;
	case S_va: 
		strcat(result, "#v/");
		break;
	case S_ta:
		strcat(result, "#/");
		break;
	}
	return dupliquechaine(result);
}

PRIVATE char *
signe2text(Signe *s, int externe)
{
	BigString *bs;
	char *result, *ttt;

	bs= create_BigString();

	if (S_EST_GRAMM(*s))
		APP2BSC("=");

	switch(S_TYPE_OF(*s))
	{
	case S_hieroglyphe:
		APP2BS(affiche_un_hieroglyphe(
																	S_REPRESENTATION_OF(*s),
																	S_TYPE_HACHURE_OF(*s),
																	S_REVERT_OF(*s),
																	S_SIZE_OF(*s),
																	S_ROTATION_OF(*s),
																	S_IGNORE_OF(*s),
																	S_LONG_OF(*s)
																	));
		break;
	case S_overwrite: 
		{
			APP2BS(signe2text(S_FIRST_SIGN_OF(*s), externe));
			APP2BSC("##");
			APP2BS(signe2text(S_SECOND_SIGN_OF(*s), externe));
			break;
		}
		/* case S_ligature:
			 APP2BS(affiche_un_hieroglyphe(
			 S_LIG_NOM_OF(*s),
			 S_TYPE_HACHURE_OF(*s),
			 S_REVERT_OF(*s),
			 S_SIZE_OF(*s),
			 S_ROTATION_OF(*s)
			 ));
			 break;
		*/
	case S_inconnu:
		APP2BSC(affiche_un_hieroglyphe(
																	 S_TEXTE_OF(*s),
																	 S_TYPE_HACHURE_OF(*s),
																	 S_REVERT_OF(*s),
																	 S_SIZE_OF(*s),
																	 S_ROTATION_OF(*s),
																	 S_IGNORE_OF(*s),
																	 S_LONG_OF(*s)
																	));
		break;
	case S_romain:								/* passage en fonte romaine */
		APP2BSC("+l");
		APP2BSC(S_TEXTE_OF(*s));
		APP2BSC("+s");
		break;
	case S_textesuper:						/* texte au dessus d'un trait */
		APP2BSC("|");
		APP2BSC(S_TEXTE_OF(*s));
		break;
	case S_pointrouge:
		APP2BS(affiche_un_hieroglyphe(
																	"o",
																	S_TYPE_HACHURE_OF(*s),
																	S_REVERT_OF(*s),
																	S_SIZE_OF(*s),
																	S_ROTATION_OF(*s),
																	S_IGNORE_OF(*s),
																	S_LONG_OF(*s)
																	));
		break;
	case S_pointnoir:
		APP2BS(affiche_un_hieroglyphe(
																	"O",
																	S_TYPE_HACHURE_OF(*s),
																	S_REVERT_OF(*s),
																	S_SIZE_OF(*s),
																	S_ROTATION_OF(*s),
																	S_IGNORE_OF(*s),
																	S_LONG_OF(*s)
																	));
	 break;
	case S_haplographie:
		break;
	case S_gauchedroite:
		break;
	case S_droitegauche :
		break;
	case S_lacune :
		break;
	case S_lignelacune :
		break;
	case S_point:
		APP2BSC(".");
		break;
	case S_pointpoint:
		APP2BSC("..");
		break;
	case S_hachureg:
		APP2BSC("//");
		break;
	case S_hachureh:
		APP2BSC("h/");
		break;
	case S_hachurev:
		APP2BSC("v/");
		break;
	case S_hachuret:
		APP2BSC("/");
		break;
	}
	if (S_EST_FIN_MOT(*s))
		add_to_BigString(bs,' ');
	dissolve_BigString_in_string(&bs, &result);
	return result;
}

PRIVATE char *
lig2text(struct Liggroup *l, int externe, int breakit)
{
	BigString *bs;
	char * result;
	char *ttt;
	struct _Signinlig *s;

	bs= create_BigString();
	s= l->signs;
	while(s)
		{
			APP2BS(signe2text((s->sign),externe));
			s= s->next;
			if (s)
				{
					if (!breakit)
						APP2BSC("&");
					else
						APP2BSC("\n");
				}
		}
	dissolve_BigString_in_string(&bs, &result);
	return result;
}

/*
 * Prints piles of quadras or signs :
 *      A:B:C
 *
 */

/*
	If breakit is TRUE and if the text can be broken, it will be.
	*/

PRIVATE char *
vertlist2text(Vertlist *h, int externe, int breakit, int red, int shaded)
{
	BigString *bs;
	char *ttt;
	char * result;
	
	bs= create_BigString();

	if ((S_NEXT_OF(*h) == NULL))	/* un seul tage */
		{
			APP2BS(qd2txt(VL_QUADRA_OF(*h), externe, FALSE, breakit, red, shaded));
		}
	else													/* empilement */
		{			
			char charsep= ':';
			Vertlist *t= h;

			if (breakit)
				{
					charsep= '\n';
					breakit= FALSE;
				}

			if (!externe)
				{
					add_to_BigString(bs,'(');					
				}
		
			while (t)
				{
					int i;

					for (i=0; i<t->kerning; i++);	/* supressed */
					APP2BS(qd2txt(VL_QUADRA_OF(*t),FALSE, FALSE, FALSE, red, shaded));
					t=S_NEXT_OF(*t);								/* this next is vertical */
					if (t)
						{
							add_to_BigString(bs,charsep);
						}
				}
			if (!externe)
				{
					add_to_BigString(bs,')');
				}
		}
	dissolve_BigString_in_string(&bs, &result);
	return result;
}

#define QUADRA_PARENTHESE(DEBUT,externe,FIN) \
if (!breakit) {\
APP2BSC(DEBUT); \
APP2BS(qd2txt(QD_QUADRA_OF(*q), externe, TRUE, FALSE, red, shaded));\
APP2BSC(FIN); } \
else \
{ APP2BS(qd2txt(QD_QUADRA_OF(*q), externe, TRUE, FALSE, red, shaded)); breakit= FALSE;}

/*
 * externe means the quadra is not inside another one
 * Its sole purporse is to add parenthesis if needed.
 *
 * if breakit is true, then the most external separation
 * symbol, be it ':' or '*' will be made into a '\n'
 * to allow easy manipulation
 *
 */

/*
 * If breakit is true, it WILL destroy something if possible.
 *
 * This means :
 * breakit will stay TRUE until it receives a reason to become FALSE
 *
 * protected means the quadra can use '-' instead of '*' because it's
 * protected by parenthesis.
 *  
 */

PRIVATE char *
qd2txt(Quadra *q,	int externe, int protected, int breakit, int red, int externalShade)
{
	BigString *bs;
	char *ttt;
	char * result;
	char sepchar =  '*';
	int shaded;

	bs= create_BigString();

	if (protected)
		sepchar= '-';

	if (q->next)
		{
			if (!protected) externe= FALSE;
			if (breakit)
				{
					sepchar= '\n';
					breakit= FALSE;
				}
		}

	shaded= externalShade;
	
	while (q)
		{
			if (protected && q->est_grise && !shaded)
				{
					APP2BSC("-#b-");
					shaded= TRUE;
				}
			if (protected && !q->est_grise && shaded)
				{
					APP2BSC("-#e-");
					shaded= FALSE;
				}

			switch(S_TYPE_OF(*q))
				{
				case S_quadra:
					APP2BS(vertlist2text(QD_VERTLIST_OF(*q), externe, breakit, red, shaded));
					break;
					/* les constructions de type cartouche */
				case S_carto :
					QUADRA_PARENTHESE("<-", TRUE, "->");
					break;
				case S_debcarto:
					QUADRA_PARENTHESE("<b-", TRUE, "->");
					break;
				case S_milcarto :
					QUADRA_PARENTHESE("<m-", TRUE, "->");
					break;
				case S_fincarto :
					QUADRA_PARENTHESE("<e-", TRUE, "->");
					break;
				case S_serekh :
					QUADRA_PARENTHESE("<S-", TRUE, "->");
					break;
				case S_enceinte :
					QUADRA_PARENTHESE("<F-", TRUE, "->");
					break;
				case S_chateau :
					QUADRA_PARENTHESE("<H-", TRUE, "->");
					break;
				case S_debserekh :
					QUADRA_PARENTHESE("<Sb-", TRUE, "->");
					break;
				case S_milserekh :
					QUADRA_PARENTHESE("<Sm-", TRUE, "->");
					break;
				case S_finserekh :
					QUADRA_PARENTHESE("<Se-", TRUE, "->");
					break;
				case S_debenceinte :
					QUADRA_PARENTHESE("<Fb-", TRUE, "->");
					break;
				case S_milenceinte :
					QUADRA_PARENTHESE("<Fm-", TRUE, "->");
					break;
				case S_finenceinte :
					QUADRA_PARENTHESE("<Fe-", TRUE, "->");
					break;
				case S_debchateau :
					QUADRA_PARENTHESE("<Hb-", TRUE, "->");
					break;
				case S_milchateau :
					QUADRA_PARENTHESE("<Hm-", TRUE, "->");
					break;
				case S_finchateau :
					QUADRA_PARENTHESE("<He-", TRUE, "->");
					break;
					/*
						case S_enrouge:
						QUADRA_PARENTHESE("$-", FALSE, "-$");
						break;
						*/
				case S_rightleft:
					QUADRA_PARENTHESE("+rl", TRUE, "+lr");
					break;
				case S_superfetatoire:
					QUADRA_PARENTHESE("[{-", TRUE, "-}]");
					break;
				case S_efface:
					QUADRA_PARENTHESE("[[-", TRUE, "-]]");
					break;
				case S_disparu:
					QUADRA_PARENTHESE("[\"-", TRUE, "-\"]");
					break;
				case S_ajoutscribe:
					QUADRA_PARENTHESE("['-", TRUE, "-']");
					break;
				case S_ajoutauteur:
					QUADRA_PARENTHESE("[&-", TRUE, "-&]");
					break;
				case S_signe:
					APP2BS(signe2text((QD_SIGNE_OF(*q)), externe));
					break;
				case S_ligature:
					APP2BS(lig2text(QD_LIG_OF(*q), externe, breakit));
				}
			
			/* if (externe && q->hachures != NOHACHURE)*/
			if (q->hachures != NOHACHURE)
				{
					APP2BSC("#");
					if (q->hachures&HACHURE1)
						APP2BSC("1");
					if (q->hachures&HACHURE2)
						APP2BSC("2");
					if (q->hachures&HACHURE3)
						APP2BSC("3");
					if (q->hachures&HACHURE4)
						APP2BSC("4");
				}

			if (q->next)
				add_to_BigString(bs,sepchar);
			
			q= q-> next;
		}
	if (shaded && ! externalShade) {
		APP2BSC("-#e-");
	}
	dissolve_BigString_in_string(&bs, &result);
	return result;
}

/**
 * PUBLIC char *
 * quad2text(Quadra *q)
 *
 * Sends back a string in MDC format representing q
 *
 * breakit : should we break the upper level construct ?
 * red : are we in a red TEXT ?
 * shaded : are we in a shaded TEXT ?
 */

PUBLIC char *
quad2text(Quadra *q, int breakit, int red, int shaded)
{
	return qd2txt(q, TRUE, FALSE, breakit, red, shaded);
}

/*
 * Local Variables:
 * header: "quadra.h"
 * tab-width: 2
 * End:
 */
