
#define _POSIX_SOURCE 1

#include <string.h>
#include <limits.h>
#include <stdio.h>

#include "duplique.h"
#include "quadra.h"
#include "mdctexte.h"
#include "seshaux.h"
#include "mynew.h"

DECLARE_STACK(Texte)
IMPLEMENTE_STACK(Texte)

		 
#ifndef NULL
#define NULL (void *)0
#endif

#ifndef TRUE
#define TRUE 1
#define FALSE 0
#endif

/* Declaration of auxiliary functions */

PRIVATE char *
SESH_texte_to_string_aux(GeneralIndex gdebut, GeneralIndex gfin, char sep);

PRIVATE void
montexte_nl(Texte *t)
{
	t->type= T_NL;
	/* t->fini= TRUE;*/
}


/**
 *
 * PUBLIC Texte * 
 * NewTexte()
 * 
 */

PUBLIC Texte * 
NewTexte()
{
  Texte *res;
  res= NEW(Texte);
	res->type= T_TEXTE;
	res->etat= E_NOUVEAU;
	res->couleur= black;
  res->q= NULL;
	res->real_dims= FALSE;
	res->scaleh= 1.0;
	res->scalev= 1.0;
  res->x=0.0;
  res->y= 0;
  res->wh=0;
  res->hh=0;
  res->wv=0;
  res->hv=0;
	res->saut= 0;
	res->texte=NULL;
	res->first=FALSE;
	res->compteur= ULONG_MAX;
  res->next=NULL;
  res->previous=NULL;
	res->liste_des_marques= NULL;
	res->lessignes= NULL;
	res->selected= FALSE;
  return res;
}


/**
	 completely frees a text. Do not use to delete a text.
	 Only to free the text when deleting a sesh widget.
 */

PUBLIC void
FreeTexte(Texte *tt)
{
	if (tt) {
		Texte *t1;		
		t1= tt->next;
		while (t1 != NULL) 
			{
				DeleteTexte(t1);
				t1= tt->next;
			}
		DELETE(Texte,tt);
	}
}

/** 
		Destruction of a text entry. 
		All marks are moved to the previous
		text item 
*/

PUBLIC void
DeleteTexte(Texte *tt)
{
	Liste_marques *l;

	SESH_detruire_listesignes(tt->lessignes);
	detruit_quadra(tt->q);
	if (tt->texte)
		LIBERE(tt->texte);
	l= tt->liste_des_marques;
	while (l)
		{
			Liste_marques *l1;
			Texte * newhost;
			l1= l->next;
			
			/*
			 *	now, l is the current mark, and we are free to 
			 * move it as we like since l->next is kept in l1->next.
			 */
			if (strcmp(l->m->nom, "ecran") == 0)
				{
					newhost= tt->previous;
					while (newhost->type != T_NL &&
								 newhost->type != T_NP)
						newhost= newhost->previous;
				}
			else
				newhost= tt->previous;
			
			if (newhost)
				{
					if (newhost->liste_des_marques)
						newhost->liste_des_marques->previous= l;
					l->next= newhost->liste_des_marques;
					newhost->liste_des_marques= l;
					l->previous= NULL;
					l->m->index= SESH_make_general_index_to(newhost,0);
				}
			else
				{
					fprintf(stderr," y'a un shmu dans le potage !!!!!\n");
					LIBERE(l);
				}
			l= l1;
		}
	if (tt->previous)
		{
			tt->previous->next= tt->next;
		}
	if (tt->next)
		tt->next->previous= tt->previous;
	DELETE(Texte,tt);
}

/**
 */
PUBLIC Texte *
MDC_FirstTexte()
{
	Texte *t;

	t= NewTexte();
	
	montexte_nl(t);
	t->first= TRUE;

	return t;
}


/**
 */
PUBLIC void
add_page_after(Texte *txt)
{
	Texte *t;

	t= NewTexte();
	t->type= T_NP;

	t->previous= txt;
	t->next= txt->next;
	if (t->next)
		t->next->previous= t;
	t->previous->next=t;
}	

PRIVATE int											/* A AMELIORER !!!!!!!!!!!! */
bon_sens(Texte *debut, Texte *fin)
{
	int ARRIERE= TRUE;
	Texte *tt= debut;
	
	while (ARRIERE && tt)
	{
		if ((tt == fin))
			ARRIERE= FALSE;
		tt= tt->next;
	}
	if ((tt == fin))
		ARRIERE= FALSE;
	return !ARRIERE;
}

PRIVATE void
swap_Texte_ptr(Texte **a, Texte**b)
{
	Texte *tmp;
	tmp= *a;
	*a=*b;
	*b=tmp;
}

/**
 * PUBLIC void 
 * SESH_ordonner_Texte(Texte **a, Texte **b)
 * 
 * ensure that 'a' is before 'b' in the text, swapping them
 * if needed.
 */

PUBLIC void 
SESH_ordonner_Texte(Texte **a, Texte **b)
{
	if (! bon_sens(*a, *b))
		swap_Texte_ptr(a, b);
}

/**
	For details, see detruire_Texte, below
	*/

PRIVATE Texte *
local_detruire_Texte(Texte *debut, Texte *fin)
{
	Texte * tt;
	Texte * arret;
	Texte* nouveau; 	

	SESH_ordonner_Texte(&debut, &fin);
	
	if (!debut->next)
	{
		nouveau= debut;
		return nouveau;
	}

	tt= debut->next;
	arret= fin->next;		/* fin is eventually destroyed */
	while (tt != arret)
	{
		Texte *tnext;
		tnext= tt->next;
		DeleteTexte(tt);
		tt= tnext;
	}
	nouveau= debut;
	return nouveau;
}


/**
 * PUBLIC Texte *
 * detruire_Texte(Texte *debut, Texte *fin)
 * 
 * destroys the text between two Texte;
 * it returns a  safe spot in front of the destroyed zone.
 * The first text (be it debut or fin) is NOT destroyed.
 * graphically :
 *
 * debut                  fin
 *     |..................|
 *      *******************
 *      ^^^^^^^^^^^^^^^^^^^
 *        this is deleted 
 * The rationale is that a zone between Texte * 
 * designates text AFTER the first one up to (and including
 * the last one.
 */

PUBLIC Texte *
detruire_Texte(GeneralIndex gdebut, GeneralIndex gfin)
{
	return local_detruire_Texte(GETTEXTOF(gdebut), GETTEXTOF(gfin));
}


/**
 * PUBLIC Texte *
 * ajoute_texte_apres(Texte *montexte)
 **	ajoute du texte aprs un autre texte. Si le second argument est la
 ** ligne courante (et non NULL) alors, si le nouveau texte rajout
 ** devient la nouvelle fin de ligne, cela est gr. 
 *********************************************************************/

PUBLIC Texte *
ajoute_texte_apres(Texte *montexte)
{
	Texte *t;
	t= NewTexte();
	t->next= montexte->next;
	t->previous= montexte;
	montexte->next=t;
	if (t->next)
		t->next->previous= t;
	return t;
}


/**
 */
PUBLIC void
SESH_init_textes()
{
	INIT_STACK(Texte);
}
/* top level functions */

#include "bigstr.h"
#include "general.h"


/*
 *  makes a list of quadras into a displayable structure
 *  returns position of last quadra
 */


 
/**
	Inserts text after gori.
 */

PUBLIC Texte *
SESH_insere_texte(GeneralIndex gori, Texte *insere)
{
	Texte *fin; 
	Texte *t= insere;
	Texte *apres;
	
	fin= GETTEXTOF(gori);
	apres= (fin)->next;
	while (t)
	{
		t->previous= fin;
		(fin)->next= t;
		fin= t;
		t= t->next;
	}
	(fin)->next= apres;
	if (apres)
		apres->previous= fin;
	return fin;
}


/**
	Sends back a string representing g where the uppermost
	construct has been broken. Former separator are replaced by '\n'
	(newline).
	e.g. : 
     pt:p*t => p\np*t
	   p*t => p\nt
		 <-ra-ms-s-s-> => ra*ms*s*s
 */

PUBLIC char *
SESH_breaktext(GeneralIndex g)
{
	Texte *t;
	t= GETTEXTOF(g);
	if (t->q)
		{
			return quad2text(t->q, TRUE, FALSE, FALSE);
		}
	else
		{
			GeneralIndex g1;
			g1= SESH_previous_index(g);
			return SESH_texte_to_string(g,g1);
		}
}

/**
	sends back a string representing the text between debut and fin
	with Text separators ('-') replaced by newlines.
	*/

PUBLIC char *
SESH_breakspan(GeneralIndex debut, GeneralIndex fin)
{
	return SESH_texte_to_string_aux(debut, fin, '\n');
}

/**
	returns true if g is a pile of quadras,
	and else, false.
	BEWARE :
	  mn:n*w is a pile,
		mn*(n:t) is not !
 */
	
PUBLIC int
SESH_is_vertical(GeneralIndex g)
{
	Texte *t;
	t= GETTEXTOF(g);
	return (t 
					&& t->q 
					&& S_TYPE_OF(*(t->q)) == S_quadra
					&& S_NEXT_OF(*QD_VERTLIST_OF(*(t->q))) != NULL);
}

/**
 */
PUBLIC char *
SESH_texte_to_string(GeneralIndex debut, GeneralIndex fin)
{
	return SESH_texte_to_string_aux(debut, fin, '-');
}

PRIVATE void
SESH_fin_rouge_gris_aux(BigString *bs, 
												int finrouge, /* end of a red section ? */
												int fingris, /* end of a gray section ? */
												int enrouge, /* IN a red section ? */
												int engris,	/*  IN a gray section ?*/
												int rouge, /* distance to beginning of red section */
												int gris,	/* distance to beginning of gray section */
												int islast)	/* at end of text (then a '-' is needed)*/
{
	if (fingris && finrouge)
		{
			if (islast)
				append_to_BigString(bs, "-");
			if (rouge > gris)
				{
					append_to_BigString(bs, "#e-");
					append_to_BigString(bs, "$b-");
				}
			else
				{
					append_to_BigString(bs, "$b-");
					append_to_BigString(bs, "#e-");
				}
		}
	else
		{
			if (finrouge)
				{
					if (islast)
						append_to_BigString(bs, "-");

					/*
						If we are in a "gray" section, 
						 we need to close it iff
						 it started AFTER the red section
						 (that is, rouge > gris)
					*/
						 
					if (engris && rouge > gris)
						append_to_BigString(bs, "#e-");

					append_to_BigString(bs, "$b-");

					if (engris && rouge > gris)
						append_to_BigString(bs, "#b-");
				}
			if (fingris)
				{
					if (islast)
						append_to_BigString(bs, "-");

					if (enrouge  && rouge <= gris)
						append_to_BigString(bs, "$b-");
					
					append_to_BigString(bs, "#e-");

					if (enrouge  && rouge <= gris)
						append_to_BigString(bs, "$r-");
				}
		}
}

PRIVATE char *
SESH_texte_to_string_aux(GeneralIndex gdebut, GeneralIndex gfin, char sep)
{
	BigString *bs;
	char *result;
	Texte *t;
	Texte *debut, *fin;
	int enrouge= FALSE;
	int grise= FALSE;
	int dernier_rouge=0;
	int dernier_grise=0;
	
	debut= GETTEXTOF(gdebut);
	fin= GETTEXTOF(gfin);
	
	if (! bon_sens(debut,fin))
		swap_Texte_ptr(&debut, &fin);
	
	if (!debut->next)
		return NULL;

	bs= create_BigString();
	t= debut->next;

	while(t != fin->next)
	{
		/* Seuls les hiroglyphes peuvent etre griss ou en rouge ;
		 *  en particulier, les hachures ne peuvent passer une fin de ligne */
		if (t->type != T_TEXTE)
			{
				SESH_fin_rouge_gris_aux(bs, enrouge, grise, 
																FALSE, FALSE,
																dernier_rouge, dernier_grise, FALSE);
				enrouge= FALSE;
				grise= FALSE;
			}
		switch (t->type)
		{
		case T_NL:
			append_to_BigString(bs, "!\n");
			break;
		case T_NP:
			append_to_BigString(bs, "!!\n");
			break;
		case T_LIGNE_COMMENTAIRE:
			append_to_BigString(bs, "++TXT");
			append_to_BigString(bs, t->texte);
			append_to_BigString(bs, "+s\n");
			break;
		case T_TEXTE:
		{
			char *s;
			SESH_fin_rouge_gris_aux(bs, 
															!EST_ROUGE(t) && enrouge, 
															!EST_GRIS(t) && grise, 
															EST_ROUGE(t) && enrouge,
															EST_GRIS(t) && grise,
															dernier_rouge, dernier_grise, FALSE);
			if (EST_GRIS(t) && !grise)
				{
					grise= TRUE;
					append_to_BigString(bs, "#b-");
					dernier_grise= 0;
					dernier_rouge++;
				}
			if (EST_ROUGE(t) && !enrouge)
				{
					enrouge= TRUE;
					append_to_BigString(bs, "$r-");
					dernier_rouge=0;
					dernier_grise++;
				}
			if (!EST_ROUGE(t) && enrouge)
				{
					enrouge= FALSE;
				}
			if (!EST_GRIS(t) && grise)
				{
					grise= FALSE;
				}
			
			s= (char *)quad2text(t->q, FALSE, enrouge, grise);
			append_to_BigString(bs, s);
			LIBERE(s);
		}
			break;
		case T_FIXE_COORDONNEES:
		{
			char *s;
			s= (char *)ALLOUER((strlen(t->coordonnees.descr)+12 + 1)* sizeof(char));
			sprintf(s, "++COORDS=%s+s\n", t->coordonnees.descr);
			append_to_BigString(bs,s);
			LIBERE(s);
		}
			break;
		}	/* switch (t->type) */
		if (t->next != fin->next)
			add_to_BigString(bs, sep);		
		t= t->next;
	}
	SESH_fin_rouge_gris_aux(bs, enrouge, grise,
													FALSE, FALSE,
													dernier_rouge, dernier_grise, TRUE);
	dissolve_BigString_in_string(&bs, &result);
	return result;
}




/**
 */
PUBLIC void 
SESH_set_Texte_texte(Texte *t, char * texte)
{
	if (!t)
		return;
	if (t->texte)
		LIBERE(t->texte);
	t->texte= dupliquechaine(texte);
}



/* 
 * Fonctions de parcours des cadras pour retrouver les signes
 */

/**
	PUBLIC char *
	SESH_listesignes(GeneralIndex gdebut, GeneralIndex gfin, int mode)
	
	List of signs between gdebut and gfin
	mode allows to get more information. Information for each
	sign is given as a list, the first element of whom is the 
	sign's code.
	
	mode :
	  SIGNS_ONLY => list of the signs
		SIGNS_EOW  => for each sign which is a word ending, indicates it by 
		              appending the tag "WORD" after the sign.
		SIGNS_GRAM => append the tag "GRAM" if sign belongs to a grammatical ending
		
		the order of the tags is always 
		    SIGN  ["IGNORE"] ["GRAM"] ["WORD"]

		WE SHOULD MAKE THIS LIST ORIENTED
 */

PUBLIC char *
SESH_listesignes(GeneralIndex gdebut, GeneralIndex gfin, int options)
{
	BigString *bs;
	char *result;
	Texte *t;
	Texte *debut, *fin;
	int SPACE_NEEDED;
	
	debut= GETTEXTOF(gdebut);
	fin= GETTEXTOF(gfin);
	
	if (! bon_sens(debut,fin))
		swap_Texte_ptr(&debut, &fin);
	
	if (!debut->next)
		return NULL;

	bs= create_BigString();
	t= debut->next;

	SPACE_NEEDED= FALSE;

	while(t != fin->next)
	{	
		SignePix p;

		if (SESH_nieme_signe(&p, t, 1)) {
			int stillgood;
			if (SPACE_NEEDED)
				append_to_BigString(bs, " ");
			SPACE_NEEDED= TRUE;
			do {
				Signe *s;
				int addword= FALSE;
				int addgram= FALSE;
				if (p.l && (s= p.l->signe)) {
					addword= (options & SIGNS_EOW) && s->estfinmot;
					addgram= (options & SIGNS_GRAM) && s->estgram;
				}
				if (! s->ignore || (options & SIGNS_IGNORE)) {
					if (addword || addgram)
						append_to_BigString(bs, "{");
					append_to_BigString(bs, SESH_get_signe_rep_from_pix(p));
					if (addword)
						append_to_BigString(bs, " WORD");
					if (addgram)
						append_to_BigString(bs, " GRAM");
					if (addword || addgram)
						append_to_BigString(bs, "}");
				}
				stillgood= SESH_signe_suivant(&p);
				if (stillgood)
					append_to_BigString(bs, " ");
			} while (stillgood);
		} 
		t=t->next;
	}
	dissolve_BigString_in_string(&bs, &result);
	return result;
}
		
/**
 *
 * PUBLIC int
 * SESH_nombres_de_signes_dans_Texte(Texte *t);
 *
 * returns the number of signs in a quadra
 */

PUBLIC int
SESH_nombres_de_signes_dans_Texte(Texte *t)
{
	return SESH_length_listesignes(t->lessignes);
}

/**
 * Sends back a SignePix handler to the first  sign
 * RETURNS TRUE if success, else FALSE 
 */

PUBLIC int
SESH_premier_signe(SignePix *s, Texte *t)
{
	if (t->lessignes)
			s->l= t->lessignes;
	else
		s->l= NULL;
	return (s->l != NULL);
}

/**
 * Sends back a SignePix handler to the nth sign
 * RETURNS TRUE if success, else FALSE 
 */

PUBLIC int
SESH_nieme_signe(SignePix *s, Texte *t, int n)
{
	s->l= NULL;
	if (t->q)
			s->l= SESH_listesignes_nth(t->lessignes, n);
	return (s->l != FALSE);
}

/**
 * Have p point to the next sign
 * RETURNS TRUE if success, else FALSE 
 */

PUBLIC int
SESH_signe_suivant(SignePix *p)
{
	if (p->l)
		p->l= p->l->next;
	return (p->l != FALSE);
}


/**
 * Get back a pointer to the Sign pointed by p
 * or NULL if p is invalid
 */

PRIVATE Signe *
SESH_get_signe_from_pix(SignePix p)
{
	if (p.l && p.l->signe)
		return p.l->signe;
	else 
		return NULL;
}

PRIVATE Signe *
SESH_get_signe_from_index(GeneralIndex g)
{
	SignePix p;
	p= GETSIGNOF(g);
	return SESH_get_signe_from_pix(p);
}



/**
 *
 * BEWARE : the value returned should NOT be written to
 */

PUBLIC char *
SESH_get_signe_rep_from_pix(SignePix p)
{
	Signe *s;
	if ((s= SESH_get_signe_from_pix(p)))
		{
			switch (s->type) {
			case S_hieroglyphe : 
			case S_inconnu:
				return s->donnees.hieroglyphe.rep;
				break;
			default:
				return NULL;
			}
		}
	else return NULL;
}

/*********************************************************************
 *   information about individual signs 
 *
 * functions with names ending in 'p' are query functions
 *
 * functions like SESH_set_sign_YYY are used to set the value of YYY
 * property. The last argument (what) can be :
 *
 * 0 : property is unset
 * 1 : property is set
 * -1 : toggle property.
 * 
 */

PRIVATE int * 
SESH_Sign_Property_Ptr(GeneralIndex g, Signproperty prop)
{
	Signe *s;
	int *res= 0;
	s= SESH_get_signe_from_index(g);
	if (s) {
		switch (prop)
			{
			case Psize:
				res= &(s->size);
				break;
			case Protation:
				res= &(s->rotation);
				break;
			case Phachure:
				res= (int*)&(s->typehachure);
				break;
			case Prevert:
				res= &(s->revert);
				break;
			case Pword:
				res= &(s->estfinmot);
				break;
			case Pgram:
				res= &(s->estgram);
				break;
			case Pselect:
				res= &(s->selected);
				break;
			case Pignore:
				res= &(s->ignore);
				break;
			case Ptall:
				res= &(s->istall);
				break;
			case Plong:
				res= &(s->islong);
				break;
			}
	}
	return res;
}

/**
	 get a sign's property value
 */

PUBLIC int
SESH_sign_get_p(GeneralIndex g, Signproperty prop)
{
	int *val;
	val= SESH_Sign_Property_Ptr(g,prop);
	if (val)
		return  *val;
	else
		return 0;
}

/**
 * SESH_set_sign_property --
 * set or toggle the value of a sign's property.
 *
 * GeneralIndex g : the sign
 * Signproperty prop : the property to change
 * what : the property new value (an integer). 
 *        If the property is boolean,
 *        -1 means "toggle"
 *        
 *        Chars '1', '0', and 'c' can also be used.
 */
PUBLIC void
SESH_set_sign_property(GeneralIndex g, Signproperty prop, int what)
{
	int *val;
 	val= SESH_Sign_Property_Ptr(g,prop);
	if (val)
		{
			Texte *t;
			t= GETTEXTOF(g);
			switch (prop) {
			case Psize:
			case Protation:
			case Phachure:
				*val= what;							/* scalar properties */
				break;
			case Pignore:
			case Ptall:
			case Plong:
			case Prevert:
			case Pword:
			case Pgram:
			case Pselect:							/* boolean properties */
				switch (what)
					{
					case -1:
					case 'c':
						*val= 1 - *val;
						break;
					case 's':
						*val= 2;
						break;
					case '1':
						*val= 1;
					break;
					case '0':
						*val= 0;
						break;
					default:
						*val= (what != 0); /* ensures estfinmot is 0 or 1 */
					}
			}
			t->real_dims= FALSE;			/* quadra dimensions might be false */
			t->etat= E_NOUVEAU;				/* The quadra has changed and should be redrawn */
		}
}


PUBLIC void
SESH_set_selected(GeneralIndex gdebut, GeneralIndex gfin, int selectedp)
{
	Texte *debut, *fin;
	Texte * tt;

	debut= GETTEXTOF(gdebut);
	fin= GETTEXTOF(gfin);

	SESH_ordonner_Texte(&debut, &fin);
	
	if (!debut->next)
		return;

	tt= debut->next;
	while (tt != fin->next)
	{
		tt->selected= selectedp;
		tt->etat= E_NOUVEAU;				/* The quadra has changed and should be redrawn */
		tt= tt->next;
	}
}

/**
 * public GeneralIndex 
 * SESH_make_general_index_to(Texte *t, int signe)
 * 
 * Sends back an GeneralIndex to a position.
 * t is the cadrat
 * signe is the number of the sign in t
 * 0 means no real sign (the whole item)
 */

PUBLIC GeneralIndex 
SESH_make_general_index_to(Texte *t, int signe)
{
	GeneralIndex g;
	SignePix p;
	g.t= t;
	/* g.s= -1; Why -1 ??????? */
	g.s= 0;
	g.signe.l= 0;

	if (SESH_nieme_signe(&p, t, signe))
		{
			g.s= signe;
			g.signe= p;
			return g;
		}
	else if (SESH_nieme_signe(&p, t, 0))
		{
			g.s= 0;
			g.signe= p;
			return g;
		}
	return g;
}

/**
 */
PUBLIC int
SESH_not_last_index(GeneralIndex g)
{
	return (g.t->next != NULL);
}

/**
	moving in the Text
 */
PUBLIC GeneralIndex 
SESH_next_index(GeneralIndex g)
{
	if (g.t && g.t->next)
		{
			GeneralIndex gg;
			gg.t= g.t->next;
			gg.s= 0;
			return gg;
		}
	else
		return g;
}

/**
	moving in the Text
 */
PUBLIC GeneralIndex 
SESH_previous_index(GeneralIndex g)
{
	if (g.t && g.t->previous)
		{
			GeneralIndex gg;
			gg.t= g.t->previous;
			gg.s= 0;
			return gg;
		}
	else
		return g;
}

/*
 * Local Variables:
 * header: "MDCtexte.h"
 * tab-width: 2
 * c-block-comments-indent-p: nil
 * End:
 */
