/*
 * Canvas sesh item
 */

/*  2000, 2001 Serge Rosmorduc,
 * quipe EC.art, Laboratoire LRIA, Universit de Paris 8
 *
 * 15 rue Catulienne
 * 93236 St Denis Cedex
 * FRANCE
 *
 * Tout droits rservs.
 * 
 * voyez le fichier ../LICENCE.txt pour plus d'informations
 * See the ../LICENCEE.txt file for more information
 *
 * $Id: seshitem.c,v 1.1.1.1 2002/04/16 13:55:00 rosmord Exp $
 *
 * Code skeletton taken from the Text Canvas Item.
 * The Text Canvas item is :
 ** tkCanvText.c --
 **
 **	This file implements text items for canvas widgets.
 **
 ** Copyright (c) 1991-1994 The Regents of the University of California.
 ** Copyright (c) 1994-1997 Sun Microsystems, Inc.
 **
 ** See the file "license.terms" for information on usage and redistribution
 ** of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 **
 ** RCS: @(#)  tkCanvText.c,v 1.8 1999/12/21 23:55:10 hobbs Exp
 *
 */

/*
 * $Log: seshitem.c,v $
 * Revision 1.1.1.1  2002/04/16 13:55:00  rosmord
 * Imported using tkCVS
 *
 */


/* #include <stdio.h>
   #include "tkInt.h"
   #include "tkPort.h"
   #include "tkCanvas.h"
*/

/* see Tk_CreateItemType manual for gory details */

/*
 * One Sesh Item :
 *
 */

typedef struct SeshItemS {
  Tk_Item header;		/* Header, mandatory in this position */
  double x, y;			/* Position */
  Tk_Anchor anchor;		/* Anchor relative to x,y */
  /* Data below is specific to Sesh Items */
} SeshItem ;

#include <stdio.h>
#include "tkInt.h"
#include "tkCanvas.h"
#include "tkPort.h"
#include "default.h"

/*
 * The structure below defines the record for each text item.
 */

typedef struct SeshItemItem  {
  Tk_Item header;		/* Generic stuff that's the same for all
										 * types.  MUST BE FIRST IN STRUCTURE. */
  
  Tk_CanvasSeshItemInfo *textInfoPtr;
	/* Pointer to a structure containing
	 * information about the selection and
	 * insertion cursor.  The structure is owned
	 * by (and shared with) the generic canvas
	 * code. */
  /*
   * Fields that are set by widget commands other than "configure".
   */
  
  double x, y;		/* Positioning point for text. */
  int insertPos;		/* Character index of character just before
										 * which the insertion cursor is displayed. */
  
  /*
   * Configuration settings that are updated by Tk_ConfigureWidget.
   */
  
  Tk_Anchor anchor;		/* Where to anchor text relative to (x,y). */
  Tk_TSOffset tsoffset;
  XColor *color;		/* Color for text. */
  XColor *activeColor;	/* Color for text. */
  XColor *disabledColor;	/* Color for text. */
  Tk_Font tkfont;		/* Font for drawing text. */
  Tk_Justify justify;		/* Justification mode for text. */
  Pixmap stipple;		/* Stipple bitmap for text, or None. */
  Pixmap activeStipple;	/* Stipple bitmap for text, or None. */
  Pixmap disabledStipple;	/* Stipple bitmap for text, or None. */
  
  Texte *text;			/* SeshItem for item (malloc-ed). */
  int width;			/* Width of lines for word-wrap, pixels.
									 * Zero means no word-wrap. */
  
  /*
   * Fields whose values are derived from the current values of the
   * configuration settings above.
   */
  
	TKSO_tkwin * output;

  int numChars;		/* Length of text in characters. */
  int numBytes;		/* Length of text in bytes. */
  Tk_SeshItemLayout textLayout;	/* Cached text layout information. */
  int leftEdge;		/* Pixel location of the left edge of the
									 * text item; where the left border of the
									 * text layout is drawn. */
  int rightEdge;		/* Pixel just to right of right edge of
										 * area of text item.  Used for selecting up
										 * to end of line. */
  GC gc;			/* Graphics context for drawing text. */
  GC selSeshItemGC;		/* Graphics context for selected text. */
  GC cursorOffGC;		/* If not None, this gives a graphics context
										 * to use to draw the insertion cursor when
										 * it's off.  Used if the selection and
										 * insertion cursor colors are the same.  */
} SeshItemItem;

/*
 * Information used for parsing configuration specs:
 */

static Tk_CustomOption stateOption = {
  (Tk_OptionParseProc *) TkStateParseProc,
  TkStatePrintProc, (ClientData) 2
};
static Tk_CustomOption tagsOption = {
  (Tk_OptionParseProc *) Tk_CanvasTagsParseProc,
  Tk_CanvasTagsPrintProc, (ClientData) NULL
};
static Tk_CustomOption offsetOption = {
  (Tk_OptionParseProc *) TkOffsetParseProc,
  TkOffsetPrintProc, (ClientData) (TK_OFFSET_RELATIVE)
};

static Tk_ConfigSpec configSpecs[] = {
  {TK_CONFIG_COLOR, "-activefill", (char *) NULL, (char *) NULL,
   (char *) NULL, Tk_Offset(SeshItemItem, activeColor), TK_CONFIG_NULL_OK},
  {TK_CONFIG_BITMAP, "-activestipple", (char *) NULL, (char *) NULL,
   (char *) NULL, Tk_Offset(SeshItemItem, activeStipple), TK_CONFIG_NULL_OK},
  {TK_CONFIG_ANCHOR, "-anchor", (char *) NULL, (char *) NULL,
   "center", Tk_Offset(SeshItemItem, anchor),
   TK_CONFIG_DONT_SET_DEFAULT},
  {TK_CONFIG_COLOR, "-disabledfill", (char *) NULL, (char *) NULL,
   (char *) NULL, Tk_Offset(SeshItemItem, disabledColor), TK_CONFIG_NULL_OK},
  {TK_CONFIG_BITMAP, "-disabledstipple", (char *) NULL, (char *) NULL,
   (char *) NULL, Tk_Offset(SeshItemItem, disabledStipple), TK_CONFIG_NULL_OK},
  {TK_CONFIG_COLOR, "-fill", (char *) NULL, (char *) NULL,
   "black", Tk_Offset(SeshItemItem, color), TK_CONFIG_NULL_OK},
  {TK_CONFIG_FONT, "-font", (char *) NULL, (char *) NULL,
   DEF_CANVTEXT_FONT, Tk_Offset(SeshItemItem, tkfont), 0},
  {TK_CONFIG_JUSTIFY, "-justify", (char *) NULL, (char *) NULL,
   "left", Tk_Offset(SeshItemItem, justify),
   TK_CONFIG_DONT_SET_DEFAULT},
  {TK_CONFIG_CUSTOM, "-offset", (char *) NULL, (char *) NULL,
   "0,0", Tk_Offset(SeshItemItem, tsoffset),
   TK_CONFIG_DONT_SET_DEFAULT, &offsetOption},
  {TK_CONFIG_CUSTOM, "-state", (char *) NULL, (char *) NULL,
   (char *) NULL, Tk_Offset(Tk_Item, state), TK_CONFIG_NULL_OK,
   &stateOption},
  {TK_CONFIG_BITMAP, "-stipple", (char *) NULL, (char *) NULL,
   (char *) NULL, Tk_Offset(SeshItemItem, stipple), TK_CONFIG_NULL_OK},
  {TK_CONFIG_CUSTOM, "-tags", (char *) NULL, (char *) NULL,
   (char *) NULL, 0, TK_CONFIG_NULL_OK, &tagsOption},
  {TK_CONFIG_STRING, "-text", (char *) NULL, (char *) NULL,
   "", Tk_Offset(SeshItemItem, text), 0},
  {TK_CONFIG_PIXELS, "-width", (char *) NULL, (char *) NULL,
   "0", Tk_Offset(SeshItemItem, width), TK_CONFIG_DONT_SET_DEFAULT},
  {TK_CONFIG_END, (char *) NULL, (char *) NULL, (char *) NULL,
   (char *) NULL, 0, 0}
};

/*
 * Prototypes for procedures defined in this file:
 */

static void		ComputeSeshItemBbox _ANSI_ARGS_((Tk_Canvas canvas,
																							 SeshItemItem *textPtr));
static int		ConfigureSeshItem _ANSI_ARGS_((Tcl_Interp *interp,
																						 Tk_Canvas canvas, Tk_Item *itemPtr, int argc,
																						 Tcl_Obj *CONST argv[], int flags));
static int		CreateSeshItem _ANSI_ARGS_((Tcl_Interp *interp,
																					Tk_Canvas canvas, struct Tk_Item *itemPtr,
																					int argc, Tcl_Obj *CONST argv[]));
static void		DeleteSeshItem _ANSI_ARGS_((Tk_Canvas canvas,
																					Tk_Item *itemPtr, Display *display));
static void		DisplayCanvSeshItem _ANSI_ARGS_((Tk_Canvas canvas,
																							 Tk_Item *itemPtr, Display *display, Drawable dst,
																							 int x, int y, int width, int height));
static int		GetSelSeshItem _ANSI_ARGS_((Tk_Canvas canvas,
																					Tk_Item *itemPtr, int offset, char *buffer,
																					int maxBytes));
static int		GetSeshItemIndex _ANSI_ARGS_((Tcl_Interp *interp,
																						Tk_Canvas canvas, Tk_Item *itemPtr,
																						Tcl_Obj *obj, int *indexPtr));
static void		ScaleSeshItem _ANSI_ARGS_((Tk_Canvas canvas,
																				 Tk_Item *itemPtr, double originX, double originY,
																				 double scaleX, double scaleY));
static void		SetSeshItemCursor _ANSI_ARGS_((Tk_Canvas canvas,
																						 Tk_Item *itemPtr, int index));
static int		SeshItemCoords _ANSI_ARGS_((Tcl_Interp *interp,
																					Tk_Canvas canvas, Tk_Item *itemPtr,
																					int argc, Tcl_Obj *CONST argv[]));
static void		SeshItemDeleteChars _ANSI_ARGS_((Tk_Canvas canvas,
																							 Tk_Item *itemPtr, int first, int last));
static void		SeshItemInsert _ANSI_ARGS_((Tk_Canvas canvas,
																					Tk_Item *itemPtr, int beforeThis, char *string));
static int		SeshItemToArea _ANSI_ARGS_((Tk_Canvas canvas,
																					Tk_Item *itemPtr, double *rectPtr));
static double		SeshItemToPoint _ANSI_ARGS_((Tk_Canvas canvas,
																						 Tk_Item *itemPtr, double *pointPtr));
static int		SeshItemToPostscript _ANSI_ARGS_((Tcl_Interp *interp,
																								Tk_Canvas canvas, Tk_Item *itemPtr, int prepass));
static void		TranslateSeshItem _ANSI_ARGS_((Tk_Canvas canvas,
																						 Tk_Item *itemPtr, double deltaX, double deltaY));

/*
 * The structures below defines the rectangle and oval item types
 * by means of procedures that can be invoked by generic item code.
 */

Tk_ItemType tkSeshItemType = {
  "text",			/* name */
  sizeof(SeshItemItem),		/* itemSize */
  CreateSeshItem,			/* createProc */
  configSpecs,		/* configSpecs data*/
  ConfigureSeshItem,		/* configureProc */
  SeshItemCoords,			/* coordProc */
  DeleteSeshItem,			/* deleteProc */
  DisplayCanvSeshItem,		/* displayProc */
  TK_CONFIG_OBJS,		/* flags */
  SeshItemToPoint,		/* pointProc */
  SeshItemToArea,			/* areaProc */
  SeshItemToPostscript,		/* postscriptProc */
  ScaleSeshItem,			/* scaleProc */
  TranslateSeshItem,		/* translateProc */
  (Tk_ItemIndexProc *) GetSeshItemIndex,/* indexProc */
  SetSeshItemCursor,		/* icursorProc */
  GetSelSeshItem,			/* selectionProc */
  SeshItemInsert,			/* insertProc */
  SeshItemDeleteChars,		/* dSeshItemProc */
  (Tk_ItemType *) NULL,	/* nextPtr */
};

/*
 *--------------------------------------------------------------
 *
 * CreateSeshItem --
 *
 *	This procedure is invoked to create a new sesh item
 *	in a canvas.
 *
 * Results:
 *	A standard Tcl return value.  If an error occurred in
 *	creating the item then an error message is left in
 *	the interp's result;  in this case itemPtr is left uninitialized
 *	so it can be safely freed by the caller.
 *
 * Side effects:
 *	A new text item is created.
 *
 *--------------------------------------------------------------
 */

static int
CreateSeshItem(interp, canvas, itemPtr, argc, argv)
     Tcl_Interp *interp;		/* Interpreter for error reporting. */
     Tk_Canvas canvas;		/* Canvas to hold new item. */
     Tk_Item *itemPtr;		/* Record to hold new item; header has been
													 * initialized by caller. */
     int argc;			/* Number of arguments in argv. */
     Tcl_Obj *CONST argv[];	/* Arguments describing rectangle. */
{
  SeshItemItem *SeshPtr = (SeshItemItem *) itemPtr;
  int i;
  
  if (argc==1) {
    i = 1;
  } else {
    char *arg = Tcl_GetStringFromObj(argv[1], NULL);
    if ((argc>1) && (arg[0] == '-')
				&& (arg[1] >= 'a') && (arg[1] <= 'z')) {
      i = 1;
    } else {
      i = 2;
    }
  }
  
  if (argc < i) {
    Tcl_AppendResult(interp, "wrong # args: should be \"",
										 Tk_PathName(Tk_CanvasTkwin(canvas)), " create ",
										 itemPtr->typePtr->name, " x y ?options?\"", (char *) NULL);
    return TCL_ERROR;
  }
  
  /*
   * Carry out initialization that is needed in order to clean up after
   * errors during the the remainder of this procedure.
   */
  
  textPtr->textInfoPtr = Tk_CanvasGetSeshItemInfo(canvas);
  
  textPtr->insertPos	= 0;
  
  textPtr->anchor	= TK_ANCHOR_CENTER;
  textPtr->tsoffset.flags = 0;
  textPtr->tsoffset.xoffset = 0;
  textPtr->tsoffset.yoffset = 0;
  textPtr->color	= NULL;
  textPtr->activeColor = NULL;
  textPtr->disabledColor = NULL;
  textPtr->tkfont	= NULL;
  textPtr->justify	= TK_JUSTIFY_LEFT;
  textPtr->stipple	= None;
  textPtr->activeStipple = None;
  textPtr->disabledStipple = None;
  textPtr->text	= NULL;
  textPtr->width	= 0;
  
  textPtr->numChars	= 0;
  textPtr->numBytes	= 0;
  textPtr->textLayout = NULL;
  textPtr->leftEdge	= 0;
  textPtr->rightEdge	= 0;
  textPtr->gc		= None;
  textPtr->selSeshItemGC	= None;
  textPtr->cursorOffGC = None;
  
  /*
   * Process the arguments to fill in the item record.
   */
  
  if ((SeshItemCoords(interp, canvas, itemPtr, i, argv) != TCL_OK)) {
    goto error;
  }
  if (ConfigureSeshItem(interp, canvas, itemPtr, argc-i, argv+i, 0) == TCL_OK) {
    return TCL_OK;
  }
  
 error:
  DeleteSeshItem(canvas, itemPtr, Tk_Display(Tk_CanvasTkwin(canvas)));
  return TCL_ERROR;
}

/*
 *--------------------------------------------------------------
 *
 * SeshItemCoords --
 *
 *	This procedure is invoked to process the "coords" widget
 *	command on text items.  See the user documentation for
 *	details on what it does.
 *
 * Results:
 *	Returns TCL_OK or TCL_ERROR, and sets the interp's result.
 *
 * Side effects:
 *	The coordinates for the given item may be changed.
 *
 *--------------------------------------------------------------
 */

static int
SeshItemCoords(interp, canvas, itemPtr, argc, argv)
     Tcl_Interp *interp;		/* Used for error reporting. */
     Tk_Canvas canvas;		/* Canvas containing item. */
     Tk_Item *itemPtr;		/* Item whose coordinates are to be read or
													 * modified. */
     int argc;			/* Number of coordinates supplied in argv. */
     Tcl_Obj *CONST argv[];	/* Array of coordinates: x1, y1, x2, y2, ... */
{
  SeshItemItem *textPtr = (SeshItemItem *) itemPtr;
  
  if (argc == 0) {
    Tcl_Obj *obj = Tcl_NewObj();
    Tcl_Obj *subobj = Tcl_NewDoubleObj(textPtr->x);
    Tcl_ListObjAppendElement(interp, obj, subobj);
    subobj = Tcl_NewDoubleObj(textPtr->y);
    Tcl_ListObjAppendElement(interp, obj, subobj);
    Tcl_SetObjResult(interp, obj);
  } else if (argc < 3) {
    if (argc==1) {
      if (Tcl_ListObjGetElements(interp, argv[0], &argc,
																 (Tcl_Obj ***) &argv) != TCL_OK) {
				return TCL_ERROR;
      } else if (argc != 2) {
				char buf[64 + TCL_INTEGER_SPACE];
				
				sprintf(buf, "wrong # coordinates: expected 2, got %d", argc);
				Tcl_SetResult(interp, buf, TCL_VOLATILE);
				return TCL_ERROR;
      }
    }
    if ((Tk_CanvasGetCoordFromObj(interp, canvas, argv[0], &textPtr->x) != TCL_OK)
				|| (Tk_CanvasGetCoordFromObj(interp, canvas, argv[1],
																		 &textPtr->y) != TCL_OK)) {
      return TCL_ERROR;
    }
    ComputeSeshItemBbox(canvas, textPtr);
  } else {
    char buf[64 + TCL_INTEGER_SPACE];
    
    sprintf(buf, "wrong # coordinates: expected 0 or 2, got %d", argc);
    Tcl_SetResult(interp, buf, TCL_VOLATILE);
    return TCL_ERROR;
  }
  return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * ConfigureSeshItem --
 *
 *	This procedure is invoked to configure various aspects
 *	of a Sesh item, such as its border and background colors.
 *
 * Results:
 *	A standard Tcl result code.  If an error occurs, then
 *	an error message is left in the interp's result.
 *
 * Side effects:
 *	Configuration information, such as colors and stipple
 *	patterns, may be set for itemPtr.
 *
 *--------------------------------------------------------------
 */

static int
ConfigureSeshItem(interp, canvas, itemPtr, argc, argv, flags)
     Tcl_Interp *interp;		/* Interpreter for error reporting. */
     Tk_Canvas canvas;		/* Canvas containing itemPtr. */
     Tk_Item *itemPtr;		/* Rectangle item to reconfigure. */
     int argc;			/* Number of elements in argv.  */
     Tcl_Obj *CONST argv[];	/* Arguments describing things to configure. */
     int flags;			/* Flags to pass to Tk_ConfigureWidget. */
{
  SeshItemItem *textPtr = (SeshItemItem *) itemPtr;
  XGCValues gcValues;
  GC newGC, newSelGC;
  unsigned long mask;
  Tk_Window tkwin;
  Tk_CanvasSeshItemInfo *textInfoPtr = textPtr->textInfoPtr;
  XColor *selBgColorPtr;
  XColor *color;
  Pixmap stipple;
  Tk_State state;
  
  tkwin = Tk_CanvasTkwin(canvas);
  if (Tk_ConfigureWidget(interp, tkwin, configSpecs, argc, (char **) argv,
												 (char *) textPtr, flags|TK_CONFIG_OBJS) != TCL_OK) {
    return TCL_ERROR;
  }
  
  /*
   * A few of the options require additional processing, such as
   * graphics contexts.
   */
  
  state = itemPtr->state;
  
  if (textPtr->activeColor != NULL ||
      textPtr->activeStipple != None) {
    itemPtr->redraw_flags |= TK_ITEM_STATE_DEPENDANT;
  } else {
    itemPtr->redraw_flags &= ~TK_ITEM_STATE_DEPENDANT;
  }
  
  if(state == TK_STATE_NULL) {
    state = ((TkCanvas *)canvas)->canvas_state;
  }
  
  color = textPtr->color;
  stipple = textPtr->stipple;
  if (((TkCanvas *)canvas)->currentItemPtr == itemPtr) {
    if (textPtr->activeColor!=NULL) {
      color = textPtr->activeColor;
    }
    if (textPtr->activeStipple!=None) {
      stipple = textPtr->activeStipple;
    }
  } else if (state==TK_STATE_DISABLED) {
    if (textPtr->disabledColor!=NULL) {
      color = textPtr->disabledColor;
    }
    if (textPtr->disabledStipple!=None) {
      stipple = textPtr->disabledStipple;
    }
  }
  
  newGC = newSelGC = None;
  if (textPtr->tkfont != NULL) {
    gcValues.font = Tk_FontId(textPtr->tkfont);
    mask = GCFont;
    if (color != NULL) {
      gcValues.foreground = color->pixel;
      mask |= GCForeground;
      if (stipple != None) {
				gcValues.stipple = stipple;
				gcValues.fill_style = FillStippled;
				mask |= GCStipple|GCFillStyle;
      }
      newGC = Tk_GetGC(tkwin, mask, &gcValues);
    }
    mask &= ~(GCTile|GCFillStyle|GCStipple);
    if (stipple != None) {
      gcValues.stipple = stipple;
      gcValues.fill_style = FillStippled;
      mask |= GCStipple|GCFillStyle;
    }
    gcValues.foreground = textInfoPtr->selFgColorPtr->pixel;
    newSelGC = Tk_GetGC(tkwin, mask|GCForeground, &gcValues);
  }
  if (textPtr->gc != None) {
    Tk_FreeGC(Tk_Display(tkwin), textPtr->gc);
  }
  textPtr->gc = newGC;
  if (textPtr->selSeshItemGC != None) {
    Tk_FreeGC(Tk_Display(tkwin), textPtr->selSeshItemGC);
  }
  textPtr->selSeshItemGC = newSelGC;
  
  selBgColorPtr = Tk_3DBorderColor(textInfoPtr->selBorder);
  if (Tk_3DBorderColor(textInfoPtr->insertBorder)->pixel
      == selBgColorPtr->pixel) {
    if (selBgColorPtr->pixel == BlackPixelOfScreen(Tk_Screen(tkwin))) {
      gcValues.foreground = WhitePixelOfScreen(Tk_Screen(tkwin));
    } else {
      gcValues.foreground = BlackPixelOfScreen(Tk_Screen(tkwin));
    }
    newGC = Tk_GetGC(tkwin, GCForeground, &gcValues);
  } else {
    newGC = None;
  }
  if (textPtr->cursorOffGC != None) {
    Tk_FreeGC(Tk_Display(tkwin), textPtr->cursorOffGC);
  }
  textPtr->cursorOffGC = newGC;
  
  
  /*
   * If the text was changed, move the selection and insertion indices
   * to keep them inside the item.
   */
  
  textPtr->numBytes = strlen(textPtr->text);
  textPtr->numChars = Tcl_NumUtfChars(textPtr->text, textPtr->numBytes);
  if (textInfoPtr->selItemPtr == itemPtr) {
    
    if (textInfoPtr->selectFirst >= textPtr->numChars) {
      textInfoPtr->selItemPtr = NULL;
    } else {
      if (textInfoPtr->selectLast >= textPtr->numChars) {
				textInfoPtr->selectLast = textPtr->numChars - 1;
      }
      if ((textInfoPtr->anchorItemPtr == itemPtr)
					&& (textInfoPtr->selectAnchor >= textPtr->numChars)) {
				textInfoPtr->selectAnchor = textPtr->numChars - 1;
      }
    }
  }
  if (textPtr->insertPos >= textPtr->numChars) {
    textPtr->insertPos = textPtr->numChars;
  }
  
  ComputeSeshItemBbox(canvas, textPtr);
  return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * DeleteSeshItem --
 *
 *	This procedure is called to clean up the data structure
 *	associated with a text item.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Resources associated with itemPtr are released.
 *
 *--------------------------------------------------------------
 */

static void
DeleteSeshItem(canvas, itemPtr, display)
     Tk_Canvas canvas;		/* Info about overall canvas widget. */
     Tk_Item *itemPtr;		/* Item that is being deleted. */
     Display *display;		/* Display containing window for canvas. */
{
  SeshItemItem *textPtr = (SeshItemItem *) itemPtr;
  
  if (textPtr->color != NULL) {
    Tk_FreeColor(textPtr->color);
  }
  if (textPtr->activeColor != NULL) {
    Tk_FreeColor(textPtr->activeColor);
  }
  if (textPtr->disabledColor != NULL) {
    Tk_FreeColor(textPtr->disabledColor);
  }
  Tk_FreeFont(textPtr->tkfont);
  if (textPtr->stipple != None) {
    Tk_FreeBitmap(display, textPtr->stipple);
  }
  if (textPtr->activeStipple != None) {
    Tk_FreeBitmap(display, textPtr->activeStipple);
  }
  if (textPtr->disabledStipple != None) {
    Tk_FreeBitmap(display, textPtr->disabledStipple);
  }
  if (textPtr->text != NULL) {
    ckfree(textPtr->text);
  }
  
  Tk_FreeSeshItemLayout(textPtr->textLayout);
  if (textPtr->gc != None) {
    Tk_FreeGC(display, textPtr->gc);
  }
  if (textPtr->selSeshItemGC != None) {
    Tk_FreeGC(display, textPtr->selSeshItemGC);
  }
  if (textPtr->cursorOffGC != None) {
    Tk_FreeGC(display, textPtr->cursorOffGC);
  }
}

/*
 *--------------------------------------------------------------
 *
 * ComputeSeshItemBbox --
 *
 *	This procedure is invoked to compute the bounding box of
 *	all the pixels that may be drawn as part of a text item.
 *	In addition, it recomputes all of the geometry information
 *	used to display a text item or check for mouse hits.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The fields x1, y1, x2, and y2 are updated in the header
 *	for itemPtr, and the linePtr structure is regenerated
 *	for itemPtr.
 *
 *--------------------------------------------------------------
 */

static void
ComputeSeshItemBbox(canvas, textPtr)
     Tk_Canvas canvas;		/* Canvas that contains item. */
     SeshItemItem *textPtr;		/* Item whose bbox is to be recomputed. */
{
  Tk_CanvasSeshItemInfo *textInfoPtr;
  int leftX, topY, width, height, fudge;
  Tk_State state = textPtr->header.state;
  
  if(state == TK_STATE_NULL) {
    state = ((TkCanvas *)canvas)->canvas_state;
  }
  
  Tk_FreeSeshItemLayout(textPtr->textLayout);
  textPtr->textLayout = Tk_ComputeSeshItemLayout(textPtr->tkfont,
																								 textPtr->text, textPtr->numChars, textPtr->width,
																								 textPtr->justify, 0, &width, &height);
  
  if (state == TK_STATE_HIDDEN || textPtr->color == NULL) {
    width = height = 0;
  }
  
  /*
   * Use overall geometry information to compute the top-left corner
   * of the bounding box for the text item.
   */
  
  leftX = (int) (textPtr->x + 0.5);
  topY = (int) (textPtr->y + 0.5);
  switch (textPtr->anchor) {
  case TK_ANCHOR_NW:
  case TK_ANCHOR_N:
  case TK_ANCHOR_NE:
    break;
    
  case TK_ANCHOR_W:
  case TK_ANCHOR_CENTER:
  case TK_ANCHOR_E:
    topY -= height / 2;
    break;
    
  case TK_ANCHOR_SW:
  case TK_ANCHOR_S:
  case TK_ANCHOR_SE:
    topY -= height;
    break;
  }
  switch (textPtr->anchor) {
  case TK_ANCHOR_NW:
  case TK_ANCHOR_W:
  case TK_ANCHOR_SW:
    break;
    
  case TK_ANCHOR_N:
  case TK_ANCHOR_CENTER:
  case TK_ANCHOR_S:
    leftX -= width / 2;
    break;
    
  case TK_ANCHOR_NE:
  case TK_ANCHOR_E:
  case TK_ANCHOR_SE:
    leftX -= width;
    break;
  }
  
  textPtr->leftEdge  = leftX;
  textPtr->rightEdge = leftX + width;
  
  /*
   * Last of all, update the bounding box for the item.  The item's
   * bounding box includes the bounding box of all its lines, plus
   * an extra fudge factor for the cursor border (which could
   * potentially be quite large).
   */
  
  textInfoPtr = textPtr->textInfoPtr;
  fudge = (textInfoPtr->insertWidth + 1) / 2;
  if (textInfoPtr->selBorderWidth > fudge) {
    fudge = textInfoPtr->selBorderWidth;
  }
  textPtr->header.x1 = leftX - fudge;
  textPtr->header.y1 = topY;
  textPtr->header.x2 = leftX + width + fudge;
  textPtr->header.y2 = topY + height;
}

/*
 *--------------------------------------------------------------
 *
 * DisplayCanvSeshItem --
 *
 *	This procedure is invoked to draw a sesh item in a given
 *	drawable.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	ItemPtr is drawn in drawable using the transformation
 *	information in canvas.
 *
 *--------------------------------------------------------------
 */

static void
DisplayCanvSeshItem(canvas, itemPtr, display, drawable, x, y, width, height)
     Tk_Canvas canvas;		/* Canvas that contains item. */
     Tk_Item *itemPtr;		/* Item to be displayed. */
     Display *display;		/* Display on which to draw item. */
     Drawable drawable;		/* Pixmap or window in which to draw item. */
     int x, y, width, height;	/* Describes region of canvas that must be
															 * redisplayed (not used). */
{
  SeshItemItem *textPtr;
  Tk_CanvasSeshItemInfo *textInfoPtr;
  int selFirstChar, selLastChar;
  short drawableX, drawableY;
  Pixmap stipple;
  Tk_State state = itemPtr->state;
  
  textPtr = (SeshItemItem *) itemPtr;
  textInfoPtr = textPtr->textInfoPtr;
  
  if(state == TK_STATE_NULL) {
    state = ((TkCanvas *)canvas)->canvas_state;
  }
  stipple = textPtr->stipple;
  if (((TkCanvas *)canvas)->currentItemPtr == itemPtr) {
    if (textPtr->activeStipple!=None) {
      stipple = textPtr->activeStipple;
    }
  } else if (state==TK_STATE_DISABLED) {
    if (textPtr->disabledStipple!=None) {
      stipple = textPtr->disabledStipple;
    }
  }
  
  if (textPtr->gc == None) {
    return;
  }
  
  /*
   * If we're stippling, then modify the stipple offset in the GC.  Be
   * sure to reset the offset when done, since the GC is supposed to be
   * read-only.
   */
  
  if (stipple != None) {
    Tk_CanvasSetOffset(canvas, textPtr->gc, &textPtr->tsoffset);
  }
  
  selFirstChar = -1;
  selLastChar = 0;		/* lint. */
  
  if (textInfoPtr->selItemPtr == itemPtr) {
    selFirstChar = textInfoPtr->selectFirst;
    selLastChar = textInfoPtr->selectLast;
    if (selLastChar > textPtr->numChars) {
      selLastChar = textPtr->numChars - 1;
    }
    if ((selFirstChar >= 0) && (selFirstChar <= selLastChar)) {
      int xFirst, yFirst, hFirst;
      int xLast, yLast;
      
      /*
       * Draw a special background under the selection.
       */
      
      Tk_CharBbox(textPtr->textLayout, selFirstChar, &xFirst, &yFirst,
									NULL, &hFirst);
      Tk_CharBbox(textPtr->textLayout, selLastChar, &xLast, &yLast,
									NULL, NULL);
      
      /*
       * If the selection spans the end of this line, then display
       * selection background all the way to the end of the line.
       * However, for the last line we only want to display up to the
       * last character, not the end of the line.
       */
      
      x = xFirst;
      height = hFirst;
      for (y = yFirst ; y <= yLast; y += height) {
				if (y == yLast) {
					width = xLast - x;
				} else {	    
					width = textPtr->rightEdge - textPtr->leftEdge - x;
				}
				Tk_CanvasDrawableCoords(canvas,
																(double) (textPtr->leftEdge + x
																					- textInfoPtr->selBorderWidth),
																(double) (textPtr->header.y1 + y),
																&drawableX, &drawableY);
				Tk_Fill3DRectangle(Tk_CanvasTkwin(canvas), drawable,
													 textInfoPtr->selBorder, drawableX, drawableY,
													 width + 2 * textInfoPtr->selBorderWidth,
													 height, textInfoPtr->selBorderWidth, TK_RELIEF_RAISED);
				x = 0;
      }
    }
  }
  
  /*
   * If the insertion point should be displayed, then draw a special
   * background for the cursor before drawing the text.  Note:  if
   * we're the cursor item but the cursor is turned off, then redraw
   * background over the area of the cursor.  This guarantees that
   * the selection won't make the cursor invisible on mono displays,
   * where both are drawn in the same color.
   */
  
  if ((textInfoPtr->focusItemPtr == itemPtr) && (textInfoPtr->gotFocus)) {
    if (Tk_CharBbox(textPtr->textLayout, textPtr->insertPos,
										&x, &y, NULL, &height)) {
      Tk_CanvasDrawableCoords(canvas,
															(double) (textPtr->leftEdge + x
																				- (textInfoPtr->insertWidth / 2)),
															(double) (textPtr->header.y1 + y),
															&drawableX, &drawableY);
      if (textInfoPtr->cursorOn) {
				Tk_Fill3DRectangle(Tk_CanvasTkwin(canvas), drawable,
													 textInfoPtr->insertBorder,
													 drawableX, drawableY,
													 textInfoPtr->insertWidth, height,
													 textInfoPtr->insertBorderWidth, TK_RELIEF_RAISED);
      } else if (textPtr->cursorOffGC != None) {
				/*
				 * Redraw the background over the area of the cursor,
				 * even though the cursor is turned off.  This
				 * guarantees that the selection won't make the cursor
				 * invisible on mono displays, where both may be drawn
				 * in the same color.
				 */
				
				XFillRectangle(display, drawable, textPtr->cursorOffGC,
											 drawableX, drawableY,
											 (unsigned) textInfoPtr->insertWidth,
											 (unsigned) height);
      }
    }
  }
  
  
  /*
   * Display the text in two pieces: draw the entire text item, then
   * draw the selected text on top of it.  The selected text then
   * will only need to be drawn if it has different attributes (such
   * as foreground color) than regular text.
   */
  
  Tk_CanvasDrawableCoords(canvas, (double) textPtr->leftEdge,
													(double) textPtr->header.y1, &drawableX, &drawableY);
  Tk_DrawSeshItemLayout(display, drawable, textPtr->gc, textPtr->textLayout,
												drawableX, drawableY, 0, -1);
  
  if ((selFirstChar >= 0) && (textPtr->selSeshItemGC != textPtr->gc)) {
    Tk_DrawSeshItemLayout(display, drawable, textPtr->selSeshItemGC,
													textPtr->textLayout, drawableX, drawableY, selFirstChar,
													selLastChar + 1);
  }
  
  if (stipple != None) {
    XSetTSOrigin(display, textPtr->gc, 0, 0);
  }
}

/*
 *--------------------------------------------------------------
 *
 * SeshItemInsert --
 *
 *	Insert characters into a text item at a given position.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The text in the given item is modified.  The cursor and
 *	selection positions are also modified to reflect the
 *	insertion.
 *
 *--------------------------------------------------------------
 */

static void
SeshItemInsert(canvas, itemPtr, index, string)
     Tk_Canvas canvas;		/* Canvas containing text item. */
     Tk_Item *itemPtr;		/* SeshItem item to be modified. */
     int index;			/* Character index before which string is
										 * to be inserted. */
     char *string;		/* New characters to be inserted. */
{
  SeshItemItem *textPtr = (SeshItemItem *) itemPtr;
  int byteIndex, byteCount, charsAdded;
  char *new, *text;
  Tk_CanvasSeshItemInfo *textInfoPtr = textPtr->textInfoPtr;
  
  string = Tcl_GetStringFromObj((Tcl_Obj *) string, &byteCount);
  
  text = textPtr->text;
  
  if (index < 0) {
    index = 0;
  }
  if (index > textPtr->numChars) {
    index = textPtr->numChars;
  }
  byteIndex = Tcl_UtfAtIndex(text, index) - text;
  byteCount = strlen(string);
  if (byteCount == 0) {
    return;
  }
  
  new = (char *) ckalloc((unsigned) textPtr->numBytes + byteCount + 1);
  memcpy(new, text, (size_t) byteIndex);
  strcpy(new + byteIndex, string);
  strcpy(new + byteIndex + byteCount, text + byteIndex);
  
  ckfree(text);
  textPtr->text = new;
  charsAdded = Tcl_NumUtfChars(string, byteCount);
  textPtr->numChars += charsAdded;
  textPtr->numBytes += byteCount;
  
  /*
   * Inserting characters invalidates indices such as those for the
   * selection and cursor.  Update the indices appropriately.
   */
  
  if (textInfoPtr->selItemPtr == itemPtr) {
    if (textInfoPtr->selectFirst >= index) {
      textInfoPtr->selectFirst += charsAdded;
    }
    if (textInfoPtr->selectLast >= index) {
      textInfoPtr->selectLast += charsAdded;
    }
    if ((textInfoPtr->anchorItemPtr == itemPtr)
				&& (textInfoPtr->selectAnchor >= index)) {
      textInfoPtr->selectAnchor += charsAdded;
    }
  }
  if (textPtr->insertPos >= index) {
    textPtr->insertPos += charsAdded;
  }
  ComputeSeshItemBbox(canvas, textPtr);
}

/*
 *--------------------------------------------------------------
 *
 * SeshItemDeleteChars --
 *
 *	Delete one or more characters from a text item.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Characters between "first" and "last", inclusive, get
 *	deleted from itemPtr, and things like the selection
 *	position get updated.
 *
 *--------------------------------------------------------------
 */

static void
SeshItemDeleteChars(canvas, itemPtr, first, last)
     Tk_Canvas canvas;		/* Canvas containing itemPtr. */
     Tk_Item *itemPtr;		/* Item in which to delete characters. */
     int first;			/* Character index of first character to
										 * delete. */
     int last;			/* Character index of last character to
										 * delete (inclusive). */
{
  SeshItemItem *textPtr = (SeshItemItem *) itemPtr;
  int byteIndex, byteCount, charsRemoved;
  char *new, *text;
  Tk_CanvasSeshItemInfo *textInfoPtr = textPtr->textInfoPtr;
  
  text = textPtr->text;
  if (first < 0) {
    first = 0;
  }
  if (last >= textPtr->numChars) {
    last = textPtr->numChars - 1;
  }
  if (first > last) {
    return;
  }
  charsRemoved = last + 1 - first;
  
  byteIndex = Tcl_UtfAtIndex(text, first) - text;
  byteCount = Tcl_UtfAtIndex(text + byteIndex, charsRemoved)
    - (text + byteIndex);
  
  new = (char *) ckalloc((unsigned) (textPtr->numBytes + 1 - byteCount));
  memcpy(new, text, (size_t) byteIndex);
  strcpy(new + byteIndex, text + byteIndex + byteCount);
  
  ckfree(text);
  textPtr->text = new;
  textPtr->numChars -= charsRemoved;
  textPtr->numBytes -= byteCount;
  
  /*
   * Update indexes for the selection and cursor to reflect the
   * renumbering of the remaining characters.
   */
  
  if (textInfoPtr->selItemPtr == itemPtr) {
    if (textInfoPtr->selectFirst > first) {
      textInfoPtr->selectFirst -= charsRemoved;
      if (textInfoPtr->selectFirst < first) {
				textInfoPtr->selectFirst = first;
      }
    }
    if (textInfoPtr->selectLast >= first) {
      textInfoPtr->selectLast -= charsRemoved;
      if (textInfoPtr->selectLast < first - 1) {
				textInfoPtr->selectLast = first - 1;
      }
    }
    if (textInfoPtr->selectFirst > textInfoPtr->selectLast) {
      textInfoPtr->selItemPtr = NULL;
    }
    if ((textInfoPtr->anchorItemPtr == itemPtr)
				&& (textInfoPtr->selectAnchor > first)) {
      textInfoPtr->selectAnchor -= charsRemoved;
      if (textInfoPtr->selectAnchor < first) {
				textInfoPtr->selectAnchor = first;
      }
    }
  }
  if (textPtr->insertPos > first) {
    textPtr->insertPos -= charsRemoved;
    if (textPtr->insertPos < first) {
      textPtr->insertPos = first;
    }
  }
  ComputeSeshItemBbox(canvas, textPtr);
  return;
}

/*
 *--------------------------------------------------------------
 *
 * SeshItemToPoint --
 *
 *	Computes the distance from a given point to a given
 *	text item, in canvas units.
 *
 * Results:
 *	The return value is 0 if the point whose x and y coordinates
 *	are pointPtr[0] and pointPtr[1] is inside the text item.  If
 *	the point isn't inside the text item then the return value
 *	is the distance from the point to the text item.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

static double
SeshItemToPoint(canvas, itemPtr, pointPtr)
     Tk_Canvas canvas;		/* Canvas containing itemPtr. */
     Tk_Item *itemPtr;		/* Item to check against point. */
     double *pointPtr;		/* Pointer to x and y coordinates. */
{
  SeshItemItem *textPtr;
  Tk_State state = itemPtr->state;
  double value;
  
  if (state == TK_STATE_NULL) {
    state = ((TkCanvas *)canvas)->canvas_state;
  }
  textPtr = (SeshItemItem *) itemPtr;
  value =  (double) Tk_DistanceToSeshItemLayout(textPtr->textLayout,
																								(int) pointPtr[0] - textPtr->leftEdge,
																								(int) pointPtr[1] - textPtr->header.y1);
  
  if ((state == TK_STATE_HIDDEN) || (textPtr->color == NULL) ||
      (textPtr->text == NULL) || (*textPtr->text == 0)) {
    value = 1.0e36;
  }
  return value;
}

/*
 *--------------------------------------------------------------
 *
 * SeshItemToArea --
 *
 *	This procedure is called to determine whether an item
 *	lies entirely inside, entirely outside, or overlapping
 *	a given rectangle.
 *
 * Results:
 *	-1 is returned if the item is entirely outside the area
 *	given by rectPtr, 0 if it overlaps, and 1 if it is entirely
 *	inside the given area.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

static int
SeshItemToArea(canvas, itemPtr, rectPtr)
     Tk_Canvas canvas;		/* Canvas containing itemPtr. */
     Tk_Item *itemPtr;		/* Item to check against rectangle. */
     double *rectPtr;		/* Pointer to array of four coordinates
												 * (x1, y1, x2, y2) describing rectangular
												 * area.  */
{
  SeshItemItem *textPtr;
  Tk_State state = itemPtr->state;
  
  if (state == TK_STATE_NULL) {
    state = ((TkCanvas *)canvas)->canvas_state;
  }
  
  textPtr = (SeshItemItem *) itemPtr;
  return Tk_IntersectSeshItemLayout(textPtr->textLayout,
																		(int) (rectPtr[0] + 0.5) - textPtr->leftEdge,
																		(int) (rectPtr[1] + 0.5) - textPtr->header.y1,
																		(int) (rectPtr[2] - rectPtr[0] + 0.5),
																		(int) (rectPtr[3] - rectPtr[1] + 0.5));
}

/*
 *--------------------------------------------------------------
 *
 * ScaleSeshItem --
 *
 *	This procedure is invoked to rescale a text item.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Scales the position of the text, but not the size
 *	of the font for the text.
 *
 *--------------------------------------------------------------
 */

/* ARGSUSED */
static void
ScaleSeshItem(canvas, itemPtr, originX, originY, scaleX, scaleY)
     Tk_Canvas canvas;		/* Canvas containing rectangle. */
     Tk_Item *itemPtr;		/* Rectangle to be scaled. */
     double originX, originY;	/* Origin about which to scale rect. */
     double scaleX;		/* Amount to scale in X direction. */
     double scaleY;		/* Amount to scale in Y direction. */
{
  SeshItemItem *textPtr = (SeshItemItem *) itemPtr;
  
  textPtr->x = originX + scaleX*(textPtr->x - originX);
  textPtr->y = originY + scaleY*(textPtr->y - originY);
  ComputeSeshItemBbox(canvas, textPtr);
  return;
}

/*
 *--------------------------------------------------------------
 *
 * TranslateSeshItem --
 *
 *	This procedure is called to move a text item by a
 *	given amount.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The position of the text item is offset by (xDelta, yDelta),
 *	and the bounding box is updated in the generic part of the
 *	item structure.
 *
 *--------------------------------------------------------------
 */

static void
TranslateSeshItem(canvas, itemPtr, deltaX, deltaY)
     Tk_Canvas canvas;		/* Canvas containing item. */
     Tk_Item *itemPtr;		/* Item that is being moved. */
     double deltaX, deltaY;	/* Amount by which item is to be moved. */
{
  SeshItemItem *textPtr = (SeshItemItem *) itemPtr;
  
  textPtr->x += deltaX;
  textPtr->y += deltaY;
  ComputeSeshItemBbox(canvas, textPtr);
}

/*
 *--------------------------------------------------------------
 *
 * GetSeshItemIndex --
 *
 *	Parse an index into a text item and return either its value
 *	or an error.
 *
 * Results:
 *	A standard Tcl result.  If all went well, then *indexPtr is
 *	filled in with the index (into itemPtr) corresponding to
 *	string.  Otherwise an error message is left in
 *	the interp's result.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

static int
GetSeshItemIndex(interp, canvas, itemPtr, obj, indexPtr)
     Tcl_Interp *interp;		/* Used for error reporting. */
     Tk_Canvas canvas;		/* Canvas containing item. */
     Tk_Item *itemPtr;		/* Item for which the index is being
													 * specified. */
     Tcl_Obj *obj;		/* Specification of a particular character
											 * in itemPtr's text. */
     int *indexPtr;		/* Where to store converted character
											 * index. */
{
  SeshItemItem *textPtr = (SeshItemItem *) itemPtr;
  size_t length;
  int c;
  TkCanvas *canvasPtr = (TkCanvas *) canvas;
  Tk_CanvasSeshItemInfo *textInfoPtr = textPtr->textInfoPtr;
  char *string = Tcl_GetStringFromObj(obj, (int *) &length);
  
  c = string[0];
  length = strlen(string);
  
  if ((c == 'e') && (strncmp(string, "end", length) == 0)) {
    *indexPtr = textPtr->numChars;
  } else if ((c == 'i') && (strncmp(string, "insert", length) == 0)) {
    *indexPtr = textPtr->insertPos;
  } else if ((c == 's') && (strncmp(string, "sel.first", length) == 0)
						 && (length >= 5)) {
    if (textInfoPtr->selItemPtr != itemPtr) {
      Tcl_SetResult(interp, "selection isn't in item", TCL_STATIC);
      return TCL_ERROR;
    }
    *indexPtr = textInfoPtr->selectFirst;
  } else if ((c == 's') && (strncmp(string, "sel.last", length) == 0)
						 && (length >= 5)) {
    if (textInfoPtr->selItemPtr != itemPtr) {
      Tcl_SetResult(interp, "selection isn't in item", TCL_STATIC);
      return TCL_ERROR;
    }
    *indexPtr = textInfoPtr->selectLast;
  } else if (c == '@') {
    int x, y;
    double tmp;
    char *end, *p;
    
    p = string+1;
    tmp = strtod(p, &end);
    if ((end == p) || (*end != ',')) {
      goto badIndex;
    }
    x = (int) ((tmp < 0) ? tmp - 0.5 : tmp + 0.5);
    p = end+1;
    tmp = strtod(p, &end);
    if ((end == p) || (*end != 0)) {
      goto badIndex;
    }
    y = (int) ((tmp < 0) ? tmp - 0.5 : tmp + 0.5);
    *indexPtr = Tk_PointToChar(textPtr->textLayout,
															 x + canvasPtr->scrollX1 - textPtr->leftEdge,
															 y + canvasPtr->scrollY1 - textPtr->header.y1);
  } else if (Tcl_GetIntFromObj((Tcl_Interp *)NULL, obj, indexPtr) == TCL_OK) {
    if (*indexPtr < 0){
      *indexPtr = 0;
    } else if (*indexPtr > textPtr->numChars) {
      *indexPtr = textPtr->numChars;
    }
  } else {
    /*
     * Some of the paths here leave messages in the interp's result,
     * so we have to clear it out before storing our own message.
     */
    
  badIndex:
    Tcl_SetResult(interp, (char *) NULL, TCL_STATIC);
    Tcl_AppendResult(interp, "bad index \"", string, "\"",
										 (char *) NULL);
    return TCL_ERROR;
  }
  return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * SetSeshItemCursor --
 *
 *	Set the position of the insertion cursor in this item.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The cursor position will change.
 *
 *--------------------------------------------------------------
 */

/* ARGSUSED */
static void
SetSeshItemCursor(canvas, itemPtr, index)
     Tk_Canvas canvas;		/* Record describing canvas widget. */
     Tk_Item *itemPtr;		/* SeshItem item in which cursor position is to
													 * be set. */
     int index;			/* Character index of character just before
										 * which cursor is to be positioned. */
{
  SeshItemItem *textPtr = (SeshItemItem *) itemPtr;
  
  if (index < 0) {
    textPtr->insertPos = 0;
  } else  if (index > textPtr->numChars) {
    textPtr->insertPos = textPtr->numChars;
  } else {
    textPtr->insertPos = index;
  }
}

/*
 *--------------------------------------------------------------
 *
 * GetSelSeshItem --
 *
 *	This procedure is invoked to return the selected portion
 *	of a text item.  It is only called when this item has
 *	the selection.
 *
 * Results:
 *	The return value is the number of non-NULL bytes stored
 *	at buffer.  Buffer is filled (or partially filled) with a
 *	NULL-terminated string containing part or all of the selection,
 *	as given by offset and maxBytes.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

static int
GetSelSeshItem(canvas, itemPtr, offset, buffer, maxBytes)
     Tk_Canvas canvas;		/* Canvas containing selection. */
     Tk_Item *itemPtr;		/* SeshItem item containing selection. */
     int offset;			/* Byte offset within selection of first
											 * character to be returned. */
     char *buffer;		/* Location in which to place selection. */
     int maxBytes;		/* Maximum number of bytes to place at
											 * buffer, not including terminating NULL
											 * character. */
{
  SeshItemItem *textPtr = (SeshItemItem *) itemPtr;
  int byteCount; 
  char *text, *selStart, *selEnd;
  Tk_CanvasSeshItemInfo *textInfoPtr = textPtr->textInfoPtr;
  
  if ((textInfoPtr->selectFirst < 0) ||
      (textInfoPtr->selectFirst > textInfoPtr->selectLast)) {
    return 0;
  }
  text = textPtr->text;
  selStart = Tcl_UtfAtIndex(text, textInfoPtr->selectFirst);
  selEnd = Tcl_UtfAtIndex(selStart,
													textInfoPtr->selectLast + 1 - textInfoPtr->selectFirst);
  byteCount = selEnd - selStart - offset;
  if (byteCount > maxBytes) {
    byteCount = maxBytes;
  }
  if (byteCount <= 0) {
    return 0;
  }
  memcpy(buffer, selStart + offset, (size_t) byteCount);
  buffer[byteCount] = '\0';
  return byteCount;
}

/*
 *--------------------------------------------------------------
 *
 * SeshItemToPostscript --
 *
 *	This procedure is called to generate Postscript for
 *	text items.
 *
 * Results:
 *	The return value is a standard Tcl result.  If an error
 *	occurs in generating Postscript then an error message is
 *	left in the interp's result, replacing whatever used
 *	to be there.  If no error occurs, then Postscript for the
 *	item is appended to the result.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

static int
SeshItemToPostscript(interp, canvas, itemPtr, prepass)
     Tcl_Interp *interp;		/* Leave Postscript or error message here. */
     Tk_Canvas canvas;		/* Information about overall canvas. */
     Tk_Item *itemPtr;		/* Item for which Postscript is wanted. */
     int prepass;		/* 1 means this is a prepass to collect
										 * font information; 0 means final Postscript
										 * is being created. */
{
  SeshItemItem *textPtr = (SeshItemItem *) itemPtr;
  int x, y;
  Tk_FontMetrics fm;
  char *justify;
  char buffer[500];
  XColor *color;
  Pixmap stipple;
  Tk_State state = itemPtr->state;
  
  if(state == TK_STATE_NULL) {
    state = ((TkCanvas *)canvas)->canvas_state;
  }
  color = textPtr->color;
  stipple = textPtr->stipple;
  if (state == TK_STATE_HIDDEN || textPtr->color == NULL ||
      textPtr->text == NULL || *textPtr->text == 0) {
    return TCL_OK;
  } else if (((TkCanvas *)canvas)->currentItemPtr == itemPtr) {
    if (textPtr->activeColor!=NULL) {
      color = textPtr->activeColor;
    }
    if (textPtr->activeStipple!=None) {
      stipple = textPtr->activeStipple;
    }
  } else if (state==TK_STATE_DISABLED) {
    if (textPtr->disabledColor!=NULL) {
      color = textPtr->disabledColor;
    }
    if (textPtr->disabledStipple!=None) {
      stipple = textPtr->disabledStipple;
    }
  }
  
  if (Tk_CanvasPsFont(interp, canvas, textPtr->tkfont) != TCL_OK) {
    return TCL_ERROR;
  }
  if (prepass != 0) {
    return TCL_OK;
  }
  if (Tk_CanvasPsColor(interp, canvas, color) != TCL_OK) {
    return TCL_ERROR;
  }
  if (stipple != None) {
    Tcl_AppendResult(interp, "/StippleSeshItem {\n    ",
										 (char *) NULL);
    Tk_CanvasPsStipple(interp, canvas, stipple);
    Tcl_AppendResult(interp, "} bind def\n", (char *) NULL);
  }
  
  sprintf(buffer, "%.15g %.15g [\n", textPtr->x,
					Tk_CanvasPsY(canvas, textPtr->y));
  Tcl_AppendResult(interp, buffer, (char *) NULL);
  
  Tk_SeshItemLayoutToPostscript(interp, textPtr->textLayout);
  
  x = 0;  y = 0;  justify = NULL;	/* lint. */
  switch (textPtr->anchor) {
  case TK_ANCHOR_NW:	x = 0; y = 0;	break;
  case TK_ANCHOR_N:	x = 1; y = 0;	break;
  case TK_ANCHOR_NE:	x = 2; y = 0;	break;
  case TK_ANCHOR_E:	x = 2; y = 1;	break;
  case TK_ANCHOR_SE:	x = 2; y = 2;	break;
  case TK_ANCHOR_S:	x = 1; y = 2;	break;
  case TK_ANCHOR_SW:	x = 0; y = 2;	break;
  case TK_ANCHOR_W:	x = 0; y = 1;	break;
  case TK_ANCHOR_CENTER:	x = 1; y = 1;	break;
  }
  switch (textPtr->justify) {
  case TK_JUSTIFY_LEFT:	justify = "0";	break;
  case TK_JUSTIFY_CENTER: justify = "0.5";break;
  case TK_JUSTIFY_RIGHT:  justify = "1";	break;
  }
  
  Tk_GetFontMetrics(textPtr->tkfont, &fm);
  sprintf(buffer, "] %d %g %g %s %s DrawSeshItem\n",
					fm.linespace, x / -2.0, y / 2.0, justify,
					((stipple == None) ? "false" : "true"));
  Tcl_AppendResult(interp, buffer, (char *) NULL);
  
  return TCL_OK;
}

/* Let's preserve the present presentation
 * Local Variables:
 * tab-width: 2
 * c-block-comments-indent-p: nil
 * End:
 */
