/* WMF creator for test purposes */

#include <stdio.h>
#include <malloc.h>
#include "wmf.h"


/* Metafile functions */
#define META_SETBKCOLOR              0x0201
#define META_SETBKMODE               0x0102
#define META_SETMAPMODE              0x0103
#define META_SETROP2                 0x0104
#define META_SETRELABS               0x0105
#define META_SETPOLYFILLMODE         0x0106
#define META_SETSTRETCHBLTMODE       0x0107
#define META_SETTEXTCHAREXTRA        0x0108
#define META_SETTEXTCOLOR            0x0209
#define META_SETTEXTJUSTIFICATION    0x020A
#define META_SETWINDOWORG            0x020B
#define META_SETWINDOWEXT            0x020C
#define META_SETVIEWPORTORG          0x020D
#define META_SETVIEWPORTEXT          0x020E
#define META_OFFSETWINDOWORG         0x020F
#define META_SCALEWINDOWEXT          0x0410
#define META_OFFSETVIEWPORTORG       0x0211
#define META_SCALEVIEWPORTEXT        0x0412
#define META_LINETO                  0x0213
#define META_MOVETO                  0x0214
#define META_EXCLUDECLIPRECT         0x0415
#define META_INTERSECTCLIPRECT       0x0416
#define META_ARC                     0x0817
#define META_ELLIPSE                 0x0418
#define META_FLOODFILL               0x0419
#define META_PIE                     0x081A
#define META_RECTANGLE               0x041B
#define META_ROUNDRECT               0x061C
#define META_PATBLT                  0x061D
#define META_SAVEDC                  0x001E
#define META_SETPIXEL                0x041F
#define META_OFFSETCLIPRGN           0x0220
#define META_TEXTOUT                 0x0521
#define META_BITBLT                  0x0922
#define META_STRETCHBLT              0x0B23
#define META_POLYGON                 0x0324
#define META_POLYLINE                0x0325
#define META_ESCAPE                  0x0626
#define META_RESTOREDC               0x0127
#define META_FILLREGION              0x0228
#define META_FRAMEREGION             0x0429
#define META_INVERTREGION            0x012A
#define META_PAINTREGION             0x012B
#define META_SELECTCLIPREGION        0x012C
#define META_SELECTOBJECT            0x012D
#define META_SETTEXTALIGN            0x012E
#define META_DRAWTEXT                0x062F
#define META_CHORD                   0x0830
#define META_SETMAPPERFLAGS          0x0231
#define META_EXTTEXTOUT              0x0A32
#define META_SETDIBTODEV             0x0D33
#define META_SELECTPALETTE           0x0234
#define META_REALIZEPALETTE          0x0035
#define META_ANIMATEPALETTE          0x0436
#define META_SETPALENTRIES           0x0037
#define META_POLYPOLYGON             0x0538
#define META_RESIZEPALETTE           0x0139
#define META_DIBBITBLT               0x0940
#define META_DIBSTRETCHBLT           0x0B41
#define META_DIBCREATEPATTERNBRUSH   0x0142
#define META_STRETCHDIB              0x0F43
#define META_EXTFLOODFILL            0x0548
#define META_RESETDC                 0x014C
#define META_STARTDOC                0x014D
#define META_STARTPAGE               0x004F
#define META_ENDPAGE                 0x0050
#define META_ABORTDOC                0x0052
#define META_ENDDOC                  0x005E
#define META_DELETEOBJECT            0x01F0
#define META_CREATEPALETTE           0x00F7
#define META_CREATEBRUSH             0x00F8
#define META_CREATEPATTERNBRUSH      0x01F9
#define META_CREATEPENINDIRECT       0x02FA
#define META_CREATEFONTINDIRECT      0x02FB
#define META_CREATEBRUSHINDIRECT     0x02FC
#define META_CREATEBITMAPINDIRECT    0x02FD
#define META_CREATEBITMAP            0x06FE
#define META_CREATEREGION            0x06FF
#define META_UNKNOWN                 0x0529  /* FIXME: unknown meta magic */

/*
  Various other definitions from windows :
 */

typedef struct _WindowsMetaHeader
{
  WMFWORD  FileType;       /* Type of metafile (0=memory, 1=disk) */
  WMFWORD  HeaderSize;     /* Size of header in WMFWORDS (always 9) */
  WMFWORD  Version;        /* Version of Microsoft Windows used */
  WMFDWORD FileSize;       /* Total size of the metafile in WMFWORDs */
  WMFWORD  NumOfObjects;   /* Number of objects in the file */
  WMFDWORD MaxRecordSize;  /* The size of largest record in WMFWORDs */
  WMFWORD  NumOfParams;    /* Not Used (always 0) */
} WMFHEAD;

typedef struct _StandardMetaRecord
{
  WMFDWORD Size;          /* Total size of the record in WMFWORDs */
  WMFWORD  Function;      /* Function number (defined in WINDOWS.H) */
  WMFWORD  Parameters[1];  /* Parameter values passed to function */
} WMFRECORD;

typedef struct _PlaceableMetaHeader
{
  WMFDWORD Key;           /* Magic number (always 9AC6CDD7h) */
  WMFWORD  Handle;        /* Metafile HANDLE number (always 0) */
  WMFSHORT Left;          /* Left coordinate in metafile units */
  WMFSHORT Top;           /* Top coordinate in metafile units (IN TWIPS !!!) */
  WMFSHORT Right;         /* Right coordinate in metafile units */
  WMFSHORT Bottom;        /* Bottom coordinate in metafile units */
  WMFWORD  Inch;          /* Number of metafile units per inch (normally 1440) */ 
  WMFDWORD Reserved;      /* Reserved (always 0) */
  WMFWORD  Checksum;      /* Checksum value for previous 10 WMFWORDs */
} PLACEABLEMETAHEADER;

#define MAXOBJS 1024
struct WMF {
  PLACEABLEMETAHEADER meta;
  WMFHEAD head;
  FILE * f;
  int objects[MAXOBJS]; /* Are there limits for this ?? */
};

static int
AllocObject(struct WMF *w)
{
  int i;
  int res= -1;
  for (i=0; res == -1 && i< MAXOBJS; i++)
    {
      if (w->objects[i] == -1)
	res= i;
    }
  return res;
}

static void
FreeObject(struct WMF *w, int nobj)
{
  w->objects[nobj]= -1;
}

static void
compute_checksum(PLACEABLEMETAHEADER *pmh)
{
  WMFWORD *ptr;
  pmh->Checksum = 0;
  for (ptr = (WMFWORD *) pmh; ptr < (WMFWORD *)&pmh->Checksum; ptr++)
    pmh->Checksum ^= *ptr;
}

static void
writeWMFCHAR(FILE *f, char x)
{
  fwrite((char*)&x, 1, 1, f);
}

static void
writeWMFSHORT(FILE *f, WMFSHORT x)
{
  /* fwrite((char*)&x, 2, 1, f); */
  /* Portable system : */
  unsigned char s[2];
  s[0]= (unsigned short int)x & 0xFF;
  s[1]= ((unsigned short int)x & 0xFF00) >> 8 ;
  fwrite(s,2,1,f);
}

static void
writeWMFWORD(FILE *f, WMFWORD x)
{
  /* fwrite((char*)&x, 2, 1, f); */
  /* Portable system : */
  unsigned char s[2];
  s[0]= (unsigned short int)x & 0xFF;
  s[1]= ((unsigned short int)x & 0xFF00) >> 8 ;
  fwrite(s,2,1,f);  
}

static void
writeWMFDWORD(FILE *f, WMFDWORD x)
{
  /* char *s= (unsigned char *) &x;
     fwrite(s, 4, 1, f); */
  /* Portable system : */
  unsigned char s[4];
  s[0]= (unsigned int)x & 0xFF;
  s[1]= ((unsigned int)x & 0xFF00) >> 8 ;
  s[2]= ((unsigned int)x & 0xFF0000) >> 16 ;
  s[3]= ((unsigned int)x & 0xFF000000) >> 24 ;
  fwrite(s,4,1,f);  
}

static void
writeHEADER(struct WMF* w)
{
  compute_checksum(&(w->meta));
  
  writeWMFDWORD(w->f, w->meta.Key);
  writeWMFWORD(w->f, w->meta.Handle);
  writeWMFSHORT(w->f, w->meta.Left);
  writeWMFSHORT(w->f, w->meta.Top);
  writeWMFSHORT(w->f, w->meta.Right);
  writeWMFSHORT(w->f, w->meta.Bottom);
  writeWMFWORD(w->f, w->meta.Inch);
  writeWMFDWORD(w->f, w->meta.Reserved);
  writeWMFWORD(w->f, w->meta.Checksum);
  
  writeWMFWORD(w->f, w->head.FileType);
  writeWMFWORD(w->f, w->head.HeaderSize);
  writeWMFWORD(w->f, w->head.Version);
  writeWMFDWORD(w->f, w->head.FileSize);
  writeWMFWORD(w->f, w->head.NumOfObjects);
  writeWMFDWORD(w->f, w->head.MaxRecordSize);
  writeWMFWORD(w->f, w->head.NumOfParams);
}

static void
addRecord(struct WMF *w, WMFWORD function, int nbrWMFWORDs)
{  
  int size= 3+ nbrWMFWORDs;
  /* Update header values */
  w->head.NumOfObjects+= 1;
  if (w->head.MaxRecordSize < size)
    w->head.MaxRecordSize = size;
  w->head.FileSize+= size;
  /* Start writing data */
  writeWMFDWORD(w->f, size);
  writeWMFWORD(w->f, function);
}

static void
AddCommand(struct WMF * w, int func)
{
  addRecord(w, func, 0);
}

static void
AddCommandX(struct WMF * w, int func, int x)
{
  addRecord(w, func, 1);
  writeWMFWORD(w->f, x);
}

static void
AddCommandXY(struct WMF * w, int func, int x, int y)
{
  addRecord(w, func, 2);
  writeWMFWORD(w->f, y);
  writeWMFWORD(w->f, x);
}

static void
AddCommandABCD(struct WMF * w, int func, int a, int b, int c, int d)
{
  addRecord(w, func, 4);
  writeWMFWORD(w->f, d);
  writeWMFWORD(w->f, c);
  writeWMFWORD(w->f, b);
  writeWMFWORD(w->f, a);
}

static void
AddCommandL(struct WMF * w, int func, WMFDWORD x)
{
  addRecord(w, func, 2);
  writeWMFDWORD(w->f, x);
}

static void
AddCommandLXY(struct WMF * w, int func, WMFDWORD a, WMFWORD b, WMFWORD c)
{
}

static void
AddCommandXYL(struct WMF * w, int func, WMFWORD a, WMFWORD b, WMFDWORD c)
{
  addRecord(w, func, 4);
  writeWMFDWORD(w->f, c);
  writeWMFWORD(w->f, b);
  writeWMFWORD(w->f,a);
}

static void
AddCommandLINES(struct WMF *w, int func, WMFpoint *p, WMFWORD nbr)
{ 
  int i;
  addRecord(w, func, 2*nbr + 1);
  writeWMFWORD(w->f, nbr);
  for(i=0; i<nbr; i++)
    {
      writeWMFWORD(w->f, p[i].x);
      writeWMFWORD(w->f, p[i].y);
    }
}

/*
 * Now, the actual functions
 */


void
WMFPolygon(struct WMF *w, WMFpoint *p, WMFWORD nbr)
{
  AddCommandLINES(w, META_POLYGON, p, nbr);
}

void
WMFPolyLine(struct WMF *w, WMFpoint *p, WMFWORD nbr)
{
  AddCommandLINES(w, META_POLYLINE, p, nbr);
}

/*
 * Arc from an ellipse described by top,left,...
 * The Arc itself is described by xstart & co
 *
 */


void
WMFArc(struct WMF *w, WMFWORD left, WMFWORD top, WMFWORD right, WMFWORD bottom, 
       WMFWORD xstart, WMFWORD ystart, WMFWORD xend, WMFWORD yend)
{
  addRecord(w, META_ARC, 8);
  writeWMFWORD(w->f, yend);
  writeWMFWORD(w->f, xend);
  writeWMFWORD(w->f, ystart);
  writeWMFWORD(w->f, xstart);
  writeWMFWORD(w->f, bottom);
  writeWMFWORD(w->f, right);
  writeWMFWORD(w->f, top);
  writeWMFWORD(w->f, left);
}

void
WMFMoveTo(struct WMF * w, WMFWORD x, WMFWORD y)
{
  AddCommandXY(w, 0x0214, x, y);
}

void
WMFLineTo(struct WMF *w, int x, int y)
{
  AddCommandXY(w, 0x0213, x, y);
}

void
WMFEllipse(struct WMF *w, WMFWORD left, WMFWORD top, WMFWORD right, WMFWORD bottom)
{
  AddCommandABCD(w, META_ELLIPSE, left, top, right, bottom);
}

void
WMFRectangle(struct WMF *w, WMFWORD left, WMFWORD top, WMFWORD right, WMFWORD bottom)
{
  AddCommandABCD(w, META_RECTANGLE, left, top, right, bottom);
}

void
WMFRoundRectangle(struct WMF *w, WMFWORD left, WMFWORD top, WMFWORD right, WMFWORD bottom, 
		  WMFWORD wi, WMFWORD h)
{
  addRecord(w, META_ROUNDRECT, 6);
  writeWMFWORD(w->f, h);
  writeWMFWORD(w->f, wi);
  writeWMFWORD(w->f, bottom);
  writeWMFWORD(w->f, right);
  writeWMFWORD(w->f, top);
  writeWMFWORD(w->f, left);
}

void
WMFPatBlt(struct WMF *w, WMFWORD left, WMFWORD top, WMFWORD right, WMFWORD bottom, WMFDWORD rop)
{
  addRecord(w, META_PATBLT, 6);
  writeWMFDWORD(w->f, rop);
  writeWMFWORD(w->f, bottom);
  writeWMFWORD(w->f, right);
  writeWMFWORD(w->f, top);
  writeWMFWORD(w->f, left);
}


void
WMFTextOut(struct WMF *w, WMFWORD x, WMFWORD y, char * s)
{
  int i;
  int length= strlen(s);
  addRecord(w, META_TEXTOUT, (length+1)/2 + 3);
  /* count of chars in the string */
  writeWMFWORD(w->f, length);
  /* the string (16bit aligned) */
  for(i=0; i < length; i++)
    writeWMFCHAR(w->f, s[i]);
  /* ensures a even number of chars */
  if (length % 2 != 0) 
    writeWMFCHAR(w->f, 0);
  /* y */
  writeWMFWORD(w->f, y);
  /* x */
  writeWMFWORD(w->f, x);
}

void
WMFFloodFill(struct WMF * w, WMFWORD x, WMFWORD y, WMFDWORD colour)
{
  AddCommandXYL(w, META_FLOODFILL, x, y, colour);
}

void
WMFSetPixel(struct WMF * w, WMFWORD x, WMFWORD y, WMFDWORD colour)
{
  AddCommandXYL(w, META_SETPIXEL, x, y, colour);
}

void
WMFFillRgn(struct WMF * w, WMFWORD nbrush, WMFWORD nregion)
{
 AddCommandXY(w, META_FILLREGION, nbrush, nregion); 
}

void
WMFInvertRgn(struct WMF * w, WMFWORD nregion)
{
  AddCommandX(w, META_INVERTREGION, nregion);   
}

void
WMFPaintRgn(struct WMF * w, WMFWORD nregion)
{
  AddCommandX(w, META_PAINTREGION, nregion);   
}

void
WMFSelectClipRgn(struct WMF * w, WMFWORD nregion)
{
  AddCommandX(w, META_SELECTCLIPREGION, nregion);   
}

void
WMFSetWindowOrg(struct WMF *w, WMFWORD x, WMFWORD y)
{
  AddCommandXY(w, META_SETWINDOWORG, x, y);
}

void
WMFSetViewportOrg(struct WMF *w, WMFWORD x, WMFWORD y)
{
  AddCommandXY(w, META_SETVIEWPORTORG, x, y);
}

void
WMFOffsetWindowOrg(struct WMF *w, WMFWORD x, WMFWORD y)
{
  AddCommandXY(w, META_OFFSETWINDOWORG, x, y);
}

void
WMFOffsetViewportOrg(struct WMF *w, WMFWORD x, WMFWORD y)
{
  AddCommandXY(w, META_OFFSETVIEWPORTORG, x, y);
}

void
WMFScaleViewportExt(struct WMF *w,WMFWORD xnum,WMFWORD xdenom,WMFWORD ynum,WMFWORD ydenom)
{
  addRecord(w, META_SCALEVIEWPORTEXT, 4);
  writeWMFWORD(w->f, ydenom);
  writeWMFWORD(w->f, ynum);
  writeWMFWORD(w->f, xdenom);
  writeWMFWORD(w->f, xnum);
}


void
WMFScaleWindowExt(struct WMF *w,WMFWORD xnum,WMFWORD xdenom,WMFWORD ynum,WMFWORD ydenom)
{
  addRecord(w, META_SCALEWINDOWEXT, 4);
  writeWMFWORD(w->f, ydenom);
  writeWMFWORD(w->f, ynum);
  writeWMFWORD(w->f, xdenom);
  writeWMFWORD(w->f, xnum);
}

void
WMFSetMapMode(struct WMF *w, WMFWORD map)
{
  AddCommandX(w, META_SETMAPMODE, map);
}

void
WMFSetBKMode(struct WMF *w, WMFWORD mode)
{
  AddCommandX(w, META_SETBKMODE, mode);
}


void
WMFSetBKColor(struct WMF *w, WMFDWORD color)
{
  AddCommandL(w, META_SETBKCOLOR, color);
}

void
WMFSetTextColor(struct WMF *w, WMFDWORD color)
{
  AddCommandL(w, META_SETTEXTCOLOR, color);
}

void
WMFSetTextCharacterExtra(struct WMF *w, WMFWORD extra)
{
  AddCommandX(w, META_SETTEXTCOLOR, extra);
}

void
WMFSetROP2(struct WMF *w, WMFDWORD rop)
{
  AddCommandL(w, META_SETROP2, rop);
}

UWMFWORD
WMFCreatePenIndirect(struct WMF *w, UWMFWORD style, UWMFWORD width, WMFDWORD colour)
{
  addRecord(w, META_CREATEPENINDIRECT, 5);
  /* There seems to be a padding word between width and colour */
  writeWMFWORD(w->f, style);
  writeWMFWORD(w->f, width);
  writeWMFWORD(w->f, 0);
  writeWMFDWORD(w->f,colour);
  return AllocObject(w);
}

int
WMFCreateBrushIndirect(struct WMF *w, WMFWORD hatch, WMFDWORD colour, WMFWORD style)
{
  addRecord(w, META_CREATEBRUSHINDIRECT, 4);
  writeWMFWORD(w->f, style);
  writeWMFDWORD(w->f, colour);
  writeWMFWORD(w->f, hatch);
  return AllocObject(w);
}

int
WMFCreateRegion(struct WMF *w, WMFWORD left, WMFWORD top, WMFWORD right, WMFWORD bottom)
{
  AddCommandABCD(w, META_CREATEREGION, left, top, right, bottom);
  return AllocObject(w);
}

void
WMFSetWindowExt(struct WMF *w, WMFWORD x, WMFWORD y)
{
  AddCommandXY(w, META_SETWINDOWEXT, x, y);
}

void
WMFSetViewportExt(struct WMF *w, WMFWORD x, WMFWORD y)
{
  AddCommandXY(w, META_SETVIEWPORTEXT, x, y);  
}

void
WMFSelectObject(struct WMF *w, UWMFWORD nobj)
{
  AddCommandX(w, META_SELECTOBJECT, nobj);
}

void
WMFDeleteObject(struct WMF *w, UWMFWORD nobj)
{
  FreeObject(w, nobj);
  AddCommandX(w, META_DELETEOBJECT, nobj);  
}

void
WMFSaveDC(struct WMF *w)
{
  AddCommand(w, META_SAVEDC);
}

void
WMFRestoreDC(struct WMF *w, UWMFWORD lev)
{
  AddCommandX(w, META_RESTOREDC, lev);
}

void
WMFSetPolyFillMode(struct WMF *w, int fillmode)
{
  AddCommandX(w, META_SETPOLYFILLMODE, fillmode);
}

/*
  Clipping
*/
void
WMFExcludeClipRect(struct WMF *w, WMFWORD left, WMFWORD top, WMFWORD right, WMFWORD bottom)
{
  AddCommandABCD(w, META_EXCLUDECLIPRECT, left, top, right, bottom);
}

void
WMFIntersectClipRetc(struct WMF *w, WMFWORD left, WMFWORD top, WMFWORD right, WMFWORD bottom)
{
  AddCommandABCD(w, META_INTERSECTCLIPRECT, left, top, right, bottom);
}



/*
  void
  WMFSetBkColor(struct WMF *w, int x, int y)
  {
  AddCommandXY(w, 0x020B, x, y);
  }
*/


/*
  The last record in every metafile always has a function number of
  0000h, a Size of 0003h, and no Parameters array. This record is used
  to indicate the end of the record data in the metafile. The use of
  this terminator record is missing from the original WMF description
  found in the Windows SDK and is now documented in article Q99334 of
  the Microsoft Knowledge Base.  
*/

void
WMFFinish(struct WMF *w)
{
  AddCommand(w, 0x0);
  /* OK. Now, we should write the real values for the header */
  rewind(w->f);
  writeHEADER(w);
  free(w);
}

struct WMF *
WMFinit(FILE *f, int width, int height, int TYPE)
{
  int i;
  struct WMF *w;
  w= (struct WMF*) malloc(sizeof(struct WMF));
  /* unsigned char s[]= { 0xD7,  0xCD, 0xC6, 0x9A}; */
  /* Aldus Placeable file data */
  w->meta.Key= 0x9AC6CDD7; /* Always this value */
  w->meta.Handle= 0; /* always 0 */
  w->meta.Left=0; /* coordinate in twips (1/1440 inc) */
  w->meta.Top= 0;
  w->meta.Right= width;
  w->meta.Bottom= height;
  w->meta.Inch= 1440; /* metafile units per inch */
  w->meta.Reserved= 0;
  w->meta.Checksum= 0;
  /* Plain Metafile part */
  w->head.FileType= 1;
  w->head.HeaderSize= 9;
  w->head.Version= 0x0300;
  w->head.FileSize= 9;
  w->head.NumOfObjects= 0;
  w->head.MaxRecordSize= 0;
  w->head.NumOfParams= 0;
  /* administrative part */
  w->f= f;
  for (i=0; i< MAXOBJS; i++)
    w->objects[i]= -1;
  /* Ok. Now, we write the header */
  writeHEADER(w);
  /* Good. Now, commands can be written to the file */
  return w;
}

/* DEMO OF USE :
int
main()
{
  FILE *f;
  printf("%d\n", sizeof(WMFHEAD));
  if ((f= fopen("toto.wmf", "wb")) != 0)
    {
      struct WMF *w;
      WMFpoint pol[4]= {{20, 20}, {20, 40}, {40, 40}, {40, 20}};
      w= WMFinit(f);
      SetWindowOrg(&w, 0,0);
      SetWindowExt(&w, 200, 200);
      WMFMoveTo(w, 10, 10);
      WMFLineTo(w, 50, 100);
      WMFLineTo(w, 100, 50);
      WMFLineTo(w, 10, 10);
      WMFPolygon(w, pol, 4);
      WMFFinish(w);
      fclose(f);
    }
  /// writeWMF(&w, "toto.wmf");
  exit(0);
}
*/
