
package provide seshDico 0.0	;# This uses Tcl namespaces
package require stooop

package require TkseshDB

# Design remark :
#
# On ordering. 
#
# As far as translitterations are concerned, ordering doesn't work if one uses the sql capacities.
# a possible way to achieve ordering is to use a different ordering key, using for instance a mapping 
# between translitteration codes and ascii letters (for instance, mapping translitteration "A" to "a",
# "i" to "b", etc...
#
# this mapping belongs here, not to highter levels of the system.

class DicoDAL {
    set (dataSource) 0
    # Currently, we only use a static member
    proc DicoDAL {this} {
    }

    proc ~DicoDAL {this} {
    }
    
    # dicoDAL::writeOrthography --
    # 	saves a new orthography for a given translitteration
    # Parameters:
    # 	translit : a translitteration
    #	mdc : an orthography in manuel de codage
    # Options:
    # Results:
    # Side Effects:
    # 	orthography is added to already known saved orthographies
    # Use:
    #	PUBLIC
    # Called by:

    proc writeOrthography {translit mdc} {
	set db [TkseshDB::getTkseshDB]
	set translit [TkseshDB::protectText $translit]
	set ortho [TkseshDB::protectText $mdc]
	TkseshDB::update $db "insert into tmpOrtho values ('$translit', '$ortho')"
	if {[info exists (orthoHook)]} {
	    eval $(orthoHook) $translit
	}
    }

    # dicoDAL::readOrthography --
    # 	returns known orthographies for a word.
    # Parameters:
    # 	txt : the word translitteration
    # Options:
    # Results:
    # 	returns all possible written form for the given translitteration
    # Side Effects:
    # Use:
    #	PUBLIC
    # Called by:

    proc readOrthography {translit} {
	set db [TkseshDB::getTkseshDB]
	set translit [TkseshDB::protectText $translit]
	set res [TkseshDB::select $db "select ortho from tmpOrtho where translit = '$translit'"]
	return $res
    }

    # setOrthographyHook --
    #   defines a hook function that will be called 
    #   when some new orthography is saved in the dictionnary.
    #	This can be used to update a cache of already known orthographies.
    #
    # args --
    #    hook : a function that takes two arguments
    #           trans  : a translitteration
    #           val    : new orthography value
    
    proc setOrthographyHook {hook} {
	set (orthoHook) $hook
    }

    proc writeEntry {uid mid id date sid text} {
	set db [TkseshDB::getTkseshDB]
	
	if {$id ne "NULL"} {
	    set sql {
		update TmpDicoView
		set 
		entry = 'ENTRY',
		date = 'DATE',
		sid = SID
		where uid= UID 
		and   mid= MID
		and   id= ID
	    }
	} else {
	    set sql {
		insert into TmpDicoView (uid, mid, id, date, sid, entry)
		values (UID,MID,ID,'DATE',SID,'ENTRY');
	    }
	}
	
	set text [TkseshDB::protectText $text]
	
	set sql [string map [list UID $uid MID $mid SID $sid ID $id DATE $date ENTRY $text] $sql]
	# save the TmpDico part
	# save the header
	puts "doing $sql"
	puts -nonewline ".";  flush stdout
	TkseshDB::update $db $sql
    }

    proc getEntry {oid} {
	set db [TkseshDB::getTkseshDB]
	set sql "select entry from TmpDico where oid = $oid"
	puts "doing  $sql"
	return [lindex [TkseshDB::select $db $sql] 0]
    }
    
    proc getID {oid} {
	set db [TkseshDB::getTkseshDB]
	set sql \
	    "select id, uid, mid, date, sid from Header where repOid = $oid and type = 'TmpDico'"
	return [TkseshDB::select $db $sql]
    }

    proc saveIDX {tagName lookupText actualText oid} {
	set db [TkseshDB::getTkseshDB]
	set lk [TkseshDB::protectText $lookupText]
	set txt [TkseshDB::protectText $actualText]
	set sql "insert into tmpDicoIdx values ('$tagName', '$lk', '$txt', $oid)"
	TkseshDB::update $db $sql
    }

    proc deleteIDX {tagName lookupText actualText oid} {
	set db [TkseshDB::getTkseshDB]
	set lk [TkseshDB::protectText $lookupText]
	set txt [TkseshDB::protectText $actualText]
	set sql {
	    delete tmpDicoIdx 
	    where tag = 'TAG'
	    and lookup = 'LOOKUP'
	    and value = 'VALUE'
	    and ref = REF;
	}
	set sql [string map [list TAG $tagName LOOKUP $lk VALUE $txt REF $oid] $sql]
	TkseshDB::update $db $sql
    }


    proc getEntriesIdForTag {tagName lookupText findSubString} {
	if {$lookupText eq ""} {
	    return [list]
	}
	set db [TkseshDB::getTkseshDB]
	set lk [TkseshDB::protectText $lookupText]
	if {$findSubString} {
	    # PATCH FOR HIEROGLYPHIC CODES :
	    if {$tagName eq "WRI" || $tagName eq "DET" || $tagName eq "HIE"} {
		set sql "select distinct ref from tmpDicoIdx where lookup || ' ' GLOB '*$lk *' and tag = '$tagName'" 
	    } else {
		set sql "select distinct ref from tmpDicoIdx where lookup GLOB '*$lk*' and tag = '$tagName'" 
	    }
	} else {
	    set sql "select distinct ref from tmpDicoIdx where lookup = '$lk' and tag = '$tagName'" 
	}
	return [TkseshDB::select $db $sql]
    }

    proc dropAllIndexesFor {oid} {
	set db [TkseshDB::getTkseshDB]
	TkseshDB::update $db "delete from tmpDicoIdx where ref = $oid"
    }
    
    proc getLastDicoEntryOID {} {
	set db [TkseshDB::getTkseshDB]
	set l [TkseshDB::select $db "select lastOid from MaxTable where type = 'TmpDico'"]
	return [lindex $l 0]
    }

    proc PRIVRefreshDataSource {} {
	if {$(dataSource) != 0} {
	    #puts "refreshing $(dataSource)"
	    DataSource::refresh $(dataSource)
	}
	# Temporary patch !!!
	NoteDAL::PRIVRefreshDataSource
    }

    proc getOIDForID {idx} {
	set d [getDataSource]
	return [DataSource::getColIdx $d $idx "ref"]
    }

    proc getDataSource {} {
	if {$(dataSource) == 0} {
	    set db [TkseshDB::getTkseshDB]
	    set sql "select * from tmpDicoIdx where tag = 'MTRI'  order by lookup "
	    
	    set (dataSource) \
		[TkseshDB::buildDataSource $db $sql]
	    PRIVRefreshDataSource
	}
	return $(dataSource)
    }
    

    proc postEntry {} {
	PRIVRefreshDataSource
    }
}