##############################################################################
#
# Font editor built on the great CANVAS widget.
# 
#
# TO DO : 
# DONE :
# 	* undo/redo facility
#	* added scrollbar compatibility to the widget.
#	
# Notes :
#	BZR font stuff => special class. Here, we deal only with one char
#	Hiero DB system => automated saving of chars in font on demand.
#			   both BZR, metafont, etc... can be created.
#			   a user's chars are split in a number of fonts.
#			   auxiliary files can be produced (like tex files).
##############################################################################


class CharEdit {

    # -scrollregion {0 0 1000 1000}
    proc CharEdit {this parent name args} SCanvas {$parent $name -width 400 -height 400 -bg white -scrollregion {0 0 1000 1000} -takefocus 1 $args} {	
	set ($this,hasChanged) 0 ;# keep track of whether the current char has been modified.

	set ($this,ox) 0;		# center coordinates, in screen units.
	set ($this,oy) 0
	set ($this,oscale) 1;		#  original point to coordinate scale
	set ($this,scale) 0;		#  scale relative to initial size
	set ($this,rscale) 1;		#  oscale * exp(scale)
	set ($this,dscale) 1;		#  1 / rscale
	
	# These fields keep trace of the viewing window's minimal size.
	set ($this,minxview) 0
	set ($this,minyview) 0
	set ($this,maxxview) 0
	set ($this,maxyview) 0

	resetWindow $this

	setChanged $this 0

	set ($this,Contours) {}
	set ($this,selection) [new Selection $this]
	setNum $this 32
	set ($this,grid) 1
	set ($this,gridp) 0
	set ($this,guides) {} ; # graphic guides
	set ($this,oldx) 0
	set ($this,oldy) 0
	set ($this,dx) 0
	set ($this,dy) 0
	set ($this,rotationCenter) 0
	set ($this,type) "P"

	# The latest command. Simple undoing system
	set ($this,command) -1
	# l'tat spcifie l'action  mener
	
	set CharEdit::($this,state) edit
    }
    
    virtual proc creatorHook {this} {
	SCanvas::_creatorHook $this
    }

    virtual proc completorHook {this} {
	#Component::bind $this <Button-1> [list CharEdit::clickaction %O bouton1 %x %y]
	
	Component::bind $this <Button-1>\
	    [list CharEdit::clickaction $this bouton1 %x %y]
	
	Component::bind $this <B1-Motion>\
	    [list CharEdit::moveaction $this %x %y]	

	Component::bind $this <Motion>\
	    [list CharEdit::motionAction $this %x %y]	
	
	Component::bind $this <ButtonRelease-1>\
	    [list CharEdit::releaseaction $this %x %y]

	Component::bind $this <Key-Escape>\
	    [list CharEdit::abortActions $this]
	
	set ($this,bg) [Component::cget $this background]
    }
    
    proc ~CharEdit {this} {
	clear $this
	if { $CharEdit::($this,gridp) != 0 } {
	    delete $CharEdit::($this,screenpoint)
	}
	delete $CharEdit::($this,selection)
    }

    proc options {this} {}

    proc getContours {this} {
	return $CharEdit::($this,Contours)
    }

    proc clear {this} {
	foreach c  [getContours $this] {
	    delete $c
	}
	#foreach p $CharEdit::($this,points) {
	#    delete $p
	#}
    }

    proc toggleGrid {this} {
	set ($this,grid) [expr 1 - $CharEdit::($this,grid) ]
    }

    proc setgrid {this gx gy} {
	set ($this,grid) 1
	set ($this,gridp) [new Point [expr $gx * $CharEdit::($this,rscale)] \
			       [expr $gy * $CharEdit::($this,rscale)]]
			       
    }

    proc ongrid {this x y} {
	# Get the nearest active guide ????
	# (Currently, just a try)
	if {$CharEdit::($this,grid)} {
	    foreach g [getGuides $this] {
		set tmp [new Point $x $y]
		GraphGuide::attract $g $tmp
		return $tmp
	    }   
	}
	if 0 {
	    set gx [Point::x $CharEdit::($this,gridp)]
	    set gy [Point::y $CharEdit::($this,gridp)]
	    set x [expr floor(1.0* $x / $gx) * $gx]
	    set y [expr floor(1.0* $y / $gy) * $gy]
	}
	return [new Point $x $y]
    }
    
    #proc getwindow {this} {
    #return [Component::getPath $this]
    #}

    
    ################################################################################
    #
    #	       Scaling and coordinates related functions
    #
    
    proc fixscale {this} {
	set ($this,rscale) [expr $CharEdit::($this,oscale) * exp($CharEdit::($this,scale))]
	set ($this,dscale) [expr 1.0 / $CharEdit::($this,rscale)]
    }

    # Sets original scale.
    proc setscale {this sc ox oy} {
	set ($this,ox) $ox
	set ($this,oy) $oy
	set ($this,oscale) $sc
	fixscale $this
    }


    # newcenter --
    # 	Sets the image center (DEPRECATED)
    #	x, y: current *window* space coordinates of the new center
    
    proc newcenter {this x y} {
	set w [Component::getPath $this]
	# We retrieve the real center of the window ?
	set centerx [expr [$w cget -width] / 2.0]
	set centery [expr [$w cget -height] / 2.0]
	# we conpute the vector needed to place x,y at the window center
	set dx [expr  $centerx - $x]
	set dy [expr  $centery - $y]
	# we translate everything by this vector.
	set ($this,ox)	[expr $CharEdit::($this,ox) + $dx]
	set ($this,oy)	[expr $CharEdit::($this,oy) + $dy]
	$w move all $dx $dy
    }

    # newUpperLeft --
    #	Sets the image new upper left point
    #	x, y : real coordinates of the new upper left point.
    
    proc newUpperLeft {this x y} {
	set w [Component::getPath $this]
	# The change will mainly take place in canvas coordinate space,
	# that is *window space* and not real space. So, what we do is :
	set ul [screenpoint $this $x $y]
	set dx [expr -[Point::x $ul]]
	set dy [expr -[Point::y $ul]]
	delete $ul
	# we translate everything by (dx,dy).
	set ($this,ox)	[expr $CharEdit::($this,ox) + $dx]
	set ($this,oy)	[expr $CharEdit::($this,oy) + $dy]
	$w move all $dx $dy
	notifyScrollBars $this
    }

    proc scaleall {this scale} {
	set sc [expr  exp($scale - $CharEdit::($this,scale))]
	set w [Component::getPath $this]
	set centerx [expr [$w cget -width] / 2.0]
	set centery [expr [$w cget -height] / 2.0]
	set ($this,scale) $scale
	set ($this,ox)	[expr ($CharEdit::($this,ox) - $centerx) * $sc  + $centerx]
	set ($this,oy)	[expr ($CharEdit::($this,oy) - $centery) * $sc  + $centery]
	fixscale $this
	if {$sc != 0} {
	    [Component::getPath $this] scale all $centerx $centery\
		$sc $sc
	}
	notifyScrollBars $this
    }

    proc getScale {this} {
	return $($this,scale)
    }


    # screenpoint --
    # 	return a screen point from original coordinates
    #
    # 	returns : a Point object representing the original point, in screen coordinates.
    proc screenpoint {this x y} {
	set sc $CharEdit::($this,rscale)
	set xx [expr $x * $CharEdit::($this,rscale)  + $CharEdit::($this,ox)]
	set yy [expr (- $y )* $CharEdit::($this,rscale) + $CharEdit::($this,oy)]
	return [new Point $xx $yy]
    }

    # realPoint --
    # 	returns a file point from screen coordinates
    #	p : a point in screen coordinates
    #	returns : a list of x,y coordinates.

    proc realPoint {this p} {
	set x [Point::x $p]
	set y [Point::y $p]
	return [simpleRealPoint $this $x $y]
    }

    # simpleRealPoint --
    # 	returns a file point from screen coordinates
    #	x, y : screen coordinates
    #	returns : a list of x,y coordinates in real coordinate space
    
    proc simpleRealPoint {this x y} {
	set xx [expr 1.0 * ($x - $CharEdit::($this,ox)) * ($CharEdit::($this,dscale))]
	set yy [expr 1.0 * (- $y + $CharEdit::($this,oy)) * ($CharEdit::($this,dscale))]
	return [list $xx $yy]	
    }

    #
    # ScrollBar handling functions
    #

    proc setXScrollBar {this sb} {
	set ($this,xsb) $sb
    }

    proc setYScrollBar {this sb} {
	set ($this,ysb) $sb
    }

    proc getXScrollBar {this} {
	return $($this,xsb)
    }

    proc getYScrollBar {this} {
	return $($this,ysb)
    }
    
    # A) scrollbar setting functions.
    # 
    proc notifyScrollBars {this} {
	# the catch is here in case we don't have scrollbar
	catch {
	}
	# We want to call the "set" functions of the scrollbar
	# with the appropriate values.
	# First argument is the starting point, as a fraction of visible 
	# space. Second is the last visible point, in the same unit.
	# Let's start doing it the "right" way.

	# A) compute the used space.

	set p [Component::getPath $this]
	# The viewing window :
	
	set rll [simpleRealPoint $this 0 [$p cget -height]]
	set rur [simpleRealPoint $this [$p cget -width] 0]


	# get sure the total area is at least big enough to contain the actual viewing window :
	computeWindow $this [lindex $rll 0] [lindex $rll 1] [lindex $rur 0] [lindex $rur 1]

	# Ok. Now, compute this as percentage of the viewing window.

	set width $($this,viewWidth)
	set height $($this,viewHeight)

	set minx [expr ([lindex $rll 0]-$($this,minxview))/$width]
	set maxx [expr ([lindex $rur 0]-$($this,minxview))/$width]

	# Note that the Y scrollbar is pointing down.

	set maxy [expr ($($this,maxyview)-[lindex $rll 1])/$height]
	set miny [expr ($($this,maxyview)-[lindex $rur 1])/$height]
	
	[getXScrollBar $this] set $minx $maxx

	[getYScrollBar $this] set $miny $maxy
    }

    # Update the viewable window. The viewable window is expressed in 
    # real coordinates, and can only grow, except if a reset is done.
    
    proc computeWindow {this llx lly urx ury} {
	if {$llx < $($this,minxview)} {
	    set ($this,minxview) $llx
	}
	if {$lly < $($this,minyview)} {
	    set ($this,minyview) $lly
	}
	if {$urx > $($this,maxxview)} {
	    set ($this,maxxview) $urx
	}
	if {$ury > $($this,maxyview)} {
	    set ($this,maxyview) $ury
	}
	set ($this,viewWidth) [expr $($this,maxxview)-$($this,minxview)]
	set ($this,viewHeight) [expr $($this,maxyview)-$($this,minyview)]
    }

    proc resetWindow {this} {
	set ($this,minxview) -42
	set ($this,minyview) -42
	set ($this,maxxview) 42
	set ($this,maxyview) 42
	computeWindow $this 0 0 0 0
    }

    #
    # B) functions called by the scrollbar
    
    proc xview {this comm args} {
	switch -exact -- $comm {
	    moveto {
		set val [lindex $args 0]
		# Ok. Now, compute the new upper left coordinates.
		# First, the current ones :
		set oldUL [simpleRealPoint $this 0 0]
		# The x change.
		set newx [expr $($this,minxview)+$val*$($this,viewWidth)]
		newUpperLeft $this $newx [lindex $oldUL 1]
	    }
	    scroll {
		set val [lindex $args 0]
		set oldUL [simpleRealPoint $this 0 0]
		set newx [expr $val*7.0 + [lindex $oldUL 0]]
		newUpperLeft $this  $newx [lindex $oldUL 1]
	    }
	}
    }

    proc yview {this comm args} {
	switch -exact -- $comm {
	    moveto {
		set val [lindex $args 0]
		# Ok. Now, compute the new upper left coordinates.
		# First, the current ones :
		set oldUL [simpleRealPoint $this 0 0]
		# The y change.
		set newy [expr $($this,maxyview)-$val*$($this,viewHeight)]
		newUpperLeft $this [lindex $oldUL 0] $newy 
	    }
	    scroll {
		set val [lindex $args 0]
		set oldUL [simpleRealPoint $this 0 0]
		set newy [expr -$val*7.0 + [lindex $oldUL 1]]
		newUpperLeft $this [lindex $oldUL 0] $newy 
	    }
	}
    }

    proc ajouteContour {this x y {type T}} {
	set c [new Contour $this $type]
	switch -exact -- $type {
	    T {
		set pp1 [ongrid $this $x $y]
		set p1 [new ContourControlPoint $pp1 $c]
		set p2 [new ContourControlPoint $pp1 $c]
		delete $pp1
		set s1 [new LineSegment $c $p1 $p2]
		set Contour::($c,segments) [list $s1]
	    }
	    default {
		set pp1 [ongrid $this $x $y]
		set p1 [new ContourControlPoint $pp1 $c]
		set p2 [new ContourControlPoint $pp1 $c]
		delete $pp1
		set s1 [new LineSegment $c $p1 $p2]
		set s2 [new LineSegment $c $p2 $p1]
		set Contour::($c,segments) [list $s1 $s2]
		#Contour::attachsegment $c $s1
		#set Contour::($c,segments) [list $s1]
	    }
	}
    }
    
    proc addGuide {this x y {type line}} {
	set pp1 [new Point $x $y]
	set p1 [new ControlPoint $pp1 $this]
	set p2 [new ControlPoint $pp1 $this]
	delete $pp1
	set c [new LineGuide $this $p1 $p2]
	lappend CharEdit::($this,guides) $c
    }

    proc getGuides {this} {
	return $CharEdit::($this,guides)
    }

    proc auxdeleteContour {this Contour} {
	Selection::suppress $($this,selection) $Contour
	listreplace CharEdit::($this,Contours) $Contour
    }
    
    proc editclosest {this x y} {
	set can [Component::getPath $this]
	set ident [$can find  closest $x $y]
	#if {[info exists ContourControlPoint::(control,$can,$ident)]} {
	#set CharEdit::($this,selected) $ContourControlPoint::(control,$can,$ident)
	#}
    }
    
    proc clickaction {this bouton x y} {
	clearRotation $this
	focus [Component::getPath $this]
	
	set ($this,oldx) $x
	set ($this,oldy) $y
	# add	{CharEdit::ajouteContour $this $x $y $CharEdit::($this,type)}
	switch $CharEdit::($this,state) {
	    add	{command $this [new NewContourCommand $this $x $y $CharEdit::($this,type)]}
	    edit {}
	    select {}
	    center {newcenter $this $x $y}
	    delete {}
	    addguide {CharEdit::addGuide $this $x $y line}
	    hreflect {CharEdit::reflectHorizontal $this $x $y}
	    vreflect {CharEdit::reflectVertical $this $x $y}
	    rotate {CharEdit::rotateSelected $this $x $y}
	}
    }

    proc selectAll {this} {
	Selection::clear $($this,selection)
	foreach c [getContours $this] {
	    Contour::select $c
	}
    }

    proc deselectAll {this} {
	Selection::clear $($this,selection)
    }

    proc reflectHorizontal {this x y} {
	set p [new Point $x $y]
	set d [new Point 0 1]
	set t [AffineTransformation::makeLineReflection $p $d]
	foreach c [Selection::getContours $($this,selection)] {
	    Contour::appliTransformation $c $t
	}
	delete $t
	delete $d
	delete $p
    }


    proc reflectVertical {this x y} {
	set p [new Point $x $y]
	set d [new Point 1 0]
	set t [AffineTransformation::makeLineReflection $p $d]
	foreach c [Selection::getContours $($this,selection)] {
	    Contour::appliTransformation $c $t
	}
	delete $t
	delete $d
	delete $p
    }

    proc rotateSelected {this x y} {
	if {$($this,rotationCenter) != 0} {
	    set p $($this,rotationCenter)
	    # We want to compute the cos and sin of the rotation.
	    set dx [expr {$x - [Point::x $p]}]
	    set dy [expr {$y - [Point::y $p]}]
	    set d [expr {hypot($dx,$dy)}]
	    if {$d > 0} {
		set cos [expr {$dx/$d}]
		set sin [expr {$dy/$d}]
		
		set t [AffineTransformation::makeRotationByCosAndSin $p $cos $sin]
		foreach c [Selection::getContours $($this,selection)] {
		    Contour::appliTransformation $c $t
		}
		delete $t
	    }
	} else {
	    set ($this,rotationCenter) [new Point $x $y]
	}
    }
    
    proc releaseaction {this x y} {
	switch $CharEdit::($this,state) {
	    add	{}
	    edit {}
	    select {}
	}
    }
    
    proc abortActions {this} {
	#puts "aborting"
	clearRotation $this
	if {$($this,rotationCenter) != 0} {
	    delete $($this,rotationCenter)
	}
	set ($this,rotationCenter) 0
    }

    proc moveaction {this x y} {
	set ($this,dx) [expr $x - $CharEdit::($this,oldx)]
	set ($this,dy) [expr $y - $CharEdit::($this,oldy)]
	set ($this,oldx) $x
	set ($this,oldy) $y	
	switch $CharEdit::($this,state) {
	    add	{}
	    edit {}
	    select {}
	    scale {}
	}
    }

    proc motionAction {this x y} {
	clearRotation $this
	switch $CharEdit::($this,state) {
	    add	{}
	    edit {}
	    select {}
	    scale {}
	    rotate {CharEdit::drawRotationReference $this $x $y } 
	}
    }

    proc clearRotation {this} {
	set w [Component::getPath $this]
	$w delete rotation
    }

    proc drawRotationReference {this x y} {
	if {$($this,rotationCenter) != 0} {
	    set px [Point::x $($this,rotationCenter)]
	    set py [Point::y $($this,rotationCenter)]
	    set w [Component::getPath $this]
	    if {$x != $px || $y != $py} {
		#Compute the bloody angle :
		set hyp [expr {hypot($x-$px,$y-$py)}]
		
		set cosalpha [expr {($x-$px)/$hyp}]
		set angle [expr {acos($cosalpha)}]
		#puts "vector ($px,$py) => ($x,$y) ; [expr {$x-$px}], [expr {$y-$py}]"
		if {($y-$py) < 0} {
		    set angle -$angle
		}
		
		# Now, convert. Note that the y axes in screen space is oriented
		# downward, hence the minus sign.
		set angle [expr {-180.0*$angle/acos(-1)}]
		
		#puts "angle is $angle"
		
		$w create arc [expr {$px - $hyp}] [expr {$py + $hyp}] \
		    [expr {$px + $hyp}] [expr {$py - $hyp}] -start 0 -extent $angle \
		    -outline red -dash -.. -tags rotation 
	    }
	}

    }

    proc pasteAction {this txt} {
	command $this [new PasteCommand $this $txt]
    }

    
#      proc load {this font} {
#  	set ($this,fontfile) $font
#  	set ($this,char) 32
#  	set ($this,font) [sbzr_newfont $font]
#  	clear $this
#  	setchar $this 32
#      }
    
    
    # asText --
    # 	returns a text representation of the current char.
    #	The representation includes enough data to define a BZR char.

    proc asText {this} {
	set res ""
	# We use tk canvas' bounding box. As it is not that precise,
	# we first have to zoom the text a little.
	# but we need to revert to the current zoom after.
	set scale $CharEdit::($this,scale)
	foreach c [getContours $this] {
	    lappend res [Contour::asText $c]
	}
	scaleall $this 10
	# If there is no contour elt, bb will not return anything. 
	# So, we create an default bb
	foreach {llx ury  urx lly}  {0 0 0 0} {}
	# get the actual bounding box values
	foreach {llx ury  urx lly}  [[Component::getPath $this] bb Contourelt] {}
	set v1 [simpleRealPoint $this $llx $lly]
	set v2 [simpleRealPoint $this $urx $ury]
	set width [expr [lindex $v2 0] - [lindex $v1 0]]
	scaleall $this 0
	set res [concat [getNum $this] $width $v1 $v2 $res]
	scaleall $this $scale
	return $res
    }

    proc getSelection {this} {
	return [Selection::asText $($this,selection)]
    }

    proc loadimg {this name} {
	set CharEdit::($this,img) [new Image $this $name]
    }
    
    proc delimg {this} {
	delete $CharEdit::($this,img)
    }
    
    proc lower {this c} {
	set w [Component::getPath $this]
	$w lower $c
	catch {Image::lower $CharEdit::($this,img)}
    }

    # Lowers a contour 
    proc lowerContour {this c} {
	listreplace CharEdit::($this,Contours) $c
	set CharEdit::($this,Contours) [linsert $CharEdit::($this,Contours) 0 $c]
    }

    # Raises a contour

    proc raiseContour {this c} {
	listreplace CharEdit::($this,Contours) $c
	lappend CharEdit::($this,Contours) $c
    }


    # draws reference lines
    proc repere {this} {
	set p1 [screenpoint $this 0 0]
	set p2 [screenpoint $this 0 14]
	set p3 [screenpoint $this 14 0]
	set w [Component::getPath $this]
	set u [$w create line [Point::x $p1] [Point::y $p1] [Point::x $p2] [Point::y $p2] -fill red]
	set v [$w create line [Point::x $p1] [Point::y $p1] [Point::x $p3] [Point::y $p3] -fill red]
	$w addtag repere withtag $u
	$w addtag repere withtag $v
    }

    #
    # addtext -- 
    #	add a list of contours to the current text.
    #

    proc addtext {this text} {
	foreach c $text {
	    set pt [screenpoint $this [lindex $c 2] [lindex $c 3]]
	    set sens [lindex $c 1]
	    
	    set leContour [new Contour $this $sens]

	    set cpt [new ContourControlPoint $pt $leContour]
	    set premier $cpt
	    delete $pt

	    set paths [lrange $c 4 end]
	    set l [llength $paths]
	    
	    set i 0
	    foreach seg $paths {
		incr i
		switch -exact [lindex $seg 0] {
		    s { 
			set c1 [screenpoint $this [lindex $seg 1] [lindex $seg 2]]
			set c2 [screenpoint $this [lindex $seg 3] [lindex $seg 4]]
			set cc1 [new ContourControlPoint $c1 $leContour]
			set cc2 [new ContourControlPoint $c2 $leContour]
			delete $c1
			delete $c2
			if { $i == $l && [string compare $sens "T"] != 0} {
			    set cp2 $premier
			} else {
			    set p2 [screenpoint $this [lindex $seg 5] [lindex $seg 6]]
			    set cp2 [new ContourControlPoint $p2 $leContour]
			    delete $p2
			}
			Contour::attachSegment $leContour [new Spline $leContour $cpt $cc1 $cc2 $cp2]
		    }
		    l {
			if { $i == $l && [string compare $sens "T"] != 0} {
			    set cp2 $premier
			} else {
			    set p2 [screenpoint $this [lindex $seg 1] [lindex $seg 2]]
			    set cp2 [new ContourControlPoint $p2 $leContour]
			    delete $p2
			}
			Contour::attachSegment $leContour [new LineSegment $leContour $cpt $cp2]
		    }
		}
		set cpt $cp2
	    }
	}
    }

    # setText --
    # sets the displayed value to conform to txt.
    #
    # 	txt : is the text as given by astext
    
    proc setText {this txt} {
	clear $this
	setNum $this [lindex $txt 0]
	addtext $this [lrange $txt 6 end]
    }

    proc setNum {this num} {
	set ($this,char) $num
    }

    proc getNum {this} {
	return $($this,char)
    }
    # setchar --
    # 	loads a char from the current font.
    #	DEPRECATED
#      proc setchar {this char} {
#  	set ($this,char) $char
#  	# destruction of everything
#  	set crep [sbzr_getchar $CharEdit::($this,font) $char]
#  	addtext $this [lrange $crep 6 end]
#      }

#      proc save {this {name noname.bzr}} {
#  	sbzr_savefont $CharEdit::($this,font) $name 
#      }

#    
#	save the current char as 
#      proc savechar {this} {
#  	sbzr_setchar $CharEdit::($this,font) $CharEdit::($this,char) \
#  	    [asText $this] 
#      }

#      proc getchar {this} {
#  	clear $this
#  	setchar $this $CharEdit::($this,char)
#      }

#      proc nextchar {this} {
#  	CharEdit::clear $this
#  	incr CharEdit::($this,char)
#  	CharEdit::setchar $this $CharEdit::($this,char)
#      }

#      proc prevchar {this} {
#  	CharEdit::clear $this
#  	incr CharEdit::($this,char) -1
#  	CharEdit::setchar $this $CharEdit::($this,char)
#      }
    
    proc invertlr {this} {
	
	[Component::getPath $this] delete withtag repere
	foreach {x1 y1 x2 y2} [[Component::getPath $this] bbox all] {}
	set centerx [expr ($x1 + $x2) / 2.0]
	set centery [expr ($y1 + $y2) / 2.0]
	[Component::getPath $this] scale all $centerx $centery -1 1
	repere $this
    }

    proc invertud {this} {
	[Component::getPath $this] delete withtag repere
	foreach {x1 y1 x2 y2} [[Component::getPath $this] bbox all] {}
	set centerx [expr ($x1 + $x2) / 2.0]
	set centery [expr ($y1 + $y2) / 2.0]
	[Component::getPath $this] scale all $centerx $centery 1 -1
	repere $this
    }

    # Command System.
    #
    
    proc command {this command} {
	if {$($this,command)  != -1} {
	    delete $($this,command)
	}
	set ($this,command) $command
	MyCommand::execute $command
    }

    proc undo {this} {
	MyCommand::undo $($this,command)
    }

    proc getPhoto {this} {
	return $($this,photo)
    }

    proc makePhoto {this} {
	set ($this,photo) [asText $this]
    }

    proc setChanged {this changed} {
	set ($this,changed) $changed
    }
    
    proc getChanged {this} {
	return $($this,changed)
    }

    proc setNoFill {this} {
	set ($this,type) "S"
    }

    proc setOpenCurve {this} {
	set ($this,type) "T"
    }

    proc setFilled {this} {
	set ($this,type) "P"
    }

    proc setUnFilled {this} {
	set ($this,type) "N"
    }

    proc getFill {this} {
	return $($this,type)
    }
}
