
# Control points : attached to segments which are attached to Contours.
# Move to :
# 	ContourControlPoint
#	etc...
#	all derived from ControlPoint


# A ControlPoint is a point with associated controled forms

class ControlPoint {

    proc ControlPoint {this p win} \
	    GraphElt {$win} Point {[Point::x $p] [Point::y $p]} {
	#set w $GraphElt::($owner,window)
	set w $win
	
	# Controled forms are stored in this variable :
	set ($this,controled) {}
	
	# build our representation
	
	GraphElt::draw $this

	set can $GraphElt::($this,canvas)
	set draw [GraphElt::getDrawing $this]

	$can bind $draw <Enter> [list GraphElt::setRed $this]
	$can bind $draw <Leave> [list GraphElt::unsetRed $this]
	$can bind $draw <B1-Motion> "ControlPoint::moveaction $this %x %y; break"
	$can bind $draw <B1-ButtonRelease> "ControlPoint::releaseAction $this %x %y; break"
	$can bind $draw <Button-1>  "ControlPoint::clickaction $this %x %y; break"
    } 

    proc ~ControlPoint {this} {
	set w $GraphElt::($this,window)
	unset ($this,controled)
	# listreplace CharEdit::($w,points) $this
	# destructeur
    }

    # Action to be taken when the point is moved.
    # Undoing for edit is controled by releaseAction
    
    proc moveaction {this x y} {
	set state $CharEdit::($GraphElt::($this,window),state)
	switch -exact $state {
	    {edit}  {move $this $x $y}
	}
    }

    # Action to be taken when the point is clicked
    
    proc clickaction {this x y} {
	set state $CharEdit::($GraphElt::($this,window),state)
	#{delete} {deletePoint $this; puts "destruction de point"}
	set editor [GraphElt::getCharEdit $this]
	switch -exact $state {
	    {edit} {CharEdit::makePhoto $editor}
	    {select}  {}
	    {delete} {
		CharEdit::command $editor [new PointDeleteCommand $this]
	    }
	}
    }

    # Action to be taken when the mouse is released
    
    proc releaseAction {this x y} {
	set state $CharEdit::($GraphElt::($this,window),state)
	switch -exact $state {
	    {select}  {}
	    {delete} {}
	    {edit} {
		set editor [GraphElt::getCharEdit $this]
		CharEdit::command $editor [new MoveCommand $this $x $y [CharEdit::getPhoto $editor]]
	    }
	}
    }
    
    # State that point controls the "controled" GraphElt.
    # For most purposes, this function should probably be redefined
    
    virtual proc control {this controled} {
	set can $GraphElt::($this,canvas)
	# Each GraphElt has an associated tag which allows to
	# change it in a global way.
	# So, give this tag to the point
	# GraphElt::tagAsMember $this $Segment::($controled,owner)
	GraphElt::tagAsMember $this $controled
	#lappend ($this,controled) $segment
	lappend ($this,controled) $controled
    }

    proc getControled {this} {
	return $ControlPoint::($this,controled)
    }
    # Unlinked a controled item from the point. If nothing is controled, destroy the point.
    
    proc detachAndDeleteIfOrphan {this controled} {
	detach $this $controled
	if {[llength $ControlPoint::($this,controled)] == 0} {
	    delete $this
	}
    }

    # Unlinked a controled item from the point.

    proc detach {this controled} {
	listreplace ControlPoint::($this,controled) $controled
    }

    # moves the point to x, y.
    proc move {this x y} {
	set p [CharEdit::ongrid $GraphElt::($this,window) $x $y]
	set x [Point::x $p]
	set y [Point::y $p]
	delete $p
	setcoords $this $x $y
    }

    proc setcoords {this x y} {
	$GraphElt::($this,canvas) coords [GraphElt::getDrawing $this] $x $y
	foreach i $ControlPoint::($this,controled) {
	    GraphElt::update $i
	}
    }

    virtual proc setRed {this} {	
    }

    virtual proc unsetRed {this} {
    }
    
    proc createDrawing {this} {
	set can $GraphElt::($this,canvas)
	set x $Point::($this,x)
	set y $Point::($this,y)
	set circ [$can create image $x $y -anchor center -image $ControlPoint::(image)]
	$can addtag ControlPoint withtag $circ 
	unset Point::($this,x)
	unset Point::($this,y)
	return $circ
    }

    virtual proc update {this} {}
    
    proc x {this} {
	set cs [$GraphElt::($this,canvas) coords [GraphElt::getDrawing $this]]
	return [lindex $cs 0] 
    }

    proc y {this} {
	set cs [$GraphElt::($this,canvas) coords [GraphElt::getDrawing $this]]
	return [lindex $cs 1]
    }

    virtual proc middle {this m} {
	set p [Point::middle $this $m]
	set ret [new ControlPoint $p $ControlPoint::($this,owner)]
	delete $p
	return $ret
    }

    proc asText {this} {
	set c [CharEdit::realPoint $GraphElt::($this,window) $this]
 	return [format {%.4f %.4f} [lindex $c 0] [lindex $c 1]]
    }
    
    # deletes a Point.
    virtual proc deletePoint {this} {
	#deletes a control Point
	# If the Point is an internal control, ask for its deletion.
    }

    set ControlPoint::(image) [image create bitmap -data {
	#define square_width 8
	#define square_height 8
	static unsigned char square_bits[] = {
	    0xff, 0x81, 0x81, 0x81, 0x81, 0x81, 0x81, 0xff, };
	} ]
}
