#
# widget to display and select BZR font chars.
#

package require seshw

class FontDisplay {

    proc FontDisplay {this parent appli name args} SScrolledWindow {$parent $name $args} {
	
	set ($this,appli) $appli
	set ($this,modified) 0
	set ($this,drawingSystem) 0
    }

    proc ~FontDisplay {this} {
    }
    
    #virtual proc creatorHook {this} {
    #	SScrolledWindow::_creatorHook $this
    #}

    virtual proc completorHook {this} {
	# A note on ScrollableFrame : ScrollableFrame *doesn't* have scrollbars !
	# it's meant to be used with ScrolledWindow
	
	set frame [Component::buildNewComponent SScrollableFrame $this -auto]
	set ($this,frame) [FrameContainer::getFrame $frame]
	SScrolledWindow::setWidget $this $frame

	setChanged $this 0
	
	for {set i 0} {$i < 256} {incr i} {
	    set ($this,label,$i) [label [Component::getPath $($this,frame)].l$i -text $i]
	    set ($this,button,$i) [GlyphButton::create [Component::getPath $($this,frame)].b$i $i $this]
	    grid $($this,label,$i) $($this,button,$i)
	}
	
	# We used to pack buttons one by one. It's soooooo expensive.
	# It's much faster to pack all buttons in one instruction :
	# eval pack [listButtons $w]
	# well, grid is fast enough, so we use grid. But we keep the previous instruction
	# as a useful reminder.
    }
    

    proc openChar {b} {
	set fontDisplay [GlyphButton::getParent $b]
	set app [getAppli $fontDisplay]
	#set font [getFont $fontDisplay]
	set num [GlyphButton::getCode $b]
	#set txt [BZRFont::getChar $font $num]
	# The following line creates a dependency between FontDisplay and the CharEditAppli.
	# This should not be. A simple solution would be to use a callBack
	# CharEditAppli::editFontChar $app $txt [list FontDisplay::setChar $fontDisplay $num] "[BZRFont::getFileName $font]($num)" $fontDisplay
	CharEditAppli::editFontChar $app $fontDisplay $num 
    }
    
    # Part of setChar should move to CharEditAppli.
    # Which should act as a MEDIATOR.
    # FontDisplay tells MEDIATOR : edit char so and so from me.
    # MEDIATOR creates a 

    proc setChar {this num txt} {
	BZRFont::setChar [getFont $this] $num $txt
	updateChar $this $num
	setChanged $this 1
    }

    proc getAppli {this} {
	return $($this,appli)
    }
    
    proc getButton {this i} {
	return $($this,button,$i)
    }
    
    proc updateChar {this i} {
	set f [getFont $this]
	if {[BZRFont::hasChar $f $i]} {
	    set txt [BZRFont::getChar $f $i]
	    GlyphButton::update [getButton $this $i] $txt
	}
    }

    proc setFont {this f} {
	set ($this,font) $f
	for {set i 0} {$i < 256} {incr i} {
	    updateChar $this $i
	}
	setChanged $this 0
    }
    
    proc getFont {this} {
	return $($this,font)
    }

    proc getChanged {this} {
	return $($this,changed)
    }

    proc setChanged {this v} {
	set ($this,changed) $v
    }
}


# As we create many buttons, and as object creation is a bit slow in Stooop, 
# we decided to use a simpler paradigm for our buttons
# Note : $this == the path here.


namespace eval GlyphButton {
    variable VAR
    variable drawingSystem
    
    proc create {this code fontDisplay} {
	variable VAR
	set VAR($this,code) $code
	set VAR($this,parent) $fontDisplay
	set img [image create bitmap [getImageName $this]]
	button $this -width 40 -height 40 -image [getImageName $this]
	#button $this -width 160 -height 160 -image [getImageName $this]
	bind $this <Double-Button-1> [list FontDisplay::openChar $this]
	return $this
    }
    
    proc delete {this} {
	array unset VAR $this,*
    }
    
    proc getImageName {this} {
	return "fonticon $this"
    }
    
    proc getCode {this} {
	variable VAR
	return $VAR($this,code)
    }

    proc getParent {this} {
	variable VAR
	return $VAR($this,parent)
    }
    proc getDrawingSystem {} {
	variable drawingSystem
	if {! [info exists drawingSystem]} {
	    seshw::create_passive .dummyDrawer -htline 40
	}
	set drawingSystem .dummyDrawer
	return $drawingSystem
    }

    proc update {this txt} {
	set d [getDrawingSystem]
	tksesh_defchar temp $txt
	seshw::set_text $d temp
	seshw::autosize $d
	set img [image create bitmap [getImageName $this] -data [$d toxbm 1.0 end]]
	$this configure -image [getImageName $this]
    }

}
