#! /bin/sh
#
# WARNING: DO NOT USE tcl < 8.0 !
#
# spline.tcl - Copyright (c) 2001 by Helmar Wodtke
#
# edited for the CharEdit system by Serge Rosmorduc

# Useful functions :

# flatten -- transforms a spline into lines
# point_on_spline x y a smooth? delta? -- returns the "t" value for point x,y on the spline defined by
#                 coordinates a
# subdivide C N : C represents a spline, N is a value between 0 and 1.
#	returns 14 coordinates.
#       The first 8 correspond to the first spline.
#       The last 8 correspond to second spline. Coords 6 and 7 are shared by the two splines !


namespace eval SplineAux {
    proc vec {a p} {
	return [lrange $a [expr $p*2] [expr $p*2+1]]
    }
    
    proc scale {f p} {
	set r {}
	foreach n $p {
	    lappend r [expr $n*$f]
	}
	return $r
    }
    
    proc translate {p1 p2} {
	set l [llength $p1]
	set r {}
	for {set i 0} {$i < $l} {incr i} {
		lappend r [expr [lindex $p1 $i]+[lindex $p2 $i]]
	}
	return $r
    }
    
    proc subdivide {p {t .5}} {	# must give 8 for p, returns 14
	
	set t2 [expr 1-$t]
	set r [translate [scale $t2 [vec $p 1]] [scale $t [vec $p 2]]]
	set q1 [translate [scale $t2 [vec $p 0]] [scale $t [vec $p 1]]]
	set q2 [translate [scale $t2 $q1] [scale $t $r]]
	set q5 [translate [scale $t2 [vec $p 2]] [scale $t [vec $p 3]]]
	set q4 [translate [scale $t2 $r] [scale $t $q5]]
	set q3 [translate [scale $t2 $q2] [scale $t $q4]]
	
	return [concat [lrange $p 0 1] $q1 $q2 $q3 $q4 $q5 [lrange $p 6 7]]
    }
    
    proc subdiv2 {x1 y1 x2 y2 x3 y3 x4 y4} {	# this is a faster subdiv 
	if {$x1 == $x2 &&
	$y1 == $y2 &&
	$x3 == $x4 &&
	$y3 == $y4} {	# okay, this is stupid
	    return [list $x1 $y1 $x2 $y2 $x3 $y3]
	}
	set p1x [expr ($x2+$x1)/2.0]
	set p1y [expr ($y2+$y1)/2.0]
	set p2x [expr ($x3+$x2)/2.0]
	set p2y [expr ($y3+$y2)/2.0]
	set p3x [expr ($x4+$x3)/2.0]
	set p3y [expr ($y4+$y3)/2.0]
	set m1x [expr ($p2x+$p1x)/2.0]
	set m1y [expr ($p2y+$p1y)/2.0]
	set m2x [expr ($p3x+$p2x)/2.0]
	set m2y [expr ($p3y+$p2y)/2.0]
	set m3x [expr ($m2x+$m1x)/2.0]
	set m3y [expr ($m2y+$m1y)/2.0]
	return [list $x1 $y1 $p1x $p1y $m1x $m1y $m3x $m3y \
		$m2x $m2y $p3x $p3y]
    }
    
    proc smooth coords {
	set ll [llength $coords]
	incr ll -2
	set i 0
	set r {}
	while {$ll > $i} {
	    set r [concat $r \
		    [eval subdiv2 [lrange $coords $i [expr $i+7]]]]
	    incr i 6
	}
	lappend r [lindex $coords $i] [lindex $coords [expr $i+1]]
	return $r
    }
    
    proc simplify a {
	set r [lrange $a 0 1]
	set ll [llength $a]
	set ll [expr $ll-3-$ll%4]
	foreach {x1 y1 x2 y2} [lrange $a 2 $ll] {
		lappend r [expr ($x1+$x2)/2.0] [expr ($y1+$y2)/2.0]
	}
	return [concat $r [lrange $a [expr $ll+1] end]]
    }
    proc flatten {a {steps 3}} {
	
	while {$steps} {
	    set a [smooth $a]
	    incr steps -1
	}
	return [simplify $a]
	#	return $a
    }
    proc vprec args {
	foreach v $args {
	    upvar 1 $v var
	    set var [expr $var+0.0]
	}
	return
    }
    proc point_on_line_seg {x y x1 y1 x2 y2 {delta 1.0}} {
	vprec x y x1 y1 x2 y2 delta
	set dx [expr $x2-$x1]
	set dy [expr $y2-$y1]
	if {[expr abs($dx)]>[expr abs($dy)]} {
	    set sig [expr $x<$x1?-1:1]
	    if {[expr $sig*$x1-$delta <= $sig*$x && \
		    $sig*$x <= $sig*$x2+$delta]} {
		set by [expr ($x-$x1)*($dy/$dx)+$y1]
		return [expr $by-$delta <= $y && \
			$y <= $by+$delta]
	    }
	} elseif {[expr abs($dy)]>0.0001} {
	    set sig [expr $y<$y1?-1:1]
	    if {[expr $sig*$y1-$delta <= $sig*$y && \
		    $sig*$y <= $sig*$y2+$delta]} {
		set bx [expr ($y-$y1)*($dx/$dy)+$x1]
		return [expr $bx-$delta <= $x && \
			$x <= $bx+$delta]
	    }
	} else {
	    return [expr $x1-$delta <= $x &&
	    $x <= $x1+$delta &&
	    $y1-$delta <= $y &&
	    $y <= $y1+$delta]
	}
	return 0
    }
    proc line_length {x y args} {
	set ll [llength $args]
	set len 0.0
	for {set i 0} {$i < $ll} {} {
	    set x1 [lindex $args $i] ; incr i
	    set y1 [lindex $args $i] ; incr i
	    set a [expr $x-$x1]
	    set b [expr $y-$y1]
	    set len [expr $len+sqrt($a*$a + $b*$b)]
	    set x $x1
	    set y $y1
	}
	return $len
    }
    proc point_on_coords {x y coords {delta 1.0}} {
	set x1 [lindex $coords 0]
	set y1 [lindex $coords 1]
	set ll [llength $coords]
	for {set i 2} {$i < $ll} {} {
	    set x2 [lindex $coords $i] ; incr i
	    set y2 [lindex $coords $i] ; incr i
	    if [point_on_line_seg $x $y $x1 $y1 $x2 $y2 $delta] {
		return $i
	    }
	    set x1 $x2 ; set y1 $y2
	}
	return 0
    }
    proc binfind {fu t {p 1.0} {steps 11}} {
	set d $p
	set p [expr $p/2.0]
	for {set i 0} {$i < $steps} {incr i} {
	    set d [expr $d/2.0]
	    set c [eval $fu]
	    set p [expr ($c>$t)?$p-$d/2.0:$p+$d/2.0]
	}
	return $p
    }
    proc search1_ {a smo} {
	upvar 1 p p
	set r [eval line_length \
		[flatten [lrange [subdivide $a $p] 0 7] $smo]]
	return $r
    }
    proc point_on_spline {x y a {smo 4} {delta 2.5}} {
	set coords [flatten $a $smo]
	set n [point_on_coords $x $y $coords $delta]
	if {!$n} {return -1}
	return [binfind [list search1_ $a $smo] \
		[eval line_length [lrange $coords 0 [expr $n-3]] $x $y]]
    }
}


