# A number of utility functions.

# replaces element elt in list l by the elements in rep.
# if rep is not given, erase elt in l.

proc listreplace {l elt {rep {}}} {
    upvar $l list
    set i [lsearch -exact $list $elt]
    if { [string compare $rep {}] == 0} {
	set list [lreplace $list $i $i]	
    } else {
	set list [eval {lreplace $list $i $i} $rep]
    }
}

proc middle {a b} {
    return [expr {($a + $b) / 2.0}]
}

# Recursive drawing proc for a spline.

proc makecurve {x1 y1 x2 y2 x3 y3 x4 y4 prof} {
    if {$prof == 0} {
	return [list $x1 $y1 $x2 $y2 $x3 $y3 $x4 $y4]
    } else {
	set prof [expr $prof - 1]
	set p1x [middle $x1 $x2]
	set p1y [middle $y1  $y2]
	set p2x [middle $x2  $x3]
	set p2y [middle $y2  $y3]
	set p3x [middle $x3  $x4]
	set p3y [middle $y3  $y4]
	set m1x [middle $p1x $p2x]
	set m1y [middle $p1y $p2y]
	set m2x [middle $p3x $p2x]
	set m2y [middle $p3y $p2y]
	set m3x [middle $m1x $m2x]
	set m3y [middle $m1y $m2y]
	set l1 [makecurve $x1 $y1 $p1x $p1y $m1x $m1y $m3x $m3y $prof]
	set l2 [makecurve $m3x $m3y $m2x $m2y $p3x $p3y $x4 $y4 $prof]
	return [concat $l1 $l2]
    }
}

proc norm {p1 p2} {
    set dx [expr [Point::x $p2] - [Point::x $p1]]
    set dy [expr [Point::y $p2] - [Point::y $p1]]
    return [expr {sqrt($dx*$dx+$dy*$dy)}]
}

