# Creates the tksesh database
# Temporary stuff.
# The user should be created before, for his home directory is used
#

package provide dbmaker 1.0
package require sqlite
package require hieroglyphs
package require seshuser
package require BZRFonts
package require TkseshDB
package require stooop
package require seshDico

namespace eval dbmaker {
    variable VAR
    variable FONT
    variable VALUES
}

proc dbmaker::sourceSigns {} {
    variable VAR
    source $VAR(GLYPHSDEFS)
}

proc dbmaker::addValue {code values} {
    variable VALUES
    set VALUES($code) [split $values |]
}

proc dbmaker::addSign {code font num technum} {
    variable CHARS
    variable VAR
    lappend "CHARS($font $technum)" $code
    if {[glyphCode::isGardiner $code]} {
	set "VAR($font $technum,GardinerCode)" $code
    }
}

proc dbmaker::generateSigns {} {
    variable CHARS
    variable VAR
    variable FONT
    variable VALUES
    set lastFont ""

    set db [TkseshDB::getTkseshDB]
    
    foreach c [lsort [array names CHARS]] {
	set font [lindex $c 0]
	set num  [lindex $c 1]
	set drawing [BZRFont::getChar $FONT($font) $num]
	set guides {}
	
	if {$font ne $lastFont} {
	    puts -nonewline "$font "; flush stdout
	    set lastFont $font
	}

	# Generate the char
	set glyph "0, 0, NULL, 1, '1998 6 1 12 0 0', 0, '$drawing', '$guides'"
	# Insert the sign and create its internal code
	TkseshDB::update $db "insert into GlyphView values ($glyph)"
	set glyphid [TkseshDB::select $db "select lastOid from MaxTable where type= 'Glyph'"]

	# Now, deal with the Manuel de codage codes :
	foreach code $CHARS($c) {
	    if {[glyphCode::isGardiner $code]} {
		set hasCode "0, 0, NULL, 1, '1998 6 1 12 0 0', 0, $glyphid, '$code', 1"
		set req "insert into HasCodeView values ($hasCode)"
		TkseshDB::update $db $req
	    } else {
		# This is a phonetic code
		set gCode $VAR($c,GardinerCode)
		TkseshDB::update $db "insert into PhonCode values ('$code', '$gCode')"
	    }
	}
    }
    
    foreach {code values} [array get VALUES] {
	set date {'1998 6 1 12 0 0'}
	foreach val $values {
	    if {$val eq "*"} {continue}
	    #puts "inserting $val for $code "
	    puts -nonewline "." ; flush stdout
	    set sql "insert into HasValueView\n"
	    append sql "(uid, mid, id, repOid, date, sid, code, value, " 
	    append sql "valueType, usualValue, usualSign, period)\n"
	    append sql "values (0,0,NULL,NULL,$date,0,'$code','$val', NULL, NULL, NULL, NULL)"
	    TkseshDB::update $db $sql
	}
    }
}

proc dbmaker::loadTmpDico {} {
    variable VAR
    set db [TkseshDB::getTkseshDB]
    set f [open $VAR(DICOFILE)]
    set txt [read $f]
    close $f
    TkseshDB::begin $db
    TagHandler::declareAllTags 
    DicoWF::loadOldDicoFile 0 0 $txt 
    TkseshDB::commit $db 
}


proc dbmaker::makeVariants {} {

    # Variant definition to input into the database.
    # Note that there is currently no system 
    # dbmaker::define_variant A30 A4 : A30 is a variant of A4
    dbmaker::defineVariantSign A30 A4
    dbmaker::defineVariantSign A14 A13
    dbmaker::defineVariantSign A14A A13
    dbmaker::defineVariantSign A20 A19
    dbmaker::defineVariantSign A25 A24
    dbmaker::defineVariantSign A37 A36
    dbmaker::defineVariantSign A39 A38
    dbmaker::defineVariantSign A42 A41
    dbmaker::defineVariantSign A44 A43
    dbmaker::defineVariantSign A46 A45
    dbmaker::defineVariantSign A48 A47
    dbmaker::defineVariantSign A51 A51
    dbmaker::defineVariantSign A52 A51
    dbmaker::defineVariantSign C2 C1
    dbmaker::defineVariantSign C26 C3
    dbmaker::defineVariantSign C26A C3
    dbmaker::defineVariantSign C5 C4
    dbmaker::defineVariantSign E15 C6
    dbmaker::defineVariantSign E16 C6
    dbmaker::defineVariantSign E20 C7
    dbmaker::defineVariantSign E21 C7
    dbmaker::defineVariantSign C20 C19
    
    dbmaker::defineVariantSign D6 D5
    dbmaker::defineVariantSign D20 D19
    dbmaker::defineVariantSign D27A D27
    dbmaker::defineVariantSign D29 D28
    dbmaker::defineVariantSign D34A D24
    dbmaker::defineVariantSign D62 D61
    
    dbmaker::defineVariantSign E8A E8
    dbmaker::defineVariantSign E11 E10
    dbmaker::defineVariantSign E19 E18
    
    dbmaker::defineVariantSign F6 F5
    dbmaker::defineVariantSign F8 F7
    dbmaker::defineVariantSign F11 F10
    dbmaker::defineVariantSign F24 F23
    dbmaker::defineVariantSign F37B F37
    dbmaker::defineVariantSign F49 F46
    dbmaker::defineVariantSign F48 F46
    dbmaker::defineVariantSign F47 F46
    dbmaker::defineVariantSign G7AA G7A
    dbmaker::defineVariantSign G12 G11
    dbmaker::defineVariantSign G20 G19
    dbmaker::defineVariantSign G24 G23
    dbmaker::defineVariantSign G26A G26
    dbmaker::defineVariantSign G41 G40
    dbmaker::defineVariantSign G49 G48
    
    dbmaker::defineVariantSign H6A H6
    
    dbmaker::defineVariantSign I5A I5
    dbmaker::defineVariantSign I15 IA4
    
    dbmaker::defineVariantSign M6 M5
    dbmaker::defineVariantSign M14 M13
    dbmaker::defineVariantSign M25 M24
    dbmaker::defineVariantSign M27 M26
    dbmaker::defineVariantSign M32 M31
    dbmaker::defineVariantSign M37 M36
    
    dbmaker::defineVariantSign N3 N2
    dbmaker::defineVariantSign N6 N5
    dbmaker::defineVariantSign N10 N9
    dbmaker::defineVariantSign N17 N16
    dbmaker::defineVariantSign N21 N20
    dbmaker::defineVariantSign N39 N37
    dbmaker::defineVariantSign N38 N37
    dbmaker::defineVariantSign N42 N41
    
    dbmaker::defineVariantSign O7 O6
    dbmaker::defineVariantSign O12 O11
    dbmaker::defineVariantSign O14 O13
    dbmaker::defineVariantSign O7 O6
    dbmaker::defineVariantSign O17 O16
    dbmaker::defineVariantSign O29v O29
    dbmaker::defineVariantSign O43 O42
    dbmaker::defineVariantSign O46 O45
    dbmaker::defineVariantSign O48 O47
    
    dbmaker::defineVariantSign P7 P6
    
    dbmaker::defineVariantSign R2 R1
    dbmaker::defineVariantSign R6 R5
    dbmaker::defineVariantSign R18 R17
    dbmaker::defineVariantSign R21 R20
    dbmaker::defineVariantSign R23 R22
    dbmaker::defineVariantSign R25 R24
    
    dbmaker::defineVariantSign S6 S5
    dbmaker::defineVariantSign S16 S15
    dbmaker::defineVariantSign S17 S15
    dbmaker::defineVariantSign S20 S19
    
    dbmaker::defineVariantSign T4 T3
    dbmaker::defineVariantSign T5 T3
    dbmaker::defineVariantSign T8A T8
    dbmaker::defineVariantSign T9A T9
    dbmaker::defineVariantSign T10 T9
    dbmaker::defineVariantSign T15 T14
    dbmaker::defineVariantSign T20 T19
    dbmaker::defineVariantSign T23 T22
    dbmaker::defineVariantSign T27 T26
    dbmaker::defineVariantSign T32 T31
    dbmaker::defineVariantSign T33 T31
    dbmaker::defineVariantSign T35 T34
    
    dbmaker::defineVariantSign U2 U1
    dbmaker::defineVariantSign U3 U1
    dbmaker::defineVariantSign U5 U4
    dbmaker::defineVariantSign U7 U6
    dbmaker::defineVariantSign U12 U11
    dbmaker::defineVariantSign U14 U13
    dbmaker::defineVariantSign U18 U17
    dbmaker::defineVariantSign U20 U19
    dbmaker::defineVariantSign U25 U24
    dbmaker::defineVariantSign U27 U26
    dbmaker::defineVariantSign U29 U28
    dbmaker::defineVariantSign U35 U34
    dbmaker::defineVariantSign U40 U39

    dbmaker::defineVariantSign V8 V7
    dbmaker::defineVariantSign V10 V9
    dbmaker::defineVariantSign V14 V13
    dbmaker::defineVariantSign V18 V17
    dbmaker::defineVariantSign V23 V22
    dbmaker::defineVariantSign V25 V24
    dbmaker::defineVariantSign V27 V26
    dbmaker::defineVariantSign V31A V31
    dbmaker::defineVariantSign V34 V33
    
    dbmaker::defineVariantSign W4 W3
    dbmaker::defineVariantSign W8 W7
    dbmaker::defineVariantSign W10 W9
    dbmaker::defineVariantSign W12 W11
    dbmaker::defineVariantSign W16 W15
    dbmaker::defineVariantSign W18 W17
    
    dbmaker::defineVariantSign X3 W2
    dbmaker::defineVariantSign Y2 Y1
    dbmaker::defineVariantSign Y1v Y1
    dbmaker::defineVariantSign Y4 Y3
    
    dbmaker::defineVariantSign Z3 Z2
    dbmaker::defineVariantSign Z3A Z2
    dbmaker::defineVariantSign Z6A Z6
    dbmaker::defineVariantSign Z7 G43
    dbmaker::defineVariantSign Z10 Z9
    dbmaker::defineVariantSign Z12 A24
    dbmaker::defineVariantSign D40 A24
    dbmaker::defineVariantSign Aa12 Aa11
    dbmaker::defineVariantSign A14 Aa13
    dbmaker::defineVariantSign Aa15 Aa13
    dbmaker::defineVariantSign Aa18 Aa17
    dbmaker::defineVariantSign Aa22 Aa21
    dbmaker::defineVariantSign Aa24 Aa23
    dbmaker::defineVariantSign Aa29 Aa28
    dbmaker::defineVariantSign Aa31 Aa30
    
    dbmaker::defineVariantSign B4 B3
    dbmaker::defineVariantSign C2 C1
    dbmaker::defineVariantSign G41 G40
    dbmaker::defineVariantSign Y1v Y1
    dbmaker::defineVariantSign Z7 G43
    
    # We define sets of signs which may be use with the same meaning.
    dbmaker::defineVariantSign X8 D37
    dbmaker::defineVariantSign D38 D37 3
    dbmaker::defineVariantSign a A24 2
    dbmaker::defineVariantSign a D37 2
    dbmaker::defineVariantSign d D37 2
    
    dbmaker::defineVariantSign G19 D38
    dbmaker::defineVariantSign D37 D38 2
    dbmaker::defineVariantSign a D38 1

    dbmaker::defineVariantSign N33A Z2 3
}

proc dbmaker::makeDB {} {
    variable VAR
    variable FONT
    global env
    # Load user data ???
    seshuser::superUser

    # define variables
    set VAR(DIR) [file join $::env(SESHLIB) seshDB]
    set VAR(DBFILE) [TkseshDB::getDBFilePath]
    set VAR(SQLSRC) [file join $VAR(DIR) "tables.sql"]
    set VAR(GLYPHSDEFS) [file join $::env(SESHLIB) PrepareDB buildSQLSigns.tcl]
    set VAR(DICOFILE) [file join $::env(SESHLIB) "olddico.dic"]
    set VAR(id) 0

    # Build all fonts
    foreach font {A B C D E F G H I K L M N O P Q R S T U V W X Y Z Aa} {
	set FONT($font) [new BZRFont]
	BZRFont::load $FONT($font) [file join $::env(SESHLIB) goodbzr $font.bzr]
	
    }

    
    file delete $VAR(DBFILE)

    set db [TkseshDB::getTkseshDB]
    
    # Create the actual database
    set f [open $VAR(SQLSRC)]
    set txt [read $f]
    close $f
    TkseshDB::begin $db
    # Now, let's first go into SuperUser mode
    # Not a good idea, as actual data may change
    # Some kind of a dump would be better here.

    TkseshDB::update $db $txt

    set txt {
	insert into MaxTable
	select classname, 0, 0
	from ClassName;
    }

    TkseshDB::update $db $txt
    # Ok. Now, load the data :

    dbmaker::makeVariants
    
    source [file join $::env(SESHLIB) PrepareDB buildSQLLigatures.tcl]
    dbmaker::sourceSigns 

    dbmaker::generateSigns
    
    set txt {
	insert into Management (uid, mid, maxsid) values (0,0,1);
	update MaxTable set maxid= 0, lastOid= 0;
    }
    
    TkseshDB::update $db $txt
    TkseshDB::commit $db 

    TkseshDB::close $db
}

proc dbmaker::installDB {} {
    file copy [file join $::env(SESHLIB) seshDB.db] [TkseshDB::getDBFilePath]
}

proc dbmaker::defineVariantSign {s1 s2 {confidence 5}} {
    set db [TkseshDB::getTkseshDB]
    set s1 [TkseshDB::protectText $s1]
    set s2 [TkseshDB::protectText $s2]
    set sql {
	insert into SignRelationView 
	(uid, mid, id, date, sid, code1, code2, reltype, confidence)
	values
	(0,0,NULL,DATE,0,CODE1,CODE2,'v',CONFIDENCE)}
    set sql [string map [list DATE "'[TkseshDB::getDate]'" CODE1 '$s1' CODE2 '$s2' \
			     CONFIDENCE $confidence] $sql]
    TkseshDB::update $db $sql
}
