# seshtext.tcl --
#
#        finding and saving hieroglyphic text file 
#        from readable references.
#        SUBJECT TO CHANGE.
#
#  1999, 2003 Serge Rosmorduc,
# Universit de Paris 8
#
# Tout droits rservs.
# 
# voyez le fichier ../LICENCE.txt pour plus d'informations
# See the ../LICENCEE.txt file for more information
#
# $Id: seshtext.tcl,v 1.1.1.1 2002/04/16 13:55:01 rosmord Exp $
#
#

#
# $Log: seshtext.tcl,v $
# Revision 1.1.1.1  2002/04/16 13:55:01  rosmord
# Imported using tkCVS
#
# Revision 1.3  2000/10/31 17:19:29  rosmord
# Mac Adaptation
#
# Revision 1.2  2000/08/16 19:59:00  rosmord
# Changed from a directory based system to a database based system
#
#

# FUNCTION SUMMARY --

# known public functions are : 
# save --
# getMDCText --
# deleteText --
# directory -- get the list of available texts.
# 
# 	
# DATA SUMMARY -- 

package provide seshtext 1.0
package require seshuser
package require textDAL
package require seshutils
package require tag

namespace eval seshtext {
    variable VARS

    # seshtext::save --
    # 	saves a text into the textual database
    # Parameters:
    #	name: name of the text
    #   txt: MDC representation of the text
    #   listOfSigns : the list of signs in the text.
    #		currently, this is computed by the tksesh widget,
    #		but in the future, this argument should be suppressed 
    #		and the list should be computed here.
    # Options:
    # Results:
    #	returns the object identifier associated with the text
    #	as a list : {uid mid id type} (user id, machine id, entry id, type)
    #	type is always text.
    # Side Effects:
    #	the text is saved in the database
    # Use:
    #	PUBLIC
    # Called by:

    proc save {name txt listOfSigns} {
	set id [textDAL::saveText $name $txt $listOfSigns]
	return $id
    }
    
    # seshtext::getMDCText --
    # 	loads a text from the textual database
    # Parameters:
    #	name: name of the text
    #	ext: optional extension. hie is for the text itself, lst for
    #        the list of the text's hieroglyphs
    # Options:
    # Results:
    #	the loaded text
    # Side Effects:
    # Use:
    #	PUBLIC
    # Called by:
    
    proc getMDCText {name} {
	set txt [textDAL::getMDCTextByName $name]
	return $txt
    }
    
    # Return the first text available for a given name.
    # When the dictionary system, is updated text will be identified by oid,

    proc getOidForName {name} {
	set oid [textDAL::getOidsForName $name]
	return [lindex $oid 0]
    }

    proc getMDCTextFromListIdx {idx} {
	set oid [textDAL::getDSOid $idx]
	return [textDAL::getTextFromOid $oid]
    }

    
    # Doesn't really belong to this level. We should probably 
    # have some highter level representation for texts.  
    # Objects, for instance (a proxy would be very nice. It would
    # contain at first only the oid (or the complete identification),
    # and lazily load needed stuff.
    
    proc getMDCTextFromOid {oid} {
	return [textDAL::getTextFromOid $oid]
    }

    # seshtext::deleteText --
    # 	deletes a text from the textual database
    # Parameters:
    #	name: name of the text
    # Options:
    # Results:
    #	none
    # Side Effects:
    #	the named text is destroyed from the user base, along with the corresponding
    #	extensions.
    # Use:
    #	PUBLIC
    # Called by:

    proc deleteText {name} {
	textDAL::deleteText $name
    }


    # seshtext::directory --
    # 	deletes a text from the textual database
    # Parameters:
    #	ext: kind of texts to look for (hie, lst)
    # Options:
    # Results: the list of available texts.
    #	the
    # Side Effects:
    # Use:
    #	PUBLIC
    # Called by:
    
    # List all texts in a database.
    proc directory {{ext hie}} {
	return [textDAL::getAllTextNames]
    }

    proc getTextNamesDataSource {} {
	return [textDAL::getTextNamesDataSource]
    }

    # Search for a regexp in all texts. 
    # For each occurrence found, call "command" 
    # with arguments :
    # 
    proc findInAllTexts {reg type command} {
	variable VARS
	set VARS(command) $command
	set VARS(regexp) $reg
	textDAL::scanTextsContent $type seshtext::findInAllTextsAux
    }

    proc findInAllTextsAux {oid name content} {
	variable VARS
	puts HERE

	set reg $VARS(regexp)
	# find the data in "body". 
	# for each occurrence, calls command.

	# Suppress the first line
	regsub "^\[^\n\]*\n" $content "" content
	set txt "\n"
	append txt $content
	append txt "\n"
	# Search.  reg is supposed to suppress any line not containing the text.
	# If the text spans more than one line, the only ref. found should be to the start
	regsub -all $reg $txt {\1} txt
	# Now, process the result, reference by reference
	set txt [split $txt "\n\t"]
	# We'll got an extra empty element, due to the terminal \t of the regexp.
	# We suppress it manually ?????? (VERIFY THIS. LOOKS OBSOLETE)
	foreach ref $txt {
	    if {[string length $ref]} {
		#puts "[list $oid $name $ref]"
		eval $VARS(command) [list $oid $name $ref]
	    }
	}
    }
}
