# search --
#
#       
#       full-text search in the text database
#
# 
#
#
# 
#  1998, 2000 Serge Rosmorduc,
# quipe EC.art, Laboratoire LRIA, Universit de Paris 8
#
# 15 rue Catulienne
# 93236 St Denis Cedex
# FRANCE
#
# Tout droits rservs.
# 
# voyez le fichier ../LICENCE.txt pour plus d'informations
# See the ../LICENCEE.txt file for more information
#
# $Id: search.tcl,v 1.1.1.1 2002/04/16 13:55:01 rosmord Exp $
#
#

#
# $Log: search.tcl,v $
# Revision 1.1.1.1  2002/04/16 13:55:01  rosmord
# Imported using tkCVS
#
# Revision 1.5  2000/10/31 17:19:28  rosmord
# Mac Adaptation
#
#

package require seshw
package require legendes
package require yellow
package require seshutils
package require seshtext

package provide seshsearch 0.0	;# This uses Tcl namespaces

namespace eval seshsearch {
    namespace export create

    variable WINDOWS 0
    variable LOC
    variable TAGNUM 0
    
    array set LOC ""

    # seshsearch::create --
    #	creates a search window
    # Parameters:
    #	w	: window name
    # Options:
    # Results:
    # Side Effects:
    #	window is created
    # Use:
    #	PUBLIC
    # Called by:

    proc create {w} {
	variable LOC
	frame $w
	set LOC($w,trans) ""
	set LOC($w,trad) ""
	label $w.ltrans -text [legendes::trans 	    "Transliteration"]
	entry $w.trans -textvariable seshsearch::LOC($w,trans) -font {mdctranslitlc 18}
	label $w.ltrad -text [legendes::trans "Translation"]
	entry $w.trad -textvariable seshsearch::LOC($w,trad)
	label $w.lhie -text [legendes::trans "Hieroglyphs"]
	seshw::create_editor $w.hie -notitle -show_words 1
	yellow::create $w.lhie [legendes::trans "You can type here the hieroglyphic text you look for.\nTksesh will not differentiate signs like Y1 and Y1v.\n If you type / after a sign, it will become blue\n and be considered optional in the search."]
	set seshsearch::LOC($w,exact) 1
	checkbutton $w.exact -variable seshsearch::LOC($w,exact) -state disabled
	
	frame $w.f1
	button $w.clear -text [legendes::trans "Clear"] -command [list seshsearch::clear $w]
	button $w.go -text [legendes::trans "Find"] -command [list seshsearch::create_windows $w]
	button $w.new -text [legendes::trans "New"] -command [list seshsearch::create_windows $w 1]
	
	label $w.lexact -text [legendes::trans "Exact search"]

	frame $w.f2
	set w1 $w.f2
	text $w1.results -yscrollcommand "$w1.sb set"
	scrollbar $w1.sb -command "$w1.results yview" -orient vertical

	grid $w.ltrans - $w.trans -              -sticky ew 
	grid $w.lhie - $w.hie     -               -sticky ew 
	grid $w.ltrad - $w.trad   -               -sticky ew 
	grid $w.lexact $w.exact - -                -sticky e
	grid $w.f2 - - - -sticky news
	grid $w.f1
	grid $w1.results $w1.sb -sticky news
	
	grid rowconfigure $w1 0 -weight 1
	grid columnconfigure $w1 0 -weight 1


	pack $w.clear $w.go $w.new -side left -in $w.f1
	
	grid rowconfigure $w 4 -weight 1
	grid columnconfigure $w 1 -weight 1


	$w1.results tag configure blue -background lightblue
	bind $w1.results <Key> {break}
	bind $w1.results <Key-Down> {%W mark set insert {insert + 1 lines};break}
	bind $w1.results <Key-Up> {%W mark set insert {insert - 1 lines};break}
	bind $w1.results <Return> {seshsearch::showNext %W; break}
    }


    # seshsearch::clear --
    #	resets parameters for window
    # Parameters:
    #	w	: window name
    # Options:
    # Results:
    # Side Effects:
    # Use:
    #	PUBLIC
    # Called by:

    proc clear {w} {
	variable LOC
	set  LOC($w,trad) ""
	set LOC($w,trans) ""
	::seshw::clear_text $w.hie.s
	set seshsearch::LOC($w,exact) 1
    }


    # seshsearch::create_windows --
    #	Badly named and written function. clears the result window, and
    #	if new=0, search for text. SHOULD BE CHANGED ASAP
    # Parameters:
    #	w	: window name
    # Options:
    # Results:
    # Side Effects:
    # Use:
    #	PUBLIC
    # Called by:

    proc create_windows {w {new 0}} {
	variable WINDOWS
	variable LOC
	
	set idlist {}
	
	set w1 $w.f2
	$w1.results delete 0.0 end
	
	if {! $new} {
	    if { $LOC($w,exact) } {

		if { $LOC($w,trans) != "" } {
		    find_text $w1.results $LOC($w,trans)
		}
		
		if { $LOC($w,trad) != "" } {
		    find_text $w1.results $LOC($w,trad)
		}

		if { [$w.hie.s get 0.0 end ] != "" } {
		    find_hiero $w1.results [$w.hie.s liste_signes -word 0.0 end] -synsets
		}
	    } else {
		error 
	    }
	}
    }


    # seshsearch::freetags --
    #	delete tags in the result window
    # Parameters:
    #	w	: *result* window name
    # Options:
    # Results:
    # Side Effects:
    # Use:
    #	PRIVATE
    # Called by:
    #	find_text find_hiero
    
    proc freetags {w} {
	set l {$w tag}
	foreach t $l {
	    $w tag delete $t
	}
    }

    # seshsearch::inserttag --
    #	insert tags in the result window
    # Parameters:
    #	w	: result window name
    #	name	: text name
    #	oid	: text oid 
    #	ref	: reference in text
    # Options:
    # Results:
    # Side Effects:
    #	a t
    # Use:
    #	PRIVATE
    # Called by:
    #	find_text find_hiero


    proc inserttag {w name oid ref} {
	variable TAGNUM
	incr TAGNUM
	set tagname "t$TAGNUM"
	$w insert insert "$name, $ref" $tagname 

	set bg [$w cget -background]
	$w tag bind $tagname <Double-Button-1> [list seshw::displayText $oid $ref]
	$w tag bind $tagname <Enter> [list $w tag configure $tagname -background lightblue]
	$w tag bind $tagname <Leave> [list $w tag configure $tagname -background $bg]
    }

    proc showNext {win} {
	set tags [$win tag names "insert linestart"]
	foreach t $tags {
	    if {[string match {t[0-9]*} $t]} {
		eval [$win tag bind $t <Double-Button-1>]
	    } 
	}
	$win mark set insert {insert + 1 lines}
    }
    # Utilitary function : take a path, a pattern, and 
    # returns a string suitable for use with glob

    proc make_glob {path glob} {
	global tcl_platform
	set head ""
	set body $path
	switch -exact -- $tcl_platform(platform) {
	windows {
	      regexp {^([a-zA-Z]:)(.*)$} $path dumm head body
		}
      default {}
	}
	set pat $body
	regsub -all {(.)} $pat {\\\1} pat
	return "$head/$body/$glob"
	#regsub -all {\\\*\\\.\\t\\r\\l$} $pat1 {*.trl} pat1
    }

    # PROBABLY DOESN'T WORK. LOOK AT FIND_HIERO, WHICH WORKS NOW.
    proc find_text {w mot {type trl}} {
	# Build regular expression for search
	# the idea is that we either match a line with the word or a line without it
	# (Beware, changes in latest TCL versions)
	set reg [format {(\n[^	]*	[^\n]*%s[^\n]*)|(\n[^\n]*)} $mot]
	# replace all verbatim \n by real \n (automatically done with the new regexp)
 	#regsub -all {\\n} $reg "\n" reg
	
	find_aux $w $reg $type
    }


    proc find_hiero {w mot args} {

	# test to know if we should consider synsets
	set use_syn 0
	foreach a $args {
	    switch -- $a {
		-synsets {set use_syn 1}
	    }
	}

	# Build regular expression
	set mot1 ""
	foreach s $mot {
	    # Is this sign optional ?
	    if {[llength $s] == 2} {
		lappend mot1 "?"
		set s [lindex $s 0]
	    }
	    set s [hieroglyphs::low_level_norm $s]
	    set equiv [hieroglyphs::getAllVariantsOf $s]
	    if $use_syn {
		set equiv1 ""
		foreach e $equiv {
		    foreach v [hieroglyphs::get_all_syns $e] {
			lappend equiv1 $v
		    }
		}
		lappend mot1 [uniq $equiv1]
	    } else {
		lappend mot1 $equiv
	    }
	}


	set txtreg [hieroglyphs::list_to_regexp $mot1]
	
	# should Replace a line containing the start of the text
	# with the content of the first parenthesis (the line number).

	set reg [format {([^\t\n]+\t)[^\n]*%s [^\n]*|[^\n]*} $txtreg]
	#                ^             ^   ^      ^     ^
	#                |_line number |   |      |     |
	#                              |_ chars before the text proper
	#                                  |_ the text with a final space
	#                                     (avoids confusing A40 with A4)
	#                                         |_ line end
	#                                                |_ non matching line
	find_aux $w $reg lst
    }

    # seshsearch::find_aux --
    #	auxiliary function doing the actual search
    # Parameters:
    #	w	: *result* window name
    #	reg	: the regular expression to use
    #	type	: one of lst (sign list), trl (translit), txt (translation).
    # Options:
    # Results:
    # Side Effects:
    #	the window w is filled with references found
    # Use:
    #	PRIVATE
    # Called by:
    #	find_text find_hiero

    proc find_aux {w reg type} {
	# clear
	freetags $w
	# get list of texts
	seshtext::findInAllTexts $reg $type [list seshsearch::PRIVinsertRef $w]
    }

    
    # Callback function to pass to the search pattern.
    proc PRIVinsertRef {w oid name ref} {
	inserttag $w $name $oid $ref
	$w insert insert "\n"
    }
}


