# seshw.tcl --
#
#        basic bindings for the sesh editing widgets.
#
# Copyright (c) 1998 Serge Rosmorduc
#
# 2003/05/05 : modifications, mostly to the sign selection system.
#	we suppressed a number of dirty tricks in order to do "the right thing"
#	


package provide seshw 0.0

package require hieroglyphs
package require legendes
package require Tksesh
package require yellow
#package require signesdef
package require seshDico
package require glossary

# This one should probably move
package require seshtext
package require stooop
package require tkseshUtils

namespace eval seshw {
    variable SESHW
    # WE use PREFERREDSIGNPOS to keep track of
    # the sign which was chosen for a given code.

    variable PREFERREDSIGNPOS 
    variable CUTBUFFER
    variable CURRENTREF
    
    array set PREFERREDSIGNPOS {}
    set CUTBUFFER ""
    
    #
    # CONVENTION : everything there is built around a sesh widget.
    # In all commands, w will be the general window around the 
    # widget, and win will be the widget itself.
    #
    #
    # Format for SESHW
    # (W is the base window -- you can retrieve it with from 
    # a sesh window by using the base_window  procedure)
    #
    # $W,pos : position of the cursor in the text. 
    #       textual coordinates, not computer ones
    #
    # $W,name : name of the edited text
    #
    # $W,buffer : buffer for the current window
    # $W,code : code entered for sign being added
    # $W,possibleSigns : possible signs for current code
    # $W,dictionaryExpanded : true if the dictionary was searched for the current code
    # $W,separator : pending separator
    # $W,modifier : pending modifier for current sign (size, revert, etc.)
    # $W,saved : cadrat being modified
    # $W,retro : retrograde writing mode ?
    # $W,window : the sesh widget
    # $W,scribe : name of the person who typed the text.
    # $W,comment : free text comment. Normaly used to indicate 
    #   text origin (e.g. "collated from the original")
    #
    # $W,sign,$CODE : user-defined sign corresponding to "code"
    #             At the time being, the stuff is global to all widgets. 
    #             Should be local. the value is a list : code repr,
    #             where repr is a list based representation of the sign.
    #
    # for the coordinates :
    # SESHW($W,coords,...) and
    #   format : description of the format
    #   1, 2... : description of the ith coordinate
    #   value,NUM : value of NUMth coordinate
    #
    #
    #
    # For more documentation, see function clear_text 
    
    namespace export create_viewer;	# create a window to view a text
    namespace export create_editor;	# create a window to edit a text
    namespace export create_passive;	# small window for inclusion in other windows
    namespace export set_text;		# set the text at some value
    namespace export clear_text;	# clears a text
    namespace export fromtext;		# load a text
    namespace export totext;		# output a textual description of the window's contents
    namespace export getname;		# get the text's name
    namespace export loadText;		# load a text from the database
    namespace export savetext;		# saves a text to the database
    namespace export gethwindow;	# get the real sesh widget associated with current widget
    namespace export hachure;		# put hachures around selected text
    namespace export cartouche;		# put a cartouche around selected text
    namespace export crochets;		# put [] around selected text
    namespace export hachure_cadrat;	# hatchure the current cadrat 
    namespace export set_scribe get_scribe
    namespace export set_comment get_comment
    namespace export set_scribe_code get_scribe_code
    namespace export define_local_sign;	# defines a sign local to the current text
    namespace export edit_coords edit_coords_system
    namespace export sesh_depl;		#  moves the insertion point
    namespace export insert_text;		#  moves the insertion point
    namespace export copier couper paste_selection; #  copy/cut/paste
    namespace export get_selection;	# get the selection's value
    namespace export bind_onmove;	# define a function which will be called
					# each time the insertion point moves in the widget; 
    namespace export winlatininsert winnumlineinsert
    namespace export explode_cadrat
    #
    #if [info exists env(SESHLIB)]
    
    
    ################################################################
    # Routine for global font loading. 
    #
    #

    ################################################################
    # Creation/deletion routines.
    #
    #
    
    ################################################################
    # The create function : creates a sesh window fully equiped with 
    # scrollbars and the like.
    # The sesh widget itself is $w.s
    proc create_editor {w args} {
	eval create_viewer $w $args
	bindtags $w.s [list $w EditorSesh ViewSesh all]	
	label $w.buf -textvariable seshw::SESHW($w,buffer)
	grid $w.buf -  	
	#DicoDAL::setOrthographyHook [list seshw::orthoRefresh $w]

	return $w
    }
    
    
    proc create_viewer {w args} {
	variable SESHW
	
	frame $w

	# We should use a more generic system. But this works now :
	set title 1
	if {[string compare [lindex $args 0] "-notitle"] == 0} {
	    set title 0
	    set args [lrange $args 1 end]
	}

	scrollbar $w.x -command "$w.s xview" -orient horizontal
	scrollbar $w.y -command "$w.s yview" -orient vertical

	if {$title != 0} {
	    set f1 [frame $w.f1]	    
	    entry $f1.name -textvariable seshw::SESHW($w,name) \
		    -width 40 -relief flat -justify center
	    yellow::create $f1.name [legendes::trans "Name of the text in the database\n\"nn rn=f\" stands for \"no name\""]
	    
	    label $f1.pos -textvariable seshw::SESHW($w,pos)
	    yellow::create $f1.pos [legendes::trans "Text coordinate of the cursor"]
	}

	eval sesh $w.s \
	     	[list -yscrollcommand "$w.y set" \
		     -xscrollcommand "$w.x set" ]\
	    $args
	
	set SESHW($w,window) $w.s

	# grid configuration 
	if {$title != 0} {
	    grid $f1.name 
	    grid $f1.pos 
	    grid $f1 - -sticky ew
	}
	grid $w.s $w.y -sticky news
	grid $w.x -sticky ew
	if {$title != 0} {
	    grid rowconfigure $w 1 -weight 1
	    grid columnconfigure $w 0 -weight 1
	} else {
	    grid rowconfigure $w 0 -weight 1
	    grid columnconfigure $w 0 -weight 1
	}
	
	bindtags $w.s [list $w.s ViewSesh all]	
	clear_text $w.s
	
	selection handle $w.s seshw::seshgivesel
	
	return $w
    }
    
    # Passive hieroglyphic windows for inclusion in text windows and 
    # the like
    
    proc create_passive {w args} {
	variable SESHW
	set command [concat sesh $w -draw_cursor 0 $args]
	set result [eval $command]
	set SESHW($w,window) $w
	bindtags $result [list $w PassiveSesh all]
	return $w
    }
    
    
    ################################################################
    # For passive widgets.
    #
    
    # autosize gives to w the size that just fits the hieroglyphs it contains
    proc autosize {win} {
	set dimx [$win dimx]
	set dimy [$win dimy]
	if {$dimx  > 3 && $dimy > 0} {
	    $win configure -width $dimx -height $dimy
	}	
    }
    
    proc set_text {win txt} {
	$win delete 0.0 end
	$win insert insert $txt
	$win mark set insert 0.0
    }

    # Selection is saved as a list :
    # first element is content type
    # second element is data.
    # if content type is not recognized, then 
    # the whole list is supposed to be data
    # (for compatibility and ease of use)
    # MDC : manuel de codage code
    # SGN : new character definition. 1st elt : code. 2nd : val
    
    proc paste_selection {win} {
	set sel [selection get -selection $::env(DEFAULTSELECTION)]
	set w [base_window $win]
	if [llength $sel] {
	    switch -- [lindex $sel 0] {
		MDC {insert_text $win [lindex $sel 1]}
		SGN {define_local_sign $w [lindex $sel 1] [lindex $sel 2]}
		default {insert_text $win $sel}
	    }

	}
    }
    
    
    ################################################################
    # low level tools
    
    proc base_window {win} {
	set w $win
	regexp {(.*)\.s} $win tmp w
	return $w
    }

    proc gethwindow {w} {
	variable SESHW
	return $SESHW($w,window)
    }

    ################################################################
    # Utilities for active widgets.
    # clear and initialize

    proc clear_text {win} {
	variable SESHW
	set w [base_window $win]
	
	$win delete 0.0 end
	set SESHW($w,name) "nn rn=f"
	set SESHW($w,scribe) "nn rn=f"
	set SESHW($w,comment) ""
	set SESHW($w,scribe_code) "1234567890"
	set SESHW($w,coords,format) {$2, $1}
	set SESHW($w,buffer) ""
	set SESHW($w,possibleSigns) "" ; # Possibles values for the current code
	set SESHW($w,dictionaryExpanded) 0 ;
	set SESHW($w,separator) ""
	set SESHW($w,modifier) ""
        # selection : state of the widget as regard mouse selection
	set SESHW($w,selection) 0
        # mouse_selection : state of the widget as regard "shift" selection
	set SESHW($w,shift_selection) 0
	set SESHW($w,selectionstart) "0.0"
	set SESHW($w,shift_selectionstart) "0.0"
	set SESHW($w,saved) ""
	set SESHW($w,code) ""
	set SESHW($w,retro) 0	; # if true, we are in retrograde writing mode
	set SESHW($w,in_columns) 0
	set SESHW($w,pos) "1, 1"
	clear_local_signs $w
	$win coords format {$2, $1}
	foreach i { 1 2 3 4 5} {
	    set SESHW($w,coords,$i) NUM
	    $win coords type $i NUM
	}
    }

    # erase definitions of signs peculiar to this window
    proc clear_local_signs {w} {
	variable SESHW
	foreach def [array names SESHW $w,sign,*] {
	    unset SESHW($def)
	}
    }
    
    ################################################################
    # Selection

    # Mouse selection.

    proc start_selection {win {pos insert}} {
	variable SESHW
	set w [base_window $win]
	clear_select $win
	set SESHW($w,selectionstart) [eval $win index $pos]
    }
    
    proc end_selection {win pos} {
	variable SESHW	
	set w [base_window $win]
	clear_select $win
	set SESHW($w,selection) 1
	# $win undocumented1 insert endsel
	eval $win mark set endsel $pos
	$win undocumented0 insert endsel
    }
    
    proc sesh_arrow_sel {win pos} {
	variable SESHW
	set w [base_window $win]
	if {! $SESHW($w,shift_selection)} {
	    clear_select $win
	    set SESHW($w,shift_selectionstart) [$win index insert]
	    set SESHW($w,shift_selection) 1
	}
	$win undocumented1 $SESHW($w,shift_selectionstart) endsel
	set posi [eval $win index $pos]
	seshw::set_insert_at $win $posi
	eval $win mark set endsel insert
	$win undocumented0  $SESHW($w,shift_selectionstart) endsel
    }


    proc clear_select {win} {
	variable SESHW	
	set w [base_window $win]
	if {$SESHW($w,selection)} {
	    $w.s undocumented1 0.0 end
	    set SESHW($w,selection) 0
	}
	if {$SESHW($w,shift_selection)} {
	    $w.s undocumented1 0.0 end
	    set SESHW($w,shift_selection) 0
	}
    }
    
    proc  get_start_selection {win} {
	variable SESHW
	set w [base_window $win]
	if {$SESHW($w,selection)} {
	    set debut $SESHW($w,selectionstart)
	} else {
	    if {$SESHW($w,shift_selection)} {
		set debut $SESHW($w,shift_selectionstart)
	    } else {
		error "No selection made"
	    }
	}
	return $debut
    }
    
    proc get_selection {w args} {
	set aslist 0
	foreach arg $args {
	    switch -- $arg {
		-aslist {set aslist 1} 
		default {error "unknown arg $arg"}
	    }
	}
	set win [seshw::gethwindow $w]
	if $aslist {
	    return [$win liste_signes [get_start_selection $win] endsel]
	} else {
	    return [$win get [get_start_selection $win] endsel]
	}
    }

    ################################################################
    # Getting information stuff
    
    proc getref {w {pos insert}} {
	variable SESHW
	set ligne [$w.s coords get insert]
	#regsub {\..*} $ligne {} ligne
	set reference $SESHW($w,name)
	return [list REF $reference@$ligne]
    }
    
    proc getname {w} {
	variable SESHW
	return $SESHW($w,name)
    }

    proc get_scribe_code {w} {
	variable SESHW
	return $SESHW($w,scribe_code)
    }

    proc get_scribe {w} {
	variable SESHW
	return $SESHW($w,scribe)
    }

    proc get_comment {w} {
	variable SESHW
	return $SESHW($w,comment)
    }

    proc set_scribe_code {w n} {
	variable SESHW
	set SESHW($w,scribe_code) $n
    }

    proc set_scribe {w n} {
	variable SESHW
	set SESHW($w,scribe) $n
    }

    proc set_comment {w n} {
	variable SESHW
	set SESHW($w,comment) $n
    }

    proc copier {win} {
	variable SESHW
	global tcl_platform
	set w [base_window $win]
	if {$SESHW($w,selection)} {
	    set debut $SESHW($w,selectionstart)
	} else {
	    if {$SESHW($w,shift_selection)} {
		set debut $SESHW($w,shift_selectionstart)
	    } else return
	}
	# if { "$tcl_platform(platform)" == "unix" } {
	# 	    variable CUTBUFFER
	# 	    selection own $win
	# 	    set CUTBUFFER [$win get $debut endsel] 
	# 	} else {
	clipboard clear
	clipboard append [list MDC [$win get $debut endsel]]
	#}
    }

    proc couper {win} {
	variable SESHW
	set w [base_window $win]
	if {$SESHW($w,selection)} {
	    set debut $SESHW($w,selectionstart)
	} else {
	    if {$SESHW($w,shift_selection)} {
		set debut $SESHW($w,shift_selectionstart)
	    } else return
	}
	copier $win
	clear_select $win
	$win delete $debut endsel
	seshw::set_insert_at $win insert
    }
    
    proc seshgivesel {num1 num2} {
	variable CUTBUFFER
	return [string range $CUTBUFFER $num1 [expr $num1 + $num2 - 1]]
    }
    
    ################################################################
    # movement oriented stuff
    
    # sesh_depl --
    # args :
    # 	window
    #	where_to_go
    #
    # Moves the cursor, and keeps track of it.
    # (does some bookkeeping)
    
    proc update_insert {w} {
	variable SESHW	
	set SESHW($w,pos) [$w.s coords get insert]
	$w.s see insert
	if [info exists SESHW($w,onmove,command)] {
	    $SESHW($w,onmove,command) $w $SESHW($w,onmove,data)
	}
    }
    
    proc sesh_depl {win depl} {
	set_insert_at $win $depl
	clear_select $win
    }
    
    proc set_insert_at {win where} {
	set w [base_window $win]
	eval $win mark set insert $where
	update_insert $w
    }
    
    proc bind_onmove {w command data} {
	variable SESHW
	set SESHW($w,onmove,command) $command
	set SESHW($w,onmove,data) $data
    }
    
    ################################################################
    # loading stuff
    #
    #
    
    # loads a text from its reference. Should probably move !
    
    proc loadTextAndGotoRef {win oid ref} {
	variable SESHW
	set w [base_window $win]
	loadText $w $oid
	set_insert_at $win [list "/$ref"]
    }
    
    # Does'nt belong here in fact.
    proc displayText {oid ref} {
	set top .texte
	catch {
	    toplevel $top
	    create_editor $top.w -width 600 -height 400 -display_comments false
	    grid $top.w  -sticky news
	    grid rowconfigure $top 0 -weight 1
	    grid columnconfigure $top 0 -weight 1
	}
	loadTextAndGotoRef $top.w.s $oid $ref
	raise $top
	#traiter le cas des fichiers normaux.
    }
    
    proc savetext {w} {
	variable SESHW
	if {[string compare $SESHW($w,name) "nn rn=f"] != 0} {
	    seshtext::save $SESHW($w,name) [totext $w] [signlist $w]
	} else {
	    error [legendes::trans "the text you want to save should have a name"]
	}
    }

    # saveDirectory --
    # Save all texts in a directory.  Not logically part of seshw, but
    # by now, a sesh widget is needed for parsing.
    # Only texts in sesh (.hie extension) form will be saved. So, 
    # imported glyh and macscrobe texts should be saved beforehand.

    proc saveDirectory {w directory {progressMeter 0}} {
	set errlogs ""
	set files [glob $directory/*.hie]
	ProgressMeter::callSetMax $progressMeter [llength $files]

	set count 0
	foreach t $files {
	    incr count
	    ProgressMeter::callNotify $progressMeter $count
	
	    TkseshDB::begin [TkseshDB::getTkseshDB]
	    if [catch {
		set f [open $t]
		set txt [read $f]
		close $f
		fromtext $w $txt
		savetext $w
	    }] {
		append errlogs "$t : failed;"
	    }
	    TkseshDB::commit [TkseshDB::getTkseshDB]
	}
	ProgressMeter::callFinish $progressMeter
	return $errlogs
    }

    proc loadText {w oid} {
	set txt [seshtext::getMDCTextFromOid $oid]
	fromtext $w $txt
    }
    
    ## Text Loading proper. SHOULD STAY IN SESHW

    proc nextline {t} {
	upvar $t txt
	#set expr "^(\[^\n\r\]*)(\r?\n|\r)(.*)$"
	#regexp $expr $txt dummy line nothing txt
	set line [lindex $txt 0]
	set txt [lrange $txt 1 end]
	return $line
    }
    
    proc endoftext {txt} {
	return [expr [llength $txt] == 0]
	#return [expr [string compare $txt ""] == 0]
    }
    
    # copy a entire text in a sesh widget. Sets everything that should be set.
    proc fromtext {w txt {progress 0}} {
	variable SESHW
	
	set win $w.s
	clear_text $win
	
	set txt [split $txt "\n\r"]
	
	ProgressMeter::callSetMax $progress [llength $txt]
	set count 0

	while {! [endoftext $txt]} {
	    incr count
	    if {$count % 30 == 0} {
		ProgressMeter::callNotify $progress $count
	    }
	    
	    set ligne [nextline txt]
	    switch -regexp -- $ligne {
		{\+\+NAME (.*)\+s} {
		    regexp {\+\+NAME (.*)\+s} $ligne tmp SESHW($w,name)
		}
		{\+\+SCRIBE (.*)\+s} {
		    regexp {\+\+SCRIBE (.*)\+s} $ligne tmp SESHW($w,scribe)
		}
		{\+\+COMMENT (.*)\+s} {
		    regexp {\+\+COMMENT (.*)\+s} $ligne tmp SESHW($w,comment)
		}
		{\+\+SCRCOD (.*)\+s} {
		    regexp {\+\+SCRCOD (.*)\+s} $ligne tmp SESHW($w,scribe_code)
		}
		{\+\+COORDSYS } {
		    regexp {\+\+COORDSYS (.*)\+s} $ligne tmp pat
		    $win coords format $pat
		    set SESHW($w,coords,format) $pat
		}
		{\+\+COORD [1-9]} {
		    regexp {\+\+COORD ([1-9]) is (.*)\+s} $ligne tmp num pat
		    set ordre "$win coords type"
		    set SESHW($w,coords,$num) $pat
		    eval [concat $ordre $num $pat]
		}
		{\+\+SGN } {
		    if [regexp {\+\+SGN *([^ ]*) *(.*)\+s} $ligne tmp code val] {
			define_local_sign $w $code $val
		    }
		}
		{[-]?\+\+TXT} {
		    regexp -- {-?\+\+TXT(.*)} $ligne tmp content
		    set comm "-++TXT"
		    append comm $content
		    while {! [regexp {\+s} $content] && ! [endoftext $txt]} {
			set content [nextline txt]
			append comm $content
		    }
		    if {[regexp {\+s} $content]} {
			$win insert insert $comm
		    }
		}
		{default} {
		    $win insert insert $ligne
		}
	    }
	    
	}
	set_insert_at $win 1.0
	ProgressMeter::callFinish $progress
    }
    
    proc totext {w} {
	set result ""
	variable SESHW
	set result "++TKSESH DATABASE FILE+s\n"
	append result "++NAME $SESHW($w,name)+s\n"
	if { [string compare $SESHW($w,scribe) "nn rn=f"] != 0 } {
	    append result "++SCRIBE $SESHW($w,scribe)+s\n"
	}
	if { [string compare $SESHW($w,comment) ""] != 0 } {
	    append result "++COMMENT $SESHW($w,comment)+s\n"
	}
	if { [string compare $SESHW($w,scribe_code) "1234567890"] != 0 } {
	    append result "++SCRCOD $SESHW($w,scribe_code)+s\n"
	}
	foreach sign [array names SESHW "$w,sign,*"] {
	    set v $SESHW($sign)
	    append result "++SGN [lindex $v 0] "
	    append result [lindex $v 1]
	    append result "+s\n"
	}
	foreach i {1 2 3 4 5} {
	    if { $SESHW($w,coords,$i) != "" && 
	    $SESHW($w,coords,$i) != "NUM" } {
		append result "++COORD $i is $SESHW($w,coords,$i)+s\n"
	    }
	}
	if { $SESHW($w,coords,format) != "" } {
	    append result "++COORDSYS $SESHW($w,coords,format)+s\n"
	}
	
	append result [$w.s get 0.0 end]
    }

   
    # Returns a representation of the text as a list of signs.
    # The format is :
    # LINE_NUMBER TABULATION LIST OF SIGNS IN THE LINE SPACE \n
    # (Any sign is followed by a space)
    proc signlist {w} {
	set win $w.s
	$win mark set posi 1.0
	set txt "TKSSRC\t[getname $w]\n"
	while {[$win compare posi < end]} {
	    append txt "[$win coords get posi lineend]\t"
	    append txt [hieroglyphs::normalizeText [$win liste_signes posi posi lineend]]
	    append txt " \n"
	    $win mark set posi posi + 1 line
	}
	return $txt
    }

    ###############################################################################
    #                            Coordinates system
    
    
    proc set_coords_system {w} {
	variable SESHW
	set win $w.s
	$win coords format $SESHW($w,coords,format)
	foreach i {1 2 3 4 5} {
	    eval $win coords type $i $SESHW($w,coords,$i)
	}
    }

    ######################
    # edit_coords_system --
    #
    # arg : s hieroglyphic widget
    #
    # lauch an interactive editing window for hieroglyphic widget
    
    proc edit_coords_system {w} {
	variable SESHW
	catch {destroy .coords}
	toplevel .coords
	set edit .coords
	label $edit.l1 -text [legendes::trans "format" ]
	entry $edit.system -textvariable seshw::SESHW($w,coords,format)
	grid $edit.l1 - 
	grid $edit.system - 
	yellow::create $edit.system [legendes::trans "Format for the text coordinates\nin the current text\n\$1 means the first coordinate, and so on.\nHence \$2, \$1 means \"page number, then line number\""]
	label $edit.l21 -text [legendes::trans "coord"]
	label $edit.l22 -text [legendes::trans "values"]
	
	grid $edit.l21 $edit.l22
	foreach i {1 2 3 4 5} {
	    switch $i {
		1 {label $edit.c$i -text 1([legendes::trans "Line"])}
		2 {label $edit.c$i -text 2([legendes::trans "Page"])}
		default  {label $edit.c$i -text $i}
	    }
	    if { [array get SESHW $w,coords,$i] == "" } {
		set SESHW($w,coords,$i) NUM
	    }
	    entry $edit.v$i -textvariable seshw::SESHW($w,coords,$i)
	    yellow::create $edit.v$i \
		[legendes::trans "Possible values for this coordinate\nyou can enter NUM for numeric,\n or a list of string values\nSee year 400 Stela for an example"]
	    grid $edit.c$i $edit.v$i
	}
	button $edit.b -text [legendes::trans "set system"] -command [list seshw::set_coords_system $w]
	button $edit.c -text [legendes::trans "Close"] -command [list destroy $edit]
	grid $edit.b $edit.c
    }

    ################################################################
    # Enter coordinates
    # 
    #

    proc edit_coords {w} {
	variable SESHW
	set main .coordsedit
	catch {destroy $main}
	toplevel $main

	label $main.title -text [legendes::trans "Enter below the coordinate of the current point :\n"]
	grid $main.title -

	frame $main.f
	grid $main.f -

	# Find the variables used by the system :
	set l $seshw::SESHW($w,coords,format)
	regsub -all {\$([0-9])|.} $l {\1 } l
	regsub -all { +} $l { } l
	regsub -all { $} $l {} l
	regsub -all {^ } $l {} l
	#Now, l contains the list of used coordinates. 
	#Let's create editing windows for them
	foreach i $l {
	    if { [string compare $SESHW($w,coords,$i) "NUM"] == 0} {
		set SESHW($w,coords,value,$i) 1
		entry $main.e$i -textvariable seshw::SESHW($w,coords,value,$i) -width 4
	    } else {
		menubutton $main.e$i -menu $main.e$i.m \
		    -width 9 -justify left\
		    -takefocus 1 -textvariable seshw::SESHW($w,coords,value,$i)
		menu $main.e$i.m -tearoff 0
		foreach v $SESHW($w,coords,$i) {
		    $main.e$i.m add command -label $v -command [list set seshw::SESHW($w,coords,value,$i) $v]
		}
		set SESHW($w,coords,value,$i) [lindex $SESHW($w,coords,$i) 0]
	    }
	}
	# Now, let's build the visual representation :
	# 

	set labnum 0
	foreach t [get_coordinate_elements $w] {
	    incr labnum
	    if [string match {$[0-9]} $t ] {
		set num [string index $t 1]
		pack $main.e$num -side left -in $main.f
	    } else {
		label $main.l$labnum -text $t
		pack $main.l$labnum  -side left -in $main.f
	    }
	}
	
	button $main.ok -text [legendes::trans "Ok"]\
	    -command [list seshw::insert_coordinates_and_exit $w $main]
	button $main.cancel -text [legendes::trans "Cancel"]\
	    -command [list destroy $main ]
	grid $main.ok $main.cancel
    }
    
    # Sends back the list of elements in a text coordinate
    proc get_coordinate_elements {w} {
	variable SESHW
	regsub -all {(\$[0-9])} $seshw::SESHW($w,coords,format) "\t\\1\t" label
	regsub -all "\t+" $label "\t" label
	regsub -all "^\t" $label "" label
	regsub -all "\t$" $label "" label
	return [split $label "\t"]
    }

    proc insert_coordinates_and_exit {w edit} {
	insert_coordinates $w
	destroy $edit
    }

    proc insert_coordinates {w} {
	variable SESHW
	set l [get_coordinate_elements $w]
	foreach t $l {
	    if [string match {$[0-9]} $t ] {
		set i [string index $t 1]
		append result $SESHW($w,coords,value,$i)
	    } else {
		append result $t
	    }
	}
	set win [gethwindow $w]
	insert_text $win "-!\n++COORDS=$result+s"
    }

    ################################################################
    # creates a frame that contain various informations about
    # the text :
    
    proc create_info {w} {
	variable SESHW
	#The frame is included in the "title" part of the sesh widget.
	# IT WON'T work without a title part.
	set f $w.f1.info
	frame $f
	frame $f.f1
	set f1 $f.f1
	label $f1.l1 -text [legendes::trans "edited by: "]
	label $f1.scribe -textvariable seshw::SESHW($w,scribe)
	text $f.comment -height 2 -wrap word -yscrollcommand "$f.sb set"
	scrollbar $f.sb -command "$f.comment yview" -orient vertical
	# comment initialization
	$f.comment insert insert $SESHW($w,comment)

	grid $f.f1 -
	grid $f1.l1 $f1.scribe
	grid $f.comment $f.sb -sticky ew
	
	grid columnconfigure $f 0 -weight 1

	# To link the text widget with the variable, we 
	# use the trace function.

	set SESHW($w,commentedit) $f.comment
	trace variable seshw::SESHW($w,comment) r seshw::update_comment
	trace variable seshw::SESHW($w,comment) w seshw::update_comment
	bind $f <Destroy> [list seshw::destroy_info $w]
	
	grid $f -sticky ew
    }
    
    proc hide_info {w} {
	# Close the comment frame. Makes sure everything
	# is updated before.
	variable SESHW
	set f $w.f1.info.comment
	set txt [$f get 1.0 end-1c]
	regsub -all "\n" $txt " " txt
	set SESHW($w,comment) $txt
	destroy $w.f1.info
    }

    proc destroy_info {w} {
	variable SESHW
	unset SESHW($w,commentedit)
	trace vdelete SESHW($w,comment) r seshw::update_comment
	trace vdelete SESHW($w,comment) w seshw::update_comment
    }

    proc update_comment {name1 name2 op} {
	variable SESHW
	set f $SESHW(${name2}edit)
	switch -- $op {
	    r {
		set txt [$f get 1.0 end-1c]
		regsub -all "\n" $txt " " txt
		set SESHW($name2)  $txt
	    }
	    w {
		$f delete 1.0 end
		$f insert insert $SESHW($name2)
	    }
	}
    }

    ################################################################
    # Text entering 
    
    proc insert_text {win code} {
	set w [base_window $win]
	$win insert insert $code
	update_insert $w
    }
    
    proc addtobuffer {win code} {
	variable SESHW
	set w [base_window $win]
	#couper $win
	clear_select $win
	append SESHW($w,buffer) $code
    }
    
    proc reset_buffer {win {val ""}} {
	variable SESHW
	set w [base_window $win]
	set SESHW($w,buffer) $val
    }
    
    # Either add a new sign or try another possible sign
    proc tryNewSignOrRotateOld {win} {
	variable SESHW
	set w [base_window $win]
	if {[string compare $SESHW($w,buffer) "" ] == 0 } {
	    # If no new sign code is proposed, 
	    # change the last sign input.
	    possibleSignListNext $w
	    insertCurrentSignOrGroup $win
	} elseif {$SESHW($w,buffer) eq ":" || $SESHW($w,buffer) eq "*"} {
	    addSign $win
	    set SESHW($w,buffer) ""
	} else {
	    # if there is a new sign code, 
	    # add this sign
	    addSign $win
	}
    }
    

    
    # Adds a sign whose code corresponds to the content of buffer
    proc addSign {win} {
	variable SESHW
	variable PREFERREDSIGNPOS
	set w [base_window $win]

	clear_select $win

	# Test to find the separator to use :
	# the regexp cuts the code in buffer in :
	# separator (for the optionnal separator)
	# code (for the sign code)
	# modifier (for the sign modifier)
	
	set SESHW($w,separator) {-}
	regexp {([*:])?([a-zA-Z0-9'*]*)?(.*)} $SESHW($w,buffer) \
	    tmp SESHW($w,separator) code SESHW($w,modifier)

	# if no code was found :
	if {[string equal $code ""]} {
	    # if no separator found, do nothing,
	    # else apply the separator to the previous signs.
	    if {[string equal $SESHW($w,separator) ""]} {
		return
	    } else {
		$win mark set prevpos insert - 1 quad
		set SESHW($w,saved) [$win get insert - 1 quad insert]
		combineLastTwoCadrats $win
		return
	    }
	} else {
	    if {$SESHW($w,separator) == ""} {
		# If there is no separator, we are working on a new cadrat.
		set SESHW($w,saved) ""
		# as we will always work as if a sign had been previously
		# entered and was to be removed, we insert a dummy
		# sign now :
		$win mark set prevpos insert
		$win insert insert A
	    } else {
		# If there is a separator,
		# save the current cadrat.
		$win mark set prevpos insert - 1 quad
		set SESHW($w,saved) [$win get insert - 1 quad insert]
	    }
	    set SESHW($w,buffer) ""
	    # create list of possible signs.
	    possibleSignListBuild $w $code
	    # Insert the sign in the widget
	    insertCurrentSignOrGroup $win
	}
    }
    
    proc insertCurrentSignOrGroup {win} {
	variable SESHW

	set w [base_window $win]
	# Delete the previous attempt
	$win delete prevpos insert
	# Insert the current working cadrat.
	$win insert insert $SESHW($w,saved)
	set realCode [getPreferredSign $w]
	append realCode $SESHW($w,modifier)
	if {[ catch {addSignToCurrentCadrat $win $SESHW($w,separator) $realCode} ] != 0} {
	    # If insertion fails, we restore the current cadrat.
	    $win insert insert $SESHW($w,saved)
	}
	# Move and show the insertion cursor.
	update_insert $w
    }

    proc combineLastTwoCadrats {win} {
	variable SESHW
	set w [base_window $win]
	# Delete the previous attempt
	$win delete prevpos insert
	# Insert the current working cadrat.
	$win insert insert $SESHW($w,saved)
	if {[ catch {addSignToCurrentCadrat $win $SESHW($w,separator) ""} ] != 0} {
	    # If insertion fails, we restore the current cadrat.
	    $win insert insert $SESHW($w,saved)
	}
	# Move and show the insertion cursor.
	update_insert $w
    }
    
    # Gets the preferred sign for code in the current window.
    proc getPreferredSign {w} {
	variable SESHW
	# find a candidate sign ?
	if {[possibleSignListFinished $w]} {
	    # look in the dictionary if necessary.
	    if {! $SESHW($w,dictionaryExpanded)} {
		possibleSignListAddDictionary $w
		# there might be nothing in the dictionary anyway.
		if {[possibleSignListFinished $w]} {
		    possibleSignListReset $w
		}
	    } else {
		possibleSignListReset $w
	    }
	} 
	# return the sign
	return [possibleSignListGet $w]
    }

    proc possibleSignListBuild {w code} {
	variable SESHW
	variable PREFERREDSIGNPOS

	set SESHW($w,code) $code
	
	set possibleSigns {}
	
	# Add solutions from the glossary first, if any.
	if [Glossary::isAGlossaryCode $code] {
	    set possibleSigns [Glossary::getGlossaryEntry $code]
	}
	
	set possibleSigns [concat $possibleSigns [hieroglyphs::codeToSigns $code]]

	set SESHW($w,possibleSigns) $possibleSigns
	set SESHW($w,dictionaryExpanded) 0
	# If there is no list position for the current sign, select the first one.
	if {! [info exists PREFERREDSIGNPOS($code)]} {
	    set PREFERREDSIGNPOS($code) 0
	}
    }
    
    # advance to the next choice of sign for the current code.
    proc possibleSignListNext {w} {
	variable SESHW
	variable PREFERREDSIGNPOS

	set code $SESHW($w,code)
	# Get to the next sign for this code.
	incr PREFERREDSIGNPOS($code)
    }
    
    # returns true if the bounds of the possible sign list are past.
    proc possibleSignListFinished {w} {
	variable SESHW
	variable PREFERREDSIGNPOS

	set code $SESHW($w,code)
	# restart if we have already seen all possibilities.
	return [expr $PREFERREDSIGNPOS($code) >= [llength $SESHW($w,possibleSigns)]]
    }

    proc possibleSignListReset {w} {
	variable SESHW
	variable PREFERREDSIGNPOS
	set code $SESHW($w,code)
	set PREFERREDSIGNPOS($code) 0

    }
    
    # Gets the currently prefered possible sign for this code.
    proc possibleSignListGet {w} {
	variable PREFERREDSIGNPOS
	variable SESHW 
	set code $SESHW($w,code)
	set r [lindex $SESHW($w,possibleSigns) $PREFERREDSIGNPOS($code)]
	return $r
	
    }

    # Expands the dictionary data (if any) on the list of possible signs.
    proc possibleSignListAddDictionary {w} {
	variable PREFERREDSIGNPOS
	variable SESHW 
	set code $SESHW($w,code)

	set listdic [DicoDAL::readOrthography $code]
	set SESHW($w,possibleSigns) [concat $SESHW($w,possibleSigns) $listdic]
	set SESHW($w,dictionaryExpanded) 1
    }
    

    # plainInsertSign
    # plain insertion of a sign.
    
    proc plainInsertSign {win code} {
	if {[string equal $code ""]} {
	    return
	}
	set w [base_window $win]
	if [getRetrograde $w] {
	    append code "\\"
	}
	$win insert insert $code
    }

    proc addSignToCurrentCadrat {win sep code} {
	if {$sep == {:}} {
	    plainInsertSign $win $code
	    set toto [join [list [$win get insert - 2 quad insert - 1 quad] \
		    [$win get insert - 1 quad insert]] \
		    {:}]
	    $win delete insert insert - 2 quad
	    $win insert insert $toto
	} elseif {$sep == {*}} {
	    plainInsertSign $win $code
	    if [$win isvertical insert] {
		set toto1 [$win explode insert - 1 quad]
		set toto2 [$win get insert - 1 quad insert]
		set laste [lindex $toto1 end]
		set laste [join [list $laste $toto2] {*}]
		set toto [join [lreplace $toto1 end end $laste] {:}]
	    } else {
		set toto [join [list [$win get insert - 2 quad insert - 1 quad] \
			[$win get insert - 1 quad insert]] \
			{*}]
	    }
	    $win insert insert $toto
	    $win delete insert - 1 quad insert - 3 quad
	    $win mark set insert insert + 1 quad
	} else {
	    plainInsertSign $win $code
	}
    }
    

    
    # Text destruction
    
	proc backspace {win} {
	variable SESHW
	set w [base_window $win]
	set n [string  length $SESHW($w,buffer)]
	if [expr $n == 0] {
	    $win delete insert insert - 1 quad
	} else {
	    set  SESHW($w,buffer) [string range $SESHW($w,buffer) 0 [expr $n - 2]]
	}
	update_insert $w
    }
    
    ################################################################
    # Various complex constructions
    
    proc try_contruct {win c1 c2 regexp} {
	set deb [get_start_selection $win]
	set tmp [$win get $deb endsel]
	if {! [regexp $regexp $tmp ]} {
	    $win delete $deb endsel
	    if {[catch {$win insert endsel "$c1$tmp$c2"}] != 0} {
		$win insert endsel $tmp
	    }
	}
	clear_select $win
	seshw::set_insert_at $win insert
    }
    
    proc cartouche {win} {
	try_contruct $win {<-} {->} {<.*>}
    }

    proc startCartouche {win} {
	try_contruct $win {<b-} {->} {<.*>}
    }

    proc endCartouche {win} {
	try_contruct $win {<e-} {->} {<.*>}
    }

    proc middleCartouche {win} {
	try_contruct $win {<m-} {->} {<.*>}
    }

    proc serekh {win} {
	try_contruct $win {<S-} {->} {<.*>}
    }

    proc hout {win} {
	try_contruct $win {<H-} {->} {<.*>}
    }

    
    proc crochets {win {c1 {[[-}} {c2 {-]]}}} {
	try_contruct $win $c1 $c2 {\[.*\]}
    }
    
    proc hachure {win} {
	try_contruct $win \#b- -\#e \#b-.*-\#e
    }
    
    ###############################################################################
    # Hatchures 
    # hatchure the current cadrat 
	
    proc hachure_cadrat {win hatch} {
	set q [$win get insert insert - 1 quad]
	$win delete insert insert - 1 quad
	if [catch {
	    regsub \#\[0-9\]*\$ $q {} q
	    $win insert insert ${q}\#$hatch
	}] {
	    $win insert insert $q
	}
    }
	
    ###############################################################################
    # Fonctions diverses: 
    # passage de texte en rouge.
    
    
    proc enrouge {win} {
	set deb [get_start_selection $win]
	set texte [$win get $deb endsel]
	regsub -all {\$r\-|\$b\-} $texte {} texte
	set texte "\$r-$texte-\$b"
	$win delete $deb endsel
	$win insert endsel $texte
	clear_select $win
	seshw::set_insert_at $win insert
    }
    
    proc ennoir {win} {
	set deb [get_start_selection $win]
	set texte [$win get $deb endsel]
	regsub -all {\$r\-|\$b\-} $texte {} texte
	$win delete $deb endsel
	$win insert endsel $texte
	clear_select $win
	set_insert_at $win insert
    }
    
    
    # SHOULD BE MODIFIED TO TAKE A FEW THINGS INTO ACCOUNT -- which things ? 
    # I don't remember. Seems to work fine ? --

    proc explode_cadrat {win} {
	regsub -all "\n" [$win explode insert] {-} texte
	$win delete insert - 1 quad insert
	$win insert insert $texte
    }
    
    # given a text, sends back two lists : 
    # a all modifiers like +r ...
    proc  get_modifiers {txt} {
	error "not done yet"
    }

    #############################################################
    # Retrograde writting handling

    proc getRetrograde {w} {
	variable SESHW
	return $SESHW($w,retro)
    }

    proc toggleRetrograde {w} {
	variable SESHW
	set SESHW($w,retro) [expr 1 - [getRetrograde $w]]
    }

    #############################################################
    # column writing

    proc getInColumns {w} {
	variable SESHW
	return $SESHW($w,in_columns)
    }

    proc setInColumns {w col} {
	variable SESHW
	set SESHW($w,in_columns) $col
	[gethwindow $w] configure -in_columns $col
    }

    proc toggleColumns {w} {
	if {[getInColumns $w]} {
	    setInColumns $w 0 
	} else {
	    setInColumns $w 1
	}
    }
		       
    
    #############################################################
    # Centering signs

    proc getSignCentering {w} {
	set win [seshw::gethwindow $w]
	return [$win cget -vertical_centering]
    }

    proc toggleSignCentering {w} {
	set win [seshw::gethwindow $w]
	$win configure -vertical_centering [expr 1 - [getSignCentering $w]]
    }

    #############################################################
    # groups the text between 'insert' and 'endsel' marks 
    
    proc grh {win} {
	set deb [get_start_selection $win]
	set txt [$win listecadrats $deb endsel]
	regsub -all "\n" $txt {)*(} texte
	set texte "($texte)"
	$win delete $deb endsel
	$win insert endsel $texte
	clear_select $win
	set_insert_at $win insert
    }
    
    proc grv {win} {
	set deb [get_start_selection $win]	
	set txt [$win listecadrats $deb endsel]
	regsub -all "\n" $txt {):(} texte
	set texte "($texte)"    
	$win delete $deb endsel
	$win insert endsel $texte
	clear_select $win
	set_insert_at $win insert
    }
    
    proc grlig {win} {
	set deb [get_start_selection $win]	
	set txt [$win listecadrats $deb endsel]
	regsub -all "\n" $txt {\&} texte
	set texte "($texte)"
	$win delete $deb endsel
	$win insert endsel $texte
	clear_select $win
	set_insert_at $win insert	
    }
    
    ###############################################################################
    # Fonctions diverses: 
    # selection d'un signe.
    
    proc select_next_sign {win} {
	set nb [$win nbr_signes insert]
	set sel 0
	for {set i 1} {$sel == 0 && $i <= $nb} {incr i} {
	    if {[$win undocumented3 insert .$i]} {
		$win sign  -select 0 insert .$i
		set sel $i
	    }
	}
	incr sel
	if {$sel <= $nb} {
	    $win sign -select 1 insert .$sel
	} else {
	    $win mark set insert insert + 1 quad
	    set nb [$win nbr_signes insert]
	    set sel 0
	    for {set i 1} {$sel == 0 && $i <= $nb} {incr i} {
		if {[$win undocumented3 insert .$i]} {
		    $win sign -select 0 insert .$i 
		    set sel $i
		}
	    }
	    incr sel
	    if {$sel <= $nb} {
		$win sign -select 1 insert .$sel 
	    }
	}
    }

    ################################################################
    # Latin text functions
    # 
    
    proc winlatininsert {win} {
	frame $win.fr
	set b $win.fr
	scan [$win xypos insert] {%d %d} x y
	place $b -x 0 -y $y -anchor sw -relwidth 1.0
	entry $b.e
	bind $b.e <Return> [list seshw::latininsert $win]
	focus -force $b.e
	grid $b.e -sticky news
	grid columnconfigure $b 0 -weight 1
    }


    proc latininsert {win} {
	set code "+l[$win.fr.e get]"
	insert_text $win $code
	destroy $win.fr
	focus -force $win	
    }

    proc winnumlineinsert {win} {
	frame $win.fr
	set b $win.fr
	scan [$win xypos insert] {%d %d} x y
	place $b -x 0 -y $y -anchor sw -relwidth 1.0
	entry $b.e
	bind $b.e <Return> [list seshw::numlineinsert $win]
	focus -force $b.e 
	grid $b.e -sticky news
	grid columnconfigure $b 0 -weight 1
    }

    proc numlineinsert {win} {
	set code "|[$win.fr.e get]"
	insert_text $win $code
	destroy $win.fr
	focus -force $win
    }
    
    ################################################################
    # Sign modification commands
    # 

    proc toggle_revert_sign {win} {
	$win sign -revert c insert
    }

    proc toggle_sign {win what} {
	switch -- $what {
	    ignore {$win sign -ignore c insert}
	    wide {$win sign -long 1 insert}
	}
    }

    proc toggle_gram {win} {
	$win sign -gram c insert
    }

    proc toggle_word {win} {
	$win sign -word c insert
    }
	
    proc set_sign_size {win size} {
	$win sign -size $size insert
    }
    
    proc set_sign_rotation {win rot} {
	$win sign -rot $rot insert
    }

    ################################################################
    # Signs definitions local to the widget
	
    proc define_local_sign {w code def} {
	variable SESHW
	set SESHW($w,sign,$code) [list $code $def]
	# Global definition at the time being. Should be local
	tksesh_defchar $code $def	
    }
	

    ################################################################
    # Bindings
    
    ################################################################
    proc nothing {} {}
    # For passive widgets
    
    
    bind PassiveSesh <<Paste>> {
	seshw::clear_text %W
	seshw::paste_selection %W 
	seshw::set_insert_at %W 0.0
	seshw::autosize %W
    }
    
    ################################################################
    # For active widgets
    
    ################################################################
    # Viewer
    # Movement
    bind ViewSesh <Left> {seshw::sesh_depl %W "insert - 1 quad"}
    bind ViewSesh <Right> {seshw::sesh_depl  %W "insert + 1 quad"}
    bind ViewSesh <Home> {seshw::sesh_depl %W "insert linestart"}    
    bind ViewSesh <End> {seshw::sesh_depl %W "insert lineend"}    
    bind ViewSesh <Up> {seshw::sesh_depl %W "insert - 1 line"}    
    bind ViewSesh <Down> {seshw::sesh_depl %W "insert + 1 line"}
    
    bind ViewSesh <Shift-Left> {seshw::sesh_arrow_sel %W "insert - 1 quad"}
    bind ViewSesh <Shift-Right> {seshw::sesh_arrow_sel %W "insert + 1 quad"}

    bind ViewSesh <Control-l> {seshw::update_insert [seshw::base_window %W]}
    
    # Cut and marks
    bind ViewSesh <<Copy>> {
	seshw::copier %W
    }
    
    bind ViewSesh <Control-space> {%W mark set endsel insert}

    bind ViewSesh <Key-Escape> {seshw::nothing}

    bind ViewSesh <Key-Escape><Control-x> {
	%W mark set tmpxx insert
	seshw::set_insert_at %W endsel
	%W mark set endsel tmpxx
	break
    }
    
    bind ViewSesh <Button-1> {
	focus %W
	seshw::set_insert_at %W "@%x,%y"
	seshw::start_selection %W
    }
    
    bind ViewSesh <Double-Button-1> {
	seshw::set_insert_at %W "@%x,%y"
	%W mark set endsel insert
    }
    
    bind ViewSesh <B1-Motion> {
	seshw::end_selection %W @%x,%y
	break
    }
    
    bind ViewSesh <Shift-Button-1> {
	seshw::end_selection %W @%x,%y
	break
    }
    
    bind ViewSesh <Button-3> {
	seshw::start_selection %W insert
	seshw::end_selection %W @%x,%y
	selection own %W
	set seshw::CUTBUFFER [%W get insert endsel]
	seshw::set_insert_at %W "@%x,%y"
    }
    
    bind ViewSesh <Control-s> {
	seshw::select_next_sign %W
    }
    
    
    ################################################################
    # Editor
    
    # Inserting text 

    bind EditorSesh <Key> {seshw::addtobuffer %W  %A}
    bind EditorSesh <Key-F10> {seshw::nothing}
    bind EditorSesh <Key-Control_L>  {seshw::nothing}
    bind EditorSesh <Key-Control_R>  {seshw::nothing}
    bind EditorSesh <Key-Escape>  {seshw::nothing}
    bind EditorSesh <Key-Tab>  {seshw::nothing}
    bind EditorSesh <Key-space> {seshw::tryNewSignOrRotateOld %W}
    bind EditorSesh <Key-minus> {seshw::addSign %W}
    bind EditorSesh <Key-colon> {
	seshw::addSign %W
	seshw::reset_buffer %W ":"
    }
    
    bind EditorSesh <Key-asterisk> {
	seshw::addSign %W
	seshw::reset_buffer %W "*"
    }
    
    bind EditorSesh <Return> {
	seshw::addSign %W
	seshw::insert_text %W {-!}
    }

    bind EditorSesh <Shift-Return> {
	seshw::addSign %W
	seshw::insert_text %W {-!!}
    }
	
    bind EditorSesh <BackSpace> {
	seshw::backspace %W
    }

    bind EditorSesh <comma> {
	seshw::insert_text %W {`}
    }
    
    bind EditorSesh <semicolon> {
	seshw::insert_text %W {..}
    }
    
    bind EditorSesh <period> {
	seshw::insert_text %W {.}
    }
    
    bind EditorSesh <slash> {
	catch {seshw::toggle_word %W}
    }
    
    bind EditorSesh <equal> {
	catch {seshw::toggle_gram %W}
    }

    bind EditorSesh <Key-Escape><Key-e> {
	catch {seshw::explode_cadrat %W}
    }

    bind EditorSesh <Key-Escape><Key-h> {
	seshw::grh %W
	break
    }

    bind EditorSesh <Key-Escape><Key-l> {
	seshw::grlig %W
	break
    }

    bind EditorSesh <Key-Escape><Key-v> {
	seshw::grv %W
	break
    }

    bind EditorSesh <Key-Escape><Key-c> {
	catch {seshw::cartouche %W}
    }

    bind EditorSesh <Key-Escape><Key-z> {
	catch {seshw::hachure %W}
    }

    bind EditorSesh <Key-Escape><Key-parenleft> {
	catch {seshw::crochets %W}
    }

    bind EditorSesh <Key-Escape><Key-r> {
	catch {seshw::enrouge %W}
    }

    bind EditorSesh <Key-Escape><Key-b> {
	catch {seshw::ennoir %W}
    }

    bind EditorSesh <Control-q> {
	if {[%W cget -display_comments]}    {
	    %W configure -display_comments false
	} else {
	    %W configure -display_comments true
	}
    }

    bind EditorSesh <Left> {seshw::nothing}
    bind EditorSesh <Right> {seshw::nothing}
    bind EditorSesh <Alt-Left> {seshw::nothing}    
    bind EditorSesh <Alt-Right> {seshw::nothing}    
    bind EditorSesh <Shift-Left> {seshw::nothing}    
    bind EditorSesh <Shift-Right> {seshw::nothing}
    bind EditorSesh <Up> {seshw::nothing}    
    bind EditorSesh <Down> {seshw::nothing}
    
    bind EditorSesh <Control-l> {seshw::nothing}

    # Cut and marks
    bind EditorSesh <<Copy>> {seshw::nothing}
    
    bind EditorSesh <Control-Key-space> {seshw::nothing}
    
    bind EditorSesh <Key-Escape> {seshw::nothing}
    
    bind EditorSesh <Button-1> {seshw::nothing}
    
    bind EditorSesh <Double-Button-1> {seshw::nothing}
    
    bind EditorSesh <B1-Motion> {seshw::nothing}
    
    bind EditorSesh <Shift-Button-1> {seshw::nothing}
    
    bind EditorSesh <Button-3> {seshw::nothing}
    
    bind EditorSesh <Control-s> {seshw::nothing}
    
    
    # pasting 
    
    bind EditorSesh <<Paste>> {
	seshw::paste_selection %W 
	
    }

    bind EditorSesh <<Cut>> {
	seshw::couper %W 
    }

    bind LatinMode <Key> {seshw::latinkey %W %A; break}

    bind LatinMode <Return> {seshw::latininsert %W; break}

    bind LatinMode <BackSpace> {
	seshw::latinbackspace %W
    }
    
    ################################################################
    # For all widgets
}


