package require tclSQLDB
package require myStoopWidgets

# A browser widget.  
#
# If useFilter is true, the datasource should use an external field
# called "start" to indicate the first displayed position. Note that this can have 

package provide tkDB 1.0

class BDBBrowser {

    proc BDBBrowser {this parent name dataSource columns args} \
	SFrame {$parent $name -takefocus 1 $args} {
	    set ($this,dataSource) $dataSource
	    set ($this,columns) $columns
	    set ($this,useFilter) 0
	    set ($this,caseSensitive) 1
	}

    proc ~BDBBrowser {this} {
    }

    proc setCaseSensitive {this case} {
	set ($this,caseSensitive) $case
    }
    
    # completorHook's job is to give this widget a path
    virtual proc completorHook {this} {
	#set options [list -columns $($this,colDecls)]
	set path [Component::getPath $this]
	set ($this,entry) [Component::buildNewComponent SEntry $this -auto]
	set ($this,scroll) [Component::buildNewComponent SScrolledWindow $this -auto]
	set ($this,list) [Component::buildNewComponent BDBListBox $this -auto \
			      $($this,dataSource) $($this,columns)]
	
	#Observable::add $($this,dataSource) $this
	
	SScrolledWindow::setWidget $($this,scroll) $($this,list)
	
	Component::grid $($this,entry) -sticky news
	Component::grid $($this,scroll) -sticky news
	Component::rowConfigure $this 1 -weight 1
	Component::columnConfigure $this 0 -weight 1


	Component::bind $($this,list) <Double-Button-1> [list BDBBrowser::doCommand $this]
	#Component::bind $($this,list) <Button-1> [list focus [Component::getPath $($this,list)]]
	Component::bind $this <Button-1> [list focus [Component::getPath $($this,list)]]
	Component::bind $this <Enter> [list focus [Component::getPath $this]]
	
	Component::bind $($this,list) <Return> [list BDBBrowser::doCommand $this]
	Component::bind $($this,list) <KeyPress> [list BDBBrowser::addChar $this %A]
	Component::bind $this <KeyPress> [list BDBBrowser::addChar $this %A]
	Component::bind $($this,entry) <Return> [list BDBBrowser::setStart $this]

	#Observable::add $($this,dataSource) $this
    }

    proc doCommand {this} {
	puts "index is [getIndex $this]"
	eval $($this,command) [getIndex $this]
    }

    # setCommand --
    #   declares the command to evaluate when an item is selected.
    #		the command will be *evaluated*, with the index of the selected
    #		item in the datasource passed as argument.
    # 	command : the command to *eval*

    proc setCommand {this command} {
	set ($this,command) $command
    }

    proc getCommand {this command} {
	return $($this,command)
    }

    proc useFilter {this useFilter} {
	set ($this,useFilter) $useFilter
    }

    proc setStart {this} {
	set start [SEntry::get $($this,entry)]
	if {$($this,useFilter)} {
	    catch {
		DataSource::setExternalCol $($this,dataSource) "start" $start
	    }
	} else {
	    # Find the first corresponding entry :
	    set i 0
	    set notfound 1
	    while {$notfound && $i < [DataSource::getSize $($this,dataSource)]} {
		set v [DataSource::getColIdx $($this,dataSource) $i $($this,columns)]
		if {$($this,caseSensitive)} {
		    set match [string match "$start*" $v]
		} else {
		    set match [string match -nocase "$start*" $v]
		}
		if {$match} {
		    set notfound 0
		    puts "found $i"
		    #BDBListBox::setFirstIndex $($this,list) $i
		    BDBListBox::setCurrent $($this,list) $i
		} else {
		    incr i
		}
	    }
	}
    }
    
    proc getDataSource {this} {
	return $($this,dataSource)
    }

    proc getIndex {this} {
	return [BDBListBox::getIndex $($this,list)]
    }

    # The "pre" filter is used to solve the following problem : a
    # field in the dataSource may have a different ordering than the
    # normal one.  (for instance, Manuel de Codage transliteration
    # code is not ordered along the ASCII order). To solve this, one
    # may add a "sort key" to the entries, the key being computed
    # when the entry is created. But in this case, one needs to transform
    # the entry field content before using it to sort the data.

    proc setpreFilter {} {
    }
    
    proc setPostFilter {} {
    }

    proc setFont {this font} {
	Component::cset $($this,entry) font $font
	Component::cset $($this,list) font $font
    }
    
    proc addChar {this char} {
	puts "here $this $char"
	switch -- $char {
	    "" {}
	    "\b" {
		set t [SEntry::get $($this,entry)]
		set t [string range $t 0 end-1]
		SEntry::setText $($this,entry) $t
		setStart $this
	    }
	    "\r" {
		BDBBrowser::doCommand $this
	    }
	    default {
		SEntry::insert $($this,entry) end $char
		setStart $this
	    }
	}
    }
}