package require tclSQLDB
package require myStoopWidgets

# A widget for editing a field in a DataSource.
# The field is edited through a combobox,
# The source of whom may be either a datasource, a variable, or a list
package provide tkDB 1.0

class BDBComboBox {

    #
    # parent : the parent widget
    # name : name for this widget ("auto" lets the system name the widget)
    # args : various args to pass.
    #
    # dataSource : the DataSource the field belongs to
    # field : the field to display
    #
    
    proc BDBComboBox {this parent name dataSource field args} SComboBox {$parent $name 
	 -modifycmd [list BDBComboBox::doValidate $this] $args} Observer {} {
	     set ($this,dataSource) $dataSource
	     set ($this,field) $field
	     # Refreshing is used to avoid circular dependencies
	     set ($this,refreshing) 0
	}
    
    proc ~BDBComboBox {this} {
    }

    proc setListDataSource {this dataSource field} {
	set ($this,listDataSource) $dataSource
	set ($this,sourceField) $field
	set ($this,sourceType) "dataSource"
	PRIVRefreshListSource $this
	Observable::add $($this,listDataSource) $this
    }

    proc setListSourceList {this list} {
	set ($this,sourceType) "list"
	set ($this,list) $list
	PRIVRefreshListSource $this
    }

    proc setListSourceVariable {this variableName} {
	set ($this,sourceType) "variable"
	set ($this,variable)  $variableName
	PRIVRefreshListSource $this
    }
    
    virtual proc update {this observable msg} {
	if {$observable == $($this,dataSource)} {
	    switch -- $msg {
		"ObservableDeleted" {}
		"IndexChanged" -
		"dataChanged" -
		"rowDeleted" {doRefresh $this}
		"rowChanged" {}
		"rowCreated" {}
	    }
	}
	# No else, for both datasources may be the same.
	
	if {$($this,sourceType) eq "dataSource" && 
	    $observable == $($this,listDataSource)} {
	    PRIVRefreshListSource $this
	}
    }
    
    proc PRIVRefreshListSource {this} {
	switch -- $($this,sourceType) {
	    list {
	    }
	    dataSource {
		set ds $($this,listDataSource)
		set r {}
		for {set i 0} {$i < [DataSource::getSize $ds]} {incr i} {
		    lappend r [DataSource::getColIdx $ds $i $($this,sourceField)]
		}
		set ($this,list) $r
	    }
	    variable {
		global $($this,variable)
		set ($this,list) [set $($this,variable)]
	    }
	}
	Component::cset $this values $($this,list)
    }

    # doRefresh is called when the value of the data is modified by external sources.
    proc doRefresh {this} {
	if {! $($this,refreshing)} {
	    set ($this,refreshing) 1
	    set error [catch {
		set ($this,value) [DataSource::getCol $($this,dataSource) $($this,field)] 
	    } errorMsg]
	    if {$error} {
		set ($this,value) ""
	    }
	    # fix the editable state :
	    # set editable [DataSource::canDo $($this,dataSource) setCol $($this,field)]
	    # Component::cset $this editable $editable
	    set ($this,refreshing) 0
	}
    }
    
    proc doValidate {this} {
	if {! $($this,refreshing)} {
	    set ($this,refreshing) 1
	    puts "validating $($this,value)"
	    DataSource::setCol $($this,dataSource) $($this,field) $($this,value)
	    set ($this,refreshing) 0
	}
	return 1
    }
    
    virtual proc completorHook {this} {
	Component::cset $this textvariable BDBComboBox::($this,value)
	Observable::add $($this,dataSource) $this
	setListSourceList $this {a b c d}
	doRefresh $this
    }

}

